Package: notame
Type: Package
Title: Workflow for non-targeted LC-MS metabolic profiling
Version: 1.3.0
Authors@R: c(
    person("Anton", "Klåvus", role = c("aut", "cph"),
           comment = c(ORCID = "0000-0003-2612-0230")),
    person("Jussi", "Paananen", role = c("aut", "cph"),
           comment = c(ORCID = "0000-0001-5100-4907")),
    person("Oskari", "Timonen", role = c("aut", "cph"),
           comment = c(ORCID = "0000-0002-6317-6260")),
    person("Atte", "Lihtamo", role = "aut"),
    person("Vilhelm", "Suksi", email = "vksuks@utu.fi", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0005-1108-518X")),
    person("Retu", "Haikonen", role = "aut",
           comment = c(ORCID = "0000-0003-0830-3850")),
    person("Leo", "Lahti", role = "aut",
           comment = c(ORCID = "0000-0001-5537-637X")),
    person("Kati", "Hanhineva", role = "aut",
           comment = c(ORCID = "0000-0001-6834-7375")),
    person("Ville", "Koistinen", role = "ctb",
           comment = c(ORCID = "0000-0003-1587-8361")),
    person("Olli", "Kärkkäinen", role = "ctb",
           comment = c(ORCID = "0000-0003-0825-4956")),
    person("Artur", "Sannikov", role = "ctb",
           comment = c(ORCID = "0000-0001-7765-123X")))
Description: Provides functionality for untargeted LC-MS metabolomics
        research as specified in the associated protocol article in the
        'Metabolomics Data Processing and Data Analysis—Current Best
        Practices' special issue of the Metabolites journal (2020).
        This includes tabular data preprocessing and quality control,
        uni- and multivariate analysis as well as quality control
        visualizations, feature-wise visualizations and results
        visualizations. Raw data preprocessing and functionality
        related to biological context, such as pathway analysis, is not
        included.
License: MIT + file LICENSE
Encoding: UTF-8
biocViews: BiomedicalInformatics, Metabolomics, DataImport,
        MassSpectrometry, BatchEffect, MultipleComparison,
        Normalization, QualityControl, Visualization, Preprocessing
Depends: R (>= 4.5.0), ggplot2, SummarizedExperiment
Imports: BiocGenerics, BiocParallel, dplyr, futile.logger, methods,
        openxlsx, S4Vectors, scales, stringr, tidyr, utils
Suggests: BiocStyle, fpc, igraph, knitr, missForest, notameViz,
        notameStats, pcaMethods, RUVSeq, testthat
URL: https://github.com/hanhineva-lab/notame,
        https://hanhineva-lab.github.io/notame/
BugReports: https://github.com/hanhineva-lab/notame/issues
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Config/testthat/parallel: true
Config/pak/sysreqs: libicu-dev zlib1g-dev
Repository: https://bioc.r-universe.dev
Date/Publication: 2026-04-28 13:05:37 UTC
RemoteUrl: https://github.com/bioc/notame
RemoteRef: HEAD
RemoteSha: d7eaa1a10cbc791a51d28fabd0db77a942682bc1
NeedsCompilation: no
Packaged: 2026-05-22 09:52:51 UTC; root
Author: Anton Klåvus [aut, cph] (ORCID:
    <https://orcid.org/0000-0003-2612-0230>),
  Jussi Paananen [aut, cph] (ORCID:
    <https://orcid.org/0000-0001-5100-4907>),
  Oskari Timonen [aut, cph] (ORCID:
    <https://orcid.org/0000-0002-6317-6260>),
  Atte Lihtamo [aut],
  Vilhelm Suksi [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-1108-518X>),
  Retu Haikonen [aut] (ORCID: <https://orcid.org/0000-0003-0830-3850>),
  Leo Lahti [aut] (ORCID: <https://orcid.org/0000-0001-5537-637X>),
  Kati Hanhineva [aut] (ORCID: <https://orcid.org/0000-0001-6834-7375>),
  Ville Koistinen [ctb] (ORCID: <https://orcid.org/0000-0003-1587-8361>),
  Olli Kärkkäinen [ctb] (ORCID: <https://orcid.org/0000-0003-0825-4956>),
  Artur Sannikov [ctb] (ORCID: <https://orcid.org/0000-0001-7765-123X>)
Maintainer: Vilhelm Suksi <vksuks@utu.fi>
Built: R 4.6.0; ; 2026-05-22 09:56:10 UTC; unix

# Generated by roxygen2: do not edit by hand

export("flag<-")
export(assess_quality)
export(citations)
export(cluster_features)
export(combined_data)
export(compress_clusters)
export(correct_drift)
export(drop_flagged)
export(drop_qcs)
export(finish_log)
export(fix_MSMS)
export(fix_object)
export(flag)
export(flag_contaminants)
export(flag_detection)
export(flag_quality)
export(flag_report)
export(import_from_excel)
export(import_from_msdial)
export(impute_rf)
export(impute_simple)
export(init_log)
export(inverse_normalize)
export(join_colData)
export(join_rowData)
export(log_text)
export(mark_nas)
export(merge_notame_sets)
export(pca_bhattacharyya_dist)
export(perform_repeatability)
export(pqn_normalization)
export(quality)
export(ruvs_qc)
export(write_to_excel)
import(BiocGenerics)
import(SummarizedExperiment)
import(ggplot2, except = Position)
import(methods)
importFrom(utils,citation)

# notame - Workflow for non-targeted LC-MS metabolic profiling <img src="man/figures/notame_logo.png" align="right" width="120" />

<!-- badges: start -->

[![Platforms](https://bioconductor.org/shields/availability/devel/notame.svg)](https://bioconductor.org/packages/devel/bioc/html/notame.html)
[![rworkflows](https://github.com/hanhineva-lab/notame/actions/workflows/rworkflows.yml/badge.svg?branch=devel)](https://github.com/hanhineva-lab/notame/actions)
[![Bioc-release](http://bioconductor.org/shields/build/devel/bioc/notame.svg)](http://bioconductor.org/packages/devel/bioc/html/notame.html)
[![Bioc-age](http://bioconductor.org/shields/years-in-bioc/notame.svg)](https://bioconductor.org/packages/devel/bioc/html/notame.html)
[![Dependencies](https://bioconductor.org/shields/dependencies/devel/notame.svg)](https://bioconductor.org/packages/devel/bioc/html/notame.html)

<!-- badges: end -->

The notame packages can be used to analyze preprocessed LC-MS data in non-targeted metabolomics. Notame was developed at the [research group of nutritional metabolomics at University of Eastern Finland](https://hanhinevalab.com/home) and [Afekta Technologies](https://afekta.com/), a spinoff metabolomics company. We use notame as a way to bundle together all the preprocessing methods we use for our non-targeted LC-MS metabolomics data, so it mainly consists of methods found in other packages, and a bunch of visualizations we have found useful.

For more detailed information on how we run our LC-MS experiments and where this package fits in our workflow, you can find the paper here: ["notame": Workflow for Non-Targeted LC-MS Metabolic Profiling](https://www.mdpi.com/2218-1989/10/4/135). A huge thank you for everyone involved in the paper!

Currently, notame is developed by Afekta Technologies and was reworked for Bioconductor to promote interoperability with functionality that can complement the notame workflow. This resulted in some breaking changes (see NEWS for user-facing changes). See the vignettes and documentation for more information, for example on the [website](https://hanhineva-lab.github.io/notame/). 

### What does notame do?
Before we go into the list of features, it is good for you to know how the workflow in our lab works. The first step is to take raw data files created by the LC-MS instrument and create a peak table using a peak picking software (we use [MS-DIAL](https://systemsomicslab.github.io/compms/msdial/main.html)). After peak picking with the dedicated software, we use R for data preprocessing, quality control, statistical analysis and visualization. We then use the obtained results in identification of the actual metabolites. During the years, we ended up with various scripts that were hard to handle and update, so we decided to make notame to keep things under control. 

Here is a list of the current main functionalities of notame:

- Reading data from Excel spreadsheets created with MS-DIAL
- Data is stored in a SummarizedExperiment object that holds all the information about the features and samples along with the feature abundance matrix. This allows for a simple interface for all of the functions in notame, as there is no need to juggle with different matrices/data frames.
- Drift correction: correcting for systematic drift in the intensity of molecular features using cubic spline correction (see [Kirwan & Broadhurst et al.](https://doi.org/10.1007/s00216-013-6856-7))
- Identifying and flagging (or removing) low-quality molecular features using quality metrics defined by [Broadhurst et al.](https://doi.org/10.1007/s11306-018-1367-3)
- Imputing missing values, multiple strategies available. Random forest imputation recommended, see [Kokla et al.](https://doi.org/10.1186/s12859-019-3110-0)
- Batch effect correction: correcting for systematic variation between batches
- A novel method for clustering similar molecular features
- A bunch of statistical analyses, both feature-wise tests and multivariate models
- A rather nice set of visualizations for use in quality control, explorative analysis and interpretation of results from statistical tests


## Installation and getting started

### Bioc-release

```
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("notame")
```

### Bioc-devel

```
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# The following initializes usage of Bioc devel
BiocManager::install(version='devel')

BiocManager::install("notame")
```

## Credits and license

The first version of notame was written by Anton Klåvus for his master's thesis in Bioinformatics at Aalto university (published under former name Anton Mattsson), while working for University of Eastern Finland and Afekta Technologies. Notame is inspired by analysis scripts written by Jussi Paananen, Oskari Timonen and Anton Klåvus at University of Eastern Finland. The algorithm for clustering molecular features originating from the same compound is based on MATLAB code written by David Broadhurst. Development was picked up by Atte Lihtamo, Retu Haikonen, Vilhelm Suksi and Leo Lahti, leading to the Bioconductor release. 

If you find any bugs or other things to fix, please submit an issue on GitHub! All contributions to notame are always welcome!

notame is published under an MIT license (tl;dr: it's really permissive!)

# Write SummarizedExperiment objects from data-raw/example_data.R to file
write_to_excel(toy_notame_set, file = "inst/extdata/toy_notame_set.xlsx")
write_to_excel(hilic_neg_sample, file = "inst/extdata/hilic_neg_sample.xlsx")
write_to_excel(hilic_pos_sample, file = "inst/extdata/hilic_pos_sample.xlsx")
write_to_excel(rp_neg_sample, file = "inst/extdata/rp_neg_sample.xlsx")
write_to_excel(rp_pos_sample, file = "inst/extdata/rp_pos_sample.xlsx")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  warning = FALSE,
  message = FALSE,
  results = FALSE,
  comment = "##"
)
# To improve runtime on Windows
BiocParallel::register(BiocParallel::SerialParam())

## ----results = 'markup'-------------------------------------------------------
library(notame)
library(notameViz)
library(notameStats)
library(dplyr)
ppath <- tempdir()
init_log(log_file = file.path(ppath, "log.txt"))

## -----------------------------------------------------------------------------
data <- import_from_excel(
  file = system.file("extdata", "toy_notame_set.xlsx", package = "notame"), 
  sheet = 1, corner_row = 11, corner_column = "H", 
  split_by = c("Column", "Ion_mode"))
  
names(assays(data)) <- "abundances"

## -----------------------------------------------------------------------------
modes <- fix_object(data, split_data = TRUE, assay.type = "abundances")

## ----results = 'markup'-------------------------------------------------------
names(modes)
sapply(modes, class)

## -----------------------------------------------------------------------------
# Initialize empty list for processed objects
processed <- list()
for (i in seq_along(modes)) {
  name <- names(modes)[i]
  mode <- modes[[i]]
  # Set all zero abundances to NA
  mode <- mark_nas(mode, value = 0)
  # Flag features with low detection rate
  mode <- flag_detection(mode, group = "Group")
  # Visualize data before drift correction
  save_QC_plots(mode, prefix = paste0(ppath, "figures/", name, "_ORIG"),
                 perplexity = 5, group = "Group", time = "Time", 
                 id = "Subject_ID", color = "Group")
  # Correct drift
  corrected <- correct_drift(mode)
  # Visualize data after drift correction
  save_QC_plots(corrected, prefix = paste0(ppath, "figures/", name, "_DRIFT"),
                perplexity = 5, group = "Group", time = "Time", 
                id = "Subject_ID", color = "Group")
  # Flag low-quality features
  corrected <- corrected %>% 
    assess_quality() %>% 
    flag_quality()
  # Visualize data after removal of low-quality features
  save_QC_plots(corrected, prefix = paste0(ppath, "figures/", name, "_CLEAN"),
                perplexity = 5, group = "Group", time = "Time",
                id = "Subject_ID", color = "Group")
  # Save result of iteration
  processed[[i]] <- corrected
}

## ----results = 'markup'-------------------------------------------------------
rowData(processed[[1]])$DC_note

## -----------------------------------------------------------------------------
merged <- merge_notame_sets(processed)
save_QC_plots(merged, prefix = paste0(ppath, "figures/_FULL"),
              group = "Group", time = "Time",
              id = "Subject_ID", color = "Group")

## -----------------------------------------------------------------------------
merged_no_qc <- drop_qcs(merged)
save_QC_plots(merged_no_qc, prefix = paste0(ppath, "figures/FULL_NO_QC"),
              group = "Group", time = "Time",
              id = "Subject_ID", color = "Group")

## -----------------------------------------------------------------------------
set.seed(2024)
imputed <- impute_rf(merged_no_qc)
save_QC_plots(imputed, prefix = paste0(ppath, "figures/FULL_IMPUTED"),
              group = "Group", time = "Time",
              id = "Subject_ID", color = "Group")

## -----------------------------------------------------------------------------
base <- impute_rf(imputed, all_features = TRUE)

## -----------------------------------------------------------------------------
save(base, file = paste0(ppath, "full_data.RData"))

## -----------------------------------------------------------------------------
assay(base, "log") <- log(assay(base))
lm_results <- perform_lm(base, assay.type = "log", 
  formula_char = "Feature ~ Group")
base <- join_rowData(base, lm_results)

## -----------------------------------------------------------------------------
write_to_excel(base, file = paste0(ppath, "results.xlsx"))
finish_log()

## ----echo = FALSE, results = 'markup'-----------------------------------------
sessionInfo()

.log_text_if <- function(text, logif) {
  if (logif) {
    log_text(text)
  }
}

.create_sample_col <- function(x, id_prefix, id_column, log_messages) {
  # If id_column is provided, try to change name of the column to "Sample_ID"
  if (!is.null(id_column)) {
    .log_text_if("Checking provided sample ID column", log_messages)
    if (!id_column %in% colnames(x)) {
      .log_text_if(paste0("ID column '", id_column, "' not found"),
                   log_messages)
    } else if (!any(duplicated(x[, id_column])) && !any(is.na(x[, id_column]))){
      x$Sample_ID <- x[, id_column]
      .log_text_if(paste0("Column 'Sample_ID' created from ", id_column),
                   log_messages)
    } else {
      .log_text_if("Provided sample ID column is not valid", log_messages)
    }
  }
  # If Sample_ID is not provided explicitly, it will be created
  if (!"Sample_ID" %in% colnames(x)) {
    x$Sample_ID <- paste0(id_prefix, x$Injection_order)
    .log_text_if(paste("Sample ID generated from injection orders and prefix",
                       id_prefix), 
                 log_messages)
  } else {
    # Fill sample IDs of "QC" identifiers
    x$Sample_ID <- as.character(x$Sample_ID)
    if (any(is.na(x$Sample_ID))) {
      .log_text_if(paste("Missing values found in Sample_ID,",
                         "filling missing sample IDs of QC samples."), 
                   log_messages)
      x$Sample_ID[is.na(x$Sample_ID) & x$QC == "QC"] <- "QC"
    }
    # Fill sample IDs of "Blank" identifiers
    if (any(is.na(x$Sample_ID))) {
      .log_text_if(paste("Missing values found in Sample_ID after filling IDs",
                         "of QCs, trying to detect 'Blank' samples"), 
                  log_messages)
      blank_found <- which(vapply(x, function(y) {
        any(y == "Blank")
      }, logical(1)))
      if (length(blank_found)) {
        x$Sample_ID[is.na(x$Sample_ID) &
        x[[blank_found[1]]] == "Blank"] <- "Blank"
        x$Sample_ID[x$Sample_ID == "Blank"] <- paste0("Blank_",
          seq_len(sum(x$Sample_ID == "Blank")))
      }
      if (any(is.na(x$Sample_ID))) {
        stop("Missing values in Sample_ID after attempting",
             " filling IDs of QCs and Blanks")
      }
    }
    # Add running index to QC identifiers
    if (sum(x$Sample_ID == "QC")) {
      .log_text_if("Adding running index to 'QC' sample IDs", log_messages)
      x$Sample_ID[x$Sample_ID == "QC"] <- paste0("QC_", 
        seq_len(sum(x$Sample_ID == "QC")))
    }
    # Add running index to Blank identifiers
    if (sum(x$Sample_ID == "Blank")) {
      .log_text_if("Adding running index to 'Blank' sample IDs", log_messages)
      x$Sample_ID[x$Sample_ID == "Blank"] <- 
        paste0("Blank_", seq_len(sum(x$Sample_ID == "Blank")))
    }
    # After this, the Sample IDs should be unique
    if (any(duplicated(x$Sample_ID))) {
      stop("Sample_ID is not unique")
    }
  }
  x
}

## Helper function for checking integrity of sample data
.check_pheno_data <- function(x, pheno_injection = FALSE, pheno_ID = FALSE, 
                              pheno_QC = FALSE, pheno_factors = NULL,
                              pheno_nums = NULL, pheno_chars = NULL,   
                              pheno_cols = NULL, log_messages = FALSE) {
  .log_text_if("\nChecking sample information", log_messages)
  
  if (pheno_injection) {
    .log_text_if("Checking 'Injection_order' column in feature data",
                  log_messages)
    # Check that Injection order is included
    if (!"Injection_order" %in% colnames(x)) {
      stop("'Injection_order' not found for the samples.")
    }
    # Check that injection order can be numeric
    if (!is.numeric(x$Injection_order) && !.looks_numeric(x$Injection_order)) {
      stop("'Injection_order' is not numeric and cannot be converted to 
           numeric")
    }
    # No NAs allowed in Injection order
    if (any(is.na(x$Injection_order))) {
      stop("Missing values in Injection_order")
    }
    # Injection order should be unique
    if (length(unique(x$Injection_order)) != nrow(x)) {
      stop("Injection_order is not unique")
    }
  }

  if (pheno_ID) {
    .log_text_if("Checking 'Sample_ID' column in pheno data",
                  log_messages)           
    if (!identical(x$Sample_ID, rownames(x))) {
      stop("Sample_ID does not match rownames in pheno data")
    }
  } 
  
  if (pheno_QC) {
    .log_text_if("Checking 'QC' column in feature data",
                  log_messages)
    if (!"QC" %in% colnames(x)) {
      stop("No 'QC' column found")
    }
    if (any(is.na(x[, "QC"]))) {
      stop("QC column should not contain NAs")
    }
  }

  if (!is.null(pheno_factors)) {
    lapply(pheno_factors, function(pheno_factor) {
      if(!pheno_factor %in% colnames(x)) {
        stop(pheno_factor, " is not a column in pheno data", call. = FALSE)
      }
      if (!is.factor(x[, pheno_factor])) {
        stop(pheno_factor, " column is not a factor", call. = FALSE)
      }
      if (length(levels(x[, pheno_factor])) < 2) {
        stop("Column ", pheno_factor, " should have at least two levels!")
      }
    })
  }

  if (!is.null(pheno_nums)) {
    lapply(pheno_nums, function(pheno_num) {
      if(!pheno_num %in% colnames(x)) {
        stop(pheno_num, " is not a column in pheno data", call. = FALSE)
      }
      if (!is.numeric(x[, pheno_num])) {
        stop(pheno_num, " column is not numeric", call. = FALSE)
      }
    })
  }
  
  if (!is.null(pheno_chars)) {
    lapply(pheno_chars, function(pheno_char) {
      if(!pheno_char %in% colnames(x)) {
        stop(pheno_char, " is not a column in pheno data", call. = FALSE)
      }
      if (!is.character(x[, pheno_char])) {
        stop(pheno_char, " column is not a character", call. = FALSE)
      }
    })
  }
  
  if (!is.null(pheno_cols)) {
    lapply(pheno_cols, function(pheno_col) {
      if(!pheno_col %in% colnames(x)) {
        stop(pheno_col, " is not a column in pheno data", call. = FALSE)
      }
    })
  }
}

## Check that the position of the corner row and column is OK
.check_position <- function(x, cc, cr) {
  condition <- is.na(x[cr - 1, cc - 1]) &
    (is.numeric(utils::type.convert(x[cr + 1, cc + 1], as.is = TRUE)) |
      is.na(x[cr + 1, cc + 1])) &
    !is.na(x[cr, cc]) &
    !is.na(x[cr - 1, cc]) &
    !is.na(x[cr, cc - 1])
  if (!condition) {
    stop("The corner row and column coordinates seem to be incorrect!")
  }
}

# Check if a vector can be safely converted to numeric
.looks_numeric <- function(x) {
  stopifnot(is.atomic(x) || is.list(x))
  tryCatch(
    {
      as.numeric(x)
      TRUE
    },
    warning = function(w) FALSE
  )
}

## Check that all abundances look OK
.check_assay <- function(assay_, log_messages = FALSE) {
  .log_text_if("Checking that feature abundances only contain numeric values",
               log_messages)
  # Check that all rows are full of numbers
  non_numerics <- assay_ |> apply(1, function(x) !.looks_numeric(x))
  if (sum(non_numerics)) {
    stop(paste("Non-numeric values found in the abundances on rows",
               paste(which(non_numerics), collapse = ", ")))
  }
}

.check_feature_data <- function(feature_data, feature_ID = FALSE, 
                                check_limits = FALSE, feature_split = FALSE,
                                feature_flag = FALSE, mz_limits = c(10, 2000), 
                                rt_limits = c(0, 20), feature_cols = NULL, 
                                log_messages = FALSE) {
  .log_text_if("\nChecking feature information", log_messages)          
  if (feature_ID) {
      .log_text_if(paste0("Checking that feature IDs are unique and not stored",
                          "as numbers"),
                   log_messages)     
    fid <- feature_data$Feature_ID
    if (any(duplicated(fid))) {
      stop("Feature_ID values are not unique")
    }
    if (any(is.na(fid))) {
      stop("Missing values in Feature IDs")
    }
    fid_num <- withCallingHandlers(
      expr = as.numeric(fid), 
      warning = function(w) tryInvokeRestart("muffleWarning"))
    if (any(!is.na(fid_num))) {
      stop("Numbers are not allowed as feature IDs.", call. = FALSE)
    }
    fid_chr <- withCallingHandlers(
      expr = as.character(fid),
      warning = function(w) tryInvokeRestart("muffleWarning"))
    if (any(grepl("^[[:digit:]]", fid_chr))) {
      stop("Feature IDs can not start with numbers.", call. = FALSE)
    }
    if (!identical(feature_data$Feature_ID, rownames(feature_data))) {
      stop("Feature_ID does not match rownames in feature data")
    }
  }
  
  if (check_limits) {
    .log_text_if("Checking that m/z and retention time values are reasonable.",
                 log_messages)
    mz <- feature_data[, .find_mz_rt_cols(feature_data)$mz_col]
    rt <- feature_data[, .find_mz_rt_cols(feature_data)$rt_col]
    if (!(all(mz > mz_limits[1]) && all(mz < mz_limits[2])) ||
        !(all(rt > rt_limits[1]) && all(rt < rt_limits[2]))) {
      stop("Values in m/z or retention time columns are outside limits.")
    }
  }
  
  if (feature_split) {
    .log_text_if("Checking that feature data includes a 'Split' column",
                 log_messages)
    if (!"Split" %in% colnames(feature_data)) {
      stop("Split column not found in feature data")
    }
  }
  
  if (feature_flag) {
    .log_text_if("Checking that feature data includes a 'Flag' column",
                 log_messages)
    if (!"Flag" %in% colnames(feature_data)) {
      stop("Flag column not found in feature data")
    }
  }
  
  if (!is.null(feature_cols)) {
    lapply(feature_cols, function(feature_col) {
      if(!feature_col %in% colnames(feature_data)) {
        stop(feature_col, " is not a column in feature data", call. = FALSE)
      }
    })
  }
}

.extract_from_excel <- function(file, sheet, corner_row, corner_column, name) {
  dada <- openxlsx::read.xlsx(file, sheet, colNames = FALSE)
  
  # Define excel column order A-Z, AA - ZZ, AAA - ZZZ
  grid2 <- expand.grid(LETTERS, LETTERS)
  combinations2 <- paste0(grid2$Var2, grid2$Var1)
  grid3 <- expand.grid(LETTERS, combinations2)
  combinations3 <- paste0(grid3$Var2, grid3$Var1)
  excel_columns <- c(LETTERS, combinations2, combinations3)
  
  # If corner coordinates are omitted, try to find them automatically
  if (is.null(corner_row) || is.null(corner_column)) {
    log_text("Detecting corner position")
  }
  corner_row <- corner_row %||% which(!is.na(dada[, 1]))[1]
  corner_column <- corner_column %||% which(!is.na(dada[1, ]))[1]
  # Column can be given as a character
  cc <- ifelse(is.character(corner_column),
    which(excel_columns == corner_column), corner_column)
  cr <- corner_row
  # Check that corner is in the right spot
  .check_position(dada, cc, cr)
  log_text(paste0("Corner detected correctly at row ", cr, ", column ",
                  excel_columns[cc]))
  # Extract sample information
  log_text(paste0("\nExtracting sample information from rows 1 to ", cr,
                  " and columns ", excel_columns[cc + 1], " to ",
                  excel_columns[ncol(dada)]))
  pheno_data <- as.data.frame(t(dada[seq_len(cr), (cc + 1):ncol(dada)]),
                              stringsAsFactors = FALSE)
  log_text(paste0("Replacing spaces in sample information column names with", 
           " underscores (_)"))
  colnames(pheno_data) <- gsub(" ", "_",
                               c(dada[seq_len(cr - 1), cc], "Datafile"))
  
  # If a single mode is given, datafile will indicate the mode
  if (!is.null(name)) {
    colnames(pheno_data)[ncol(pheno_data)] <- paste0(name, "_Datafile")
    log_text(paste0('Naming the last column of sample information "', name,
                    '_Datafile"'))
  } else {
    log_text('Naming the last column of sample information "Datafile"')
  }
  
  # Extract feature information
  log_text(paste0("\nExtracting feature information from rows ", cr + 1, 
                  " to ", nrow(dada), " and columns ", excel_columns[1], 
                  " to ", excel_columns[cc]))
  feature_data <- dada[(cr + 1):nrow(dada), seq_len(cc)]
  colnames(feature_data) <- dada[cr, seq_len(cc)]
  # Extract LC-MS measurements as matrix
  log_text(paste0("\nExtracting feature abundances from rows ", cr + 1, " to ",
                  nrow(dada), " and columns ", excel_columns[cc + 1], " to ",
                  excel_columns[ncol(dada)]))
  assay_ <- dada[(cr + 1):nrow(dada), (cc + 1):ncol(dada)]

  list("pheno_data" = pheno_data, "feature_data" = feature_data, 
       "assay_" = assay_ )
}

.check_integrity <- function(
    extracted, id_prefix, id_column,
    split_by, name, mz_limits, rt_limits
) {
  pheno_data <- extracted$pheno_data
  feature_data <- extracted$feature_data
  assay_ <- extracted$assay_
  pheno_data <- .fix_pheno_data(x = pheno_data, id_prefix = id_prefix,
                                id_column = id_column, clean = TRUE,
                                log_messages = TRUE)
  assay_ <- .fix_assay(assay_, log_messages = TRUE)
  feature_data <- .fix_feature_data(feature_data, name = name, 
                                    split_by = split_by, clean = TRUE, 
                                    log_messages = TRUE)

  .check_pheno_data(x = pheno_data, pheno_injection = TRUE, pheno_ID = TRUE,
                    pheno_QC = TRUE, log_messages = TRUE)
  .check_assay(assay_, log_messages = TRUE)
  .check_feature_data(feature_data, feature_ID = TRUE, check_limits = TRUE, 
                      feature_split = TRUE, mz_limits = mz_limits, 
                      rt_limits = rt_limits, feature_flag = TRUE,
                      log_messages = TRUE)
  
  rownames(assay_) <- rownames(feature_data)
  colnames(assay_) <- rownames(pheno_data)
  
  list(
    "pheno_data" = pheno_data,
    "feature_data" = feature_data,
    "assay_" = assay_
  )
}

# Helper function to validate import arguments
.check_import_args <- function(skip_checks, split_by, name) {
  if (!skip_checks && is.null(split_by) && is.null(name)) {
    stop("Either name or split_by needs to be defined, see documentation.")
  } else if (!skip_checks && !is.null(split_by) && !is.null(name)) {
    stop("Only define split_by OR name, see documentation.")
  }
}

  
#' Read formatted Excel files
#'
#' Reads data from an Excel file of the following format:
#' \itemize{
#'   \item Left side of the sheet contains information about the features, size 
#' features x feature info columns
#'   \item Top part contains sample information, size sample info variables x 
#' samples
#'   \item The middle contains the actual abundances, size features x samples
#' }
#' Exported data from MS-DIAL converted to an excel file creates such format.
#'
#' @param file path to the Excel file
#' @param sheet the sheet number or name
#' @param id_column character, column name for unique identification of samples
#' @param corner_row integer, the bottom row of sample information,
#' usually contains data file names and feature info column names.
#' If set to NULL, will be detected automatically.
#' @param corner_column integer or character, the corresponding column number 
#' or the column name (letter) in Excel.
#' If set to NULL, will be detected automatically.
#' @param id_prefix character, prefix for autogenerated sample IDs
#' @param split_by character vector, in the case where all the modes are in the 
#' same Excel file, the column names of feature data used to separate the modes 
#' (usually Mode and Column)
#' @param name in the case where the Excel file only contains one mode, the 
#' name of the mode, such as "Hilic_neg"
#' @param mz_limits numeric vector of two, all m/z values should be in between 
#' these
#' @param rt_limits numeric vector of two, all retention time values should be 
#' in between these
#' @param skip_checks logical: skip checking and fixing of data integrity. Not 
#' recommended, but sometimes useful when you just want to read the data in as 
#' is and fix errors later. The data integrity checks are important for 
#' functioning of notame.
#'
#' @return A SummarizedExperiment with assays, rowData and colData filled from
#' the Excel file.
#'
#' @examples
#' data <- import_from_excel(
#'   file = system.file("extdata", "toy_notame_set.xlsx", 
#'   package = "notame"), sheet = 1, corner_row = 11, corner_column = "H",
#'   split_by = c("Column", "Ion_mode"))
#'
#' @export
import_from_excel <- function(file, sheet = 1, id_column = NULL, 
                          corner_row = NULL, corner_column = NULL,
                          id_prefix = "ID_", split_by = NULL, name = NULL,
                          mz_limits = c(10, 2000), rt_limits = c(0, 20),
                          skip_checks = FALSE) {
  .check_import_args(skip_checks, split_by, name)
  extracted <- .extract_from_excel(file, sheet, corner_row, 
                                    corner_column, name)
  # Skip checks
  if (!skip_checks) {
    extracted <- .check_integrity(extracted, id_prefix = id_prefix,
                     id_column = id_column, split_by = split_by, name = name,
                     mz_limits = mz_limits, rt_limits = rt_limits)
  }

  SummarizedExperiment(
    assays = as.matrix(extracted$assay_),
    rowData = extracted$feature_data,
    colData = extracted$pheno_data
  )
}

.extract_from_tsv <- function(file, name) {
  # Read the raw file
  raw <- utils::read.delim(
    file = file,
    header = FALSE,
    sep = "\t",
    stringsAsFactors = FALSE,
    fill = TRUE,
    quote = NULL,
    na.strings = c("NA", "", "null")
  )

  # first row tells us which columns hold samples
  sample_cols <- which(raw[1, ] != "" & !is.na(raw[1, ]))
  # first column tells us which rows hold features
  feature_rows <- which(raw[, 1] != "" & !is.na(raw[, 1]))
  # columns and rows overlap by one between feature/sample data
  feature_cols <- seq_len((min(sample_cols)))
  sample_rows <- seq_len((min(feature_rows)))

  # Gather feature data
  feature_data <- raw[feature_rows[-1], feature_cols, drop = FALSE]
  colnames(feature_data) <- raw[feature_rows[1], feature_cols]

  # Gather sample data
  sample_data <- as.data.frame(
    t(raw[sample_rows, sample_cols[-1]]),
    stringsAsFactors = FALSE
  )
  log_text(paste0("Replacing spaces in sample information column names with", 
           " underscores (_)"))
  colnames(sample_data) <- gsub(" ", "_", raw[sample_rows, sample_cols[1]])
  # If a single mode is given, datafile will indicate the mode
  if (!is.null(name)) {
    log_text(paste0('Naming the last column of sample information "', name,
                    '_Datafile"'))
    datafile_col <- paste0(name, "_Datafile")
  } else {
    log_text('Naming the last column of sample information "Datafile"')
    datafile_col <- "Datafile"
  }
  colnames(sample_data)[length(colnames(sample_data))] <- datafile_col
  rownames(sample_data) <- sample_data[, datafile_col]

  # Gather abundance values from intersection
  values <- raw[feature_rows[-1], sample_cols[-1]]
  values <- as.matrix(as.data.frame(lapply(values, as.numeric)))
  colnames(values) <- raw[feature_rows[1], sample_cols[-1]]

  list("pheno_data" = sample_data, "feature_data" = feature_data, 
       "assay_" = values )
}

#' Import data from a TSV file created by MS-DIAL
#'
#' Reads a tab-delimited peak table (TSV) exported from MS-DIAL. This works 
#' similarly to \code{\link{import_from_excel}}, but removes the need to 
#' manually open and re-save the TSV files in Excel. Excludes any summary
#' statistics computed by MS-DIAL, such as average abundances, since these are
#' not valid after notame processing.
#'
#' @param file path to the TSV file
#' @param id_column character, column name for unique identification of samples
#' @param id_prefix character, prefix for autogenerated sample IDs
#' @param split_by character vector, in the case where all the modes are in the 
#' same Excel file, the column names of feature data used to separate the modes 
#' (usually Mode and Column)
#' @param name in the case where the Excel file only contains one mode, the 
#' name of the mode, such as "Hilic_neg"
#' @param mz_limits numeric vector of two, all m/z values should be in between 
#' these
#' @param rt_limits numeric vector of two, all retention time values should be 
#' in between these
#' @param skip_checks logical: skip checking and fixing of data integrity. Not 
#' recommended, but sometimes useful when you just want to read the data in as 
#' is and fix errors later. The data integrity checks are important for 
#' functioning of notame.
#'
#' @return A SummarizedExperiment with assays, rowData and colData filled from
#' the TSV file.
#'
#' @examples
#' \donttest{
#' # Read a TSV peak table
#' # data <- import_from_msdial(file = "path/to/peak_table.tsv", name = "RP_neg")
#' }
#'
#' @export
import_from_msdial <- function(file, id_column = NULL,
                          id_prefix = "ID_", split_by = NULL, name = NULL,
                          mz_limits = c(10, 2000), rt_limits = c(0, 20),
                          skip_checks = FALSE) {
  .check_import_args(skip_checks, split_by, name)
  extracted <- .extract_from_tsv(file, name)
  # Skip checks
  if (!skip_checks) {
    extracted <- .check_integrity(extracted, id_prefix = id_prefix,
                     id_column = id_column, split_by = split_by, name = name,
                     mz_limits = mz_limits, rt_limits = rt_limits)
  }

  SummarizedExperiment(
    assays = as.matrix(extracted$assay_),
    rowData = extracted$feature_data,
    colData = extracted$pheno_data
  )
}

# Helper function to search for mass and retention time column names
.find_mz_rt_cols <- function(feature_data, log_messages = TRUE) {
  # Find mass and retention time columns
  mz_tags <- c("mass", "m.?z$", "molecular.?weight")
  rt_tags <- c("retention.?time", "^rt$", "(?=.*rt)(?=.*min)")
  mz_col <- NULL
  for (tag in mz_tags) {
    hits <- grepl(tag, tolower(colnames(feature_data)), perl = TRUE)
    if (any(hits)) {
      mz_col <- colnames(feature_data)[which(hits)[1]]
      break
    }
  }
  rt_col <- NULL
  for (tag in rt_tags) {
    hits <- grepl(tag, tolower(colnames(feature_data)), perl = TRUE)
    if (any(hits)) {
      rt_col <- colnames(feature_data)[which(hits)[1]]
      break
    }
  }
  # If not found, throw error
  if (is.null(mz_col)) {
    stop(paste0("No mass to charge ratio column found - should match one of:\n",
                paste(mz_tags, collapse = ", "), " (not case-sensitive)"))
  }
  if (is.null(rt_col)) {
    stop(paste0("No retention time column found - should match one of:\n",
                paste(rt_tags, collapse = ", "), " (not case-sensitive)"))
  }
  .log_text_if(paste0("Identified m/z column ", mz_col,
                      " and retention time column ", rt_col),
               log_messages)
                  
  return(list(mz_col = mz_col, rt_col = rt_col))
}

# Combines mode name, mass and retention time to create a Feature ID
.name_features <- function(feature_data) {
  cols <- .find_mz_rt_cols(feature_data)
  mz_col <- cols$mz_col
  rt_col <- cols$rt_col
  log_text(paste0("Identified m/z column ", mz_col, 
                  " and retention time column ", rt_col))
  log_text("Creating feature IDs from Split, m/z and retention time")

  round_mz <- as.numeric(feature_data[, mz_col]) |>
    as.character() |>
    gsub("[.]", "_", x = _)
  round_rt <- as.numeric(feature_data[, rt_col]) |>
    as.character() |>
    gsub("[.]", "_", x = _)
  feature_data$Feature_ID <- paste0(feature_data$Split, "_", 
                                    round_mz, "a", round_rt)
  if (anyDuplicated(feature_data$Feature_ID)) {
    duplicates <- paste0(
      feature_data$Feature_ID[duplicated(feature_data$Feature_ID)],
      collapse = ", ")
    stop("Could not create unique feature names from m/z and", 
         " retention time columns. Duplicated values: ", duplicates)
  }

  feature_data
}

#' Write results to Excel file
#'
#' Writes all the data in a SummarizedExperiment object to an 
#' Excel spreadsheet.
#' The format is similar to the one used to read data in, except for the fact 
#' that EVERYTHING NEEDS TO BE WRITTEN AS TEXT. To fix numeric values in Excel,
#' choose any cell with a number, press Ctrl + A, then go to the dropdown menu
#' in upper left corner and choose "Convert to Number". This will fix the file,
#' but can take quite a while.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param file path to the file to write
#' @param ... Additional parameters passed to
#' \code{\link[openxlsx]{write.xlsx}}
#'
#' @return None, the function is invoked for its side effect.
#'
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' data(toy_notame_set)
#' write_to_excel(toy_notame_set, file = "toy_notame_set.xlsx")
#' \dontshow{setwd(.old_wd)}
#'
#' @export
write_to_excel <- function(object, file, ...) {
  object <- .check_object(object)
  # Bottom part consists of (from left to right):
  # - feature data with results
  # - abundance values
  bottom <- as.data.frame(cbind(rowData(object), assay(object)))
  # All columns must be characters to allow combination with the top block
  bottom <- bottom |>
    dplyr::mutate(dplyr::across(dplyr::everything(), as.character))
  bottom <- rbind(colnames(bottom), bottom)
  # Top block holds the sample information
  pd <- as.data.frame(colData(object))
  datafile_cols <- colnames(pd)[grepl("Datafile", colnames(pd))]
  if (length(datafile_cols)) {
    last_datafile <- datafile_cols[length(datafile_cols)]
    pd <- pd[, c(setdiff(colnames(pd), last_datafile), last_datafile)]
    log_text(paste0("Moved ", last_datafile, " column to last",
                    " to get meaningful column names for abundances"))
  }
  top <- cbind(matrix(colnames(pd), ncol = 1), t(pd))
  # NA blocks to fill the empty space
  empty1 <- matrix(NA_character_,
                   nrow = nrow(top),
                   ncol = ncol(rowData(object)) - 1)
  top <- cbind(empty1, top)
  colnames(top) <- colnames(bottom)
  # Replace assay column names with the last column (now row) of sample info
  replace_idx <- (ncol(rowData(object)) + 1):ncol(bottom)
  bottom[1, replace_idx] <- top[nrow(top), replace_idx]
  # All combined
  big <- rbind(top[seq_len(nrow(top) - 1), ], bottom)

  openxlsx::write.xlsx(big, file = file, colNames = FALSE, ...)
}


# ------------ Methods -----------------
#' Retrieve both sample information and features
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param assay.type character, assay to be used in case of multiple assays
#' @return A data frame with sample information plus all features as columns,
#' one row per sample.
#'
#' @examples
#' data(toy_notame_set)
#' combined_data(toy_notame_set)
#'
#' @export
combined_data <- function(object, assay.type = NULL) {
  from <- .get_from_name(object, assay.type)
  cbind(as.data.frame(colData(object)), t(assay(object, from)))
}


#' Get and set the values in the flag column
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' 
#' @return Character vector of feature flags.
#'
#' @examples
#' data(toy_notame_set)
#' # Get values in flag column of rowData
#' flag(toy_notame_set)
#'
#' @export
flag <- function(object) rowData(object)$Flag

#' @rdname flag
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param value character vector, values for flag column
#'
#' @return For the endomorphism, an object with a modified flag column.
#'
#' @examples
#' data(toy_notame_set)
#' # Flag a suspicious feature manually
#' flag(toy_notame_set)[1] <- "Contaminant, known from experience"
#' @export
"flag<-" <-  function(object, value) {
  rowData(object)$Flag <- value
  if (validObject(object)) {
    return(object)
  }
}

# Wrapper helper function for checking compatibility of SummarizedExperiment
# object  with notame. The parameters with boolean arguments check for set 
# names of columns such as "Feature_ID". Arguments with default = NULL are used 
# for checking the existence and/or class of columns in pheno or feature
# data.
.check_object <- function(object, pheno_injection = FALSE, pheno_ID = FALSE, 
                          pheno_QC = FALSE, pheno_factors = NULL,
                          pheno_nums = NULL, pheno_chars = NULL,
                          pheno_cols = NULL, assay.type = NULL,
                          feature_ID = FALSE, check_limits = FALSE, 
                          feature_split = FALSE, feature_flag = FALSE,
                          feature_cols = NULL, mz_limits = c(10, 2000), 
                          rt_limits = c(0, 20)) {
                      
  .check_pheno_data(colData(object), pheno_injection = pheno_injection, 
                    pheno_ID = pheno_ID, pheno_QC = pheno_QC,
                    pheno_factors = pheno_factors, pheno_nums = pheno_nums, 
                    pheno_chars = pheno_chars, pheno_cols = pheno_cols)
  
  if (!is.null(assay.type)) {
  .check_assay(assay(object, assay.type))
  }
    
  .check_feature_data(rowData(object), feature_ID = feature_ID, 
                      check_limits = check_limits, 
                      feature_split = feature_split, 
                      feature_flag = feature_flag, mz_limits = mz_limits, 
                      rt_limits = rt_limits, feature_cols = feature_cols)

  object
}

#' Fix object for functioning of notame
#' 
#' Attempts to create missing columns needed for notame in pheno and feature 
#' data. Optionally cleans the object and splits the object by mode.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param id_prefix character, prefix for autogenerated sample IDs, see Details
#' @param id_column character, column name for unique identification of samples
#' @param split_by character vector, in the case where all the modes are in the 
#' same object, the column names of feature data used to separate the modes 
#' (usually Mode and Column)
#' @param name in the case where object only contains one mode, the 
#' name of the mode, such as "Hilic_neg"
#' @param clean boolean, whether to select best classes, reorder columns and 
#' consistently rename columns in pheno and feature
#' @param split_data logical, whether to split data by analytical mode recorded 
#' in the "Split" column of feature data. If TRUE (the default), will return a 
#' list of objects, one per analytical mode. If FALSE, will return a 
#' single object.
#' @param assay.type character, assay to be used in case of multiple assays
#'
#' @return A new SummarizedExperiment object with a single peak table. If 
#' split_data = TRUE, a list containing separate objects for analytical modes.
#'
#' @details Only specify one of \code{split_by} and \code{name}. The feature 
#' data will contain columns named "Split", used to separate features from 
#' different modes, and "Flag" for recording flagged features. Unless a column 
#' named "Feature_ID" is found in feature data, a feature ID will be generated 
#' based on the value of "Split", mass and retention time. The function will 
#' try to find columns for mass and retention time by looking at a few common 
#' alternatives, and throw 
#' an error if no matching column is found. Sample information needs to contain 
#' a row called "Injection_order", and the values need to be unique. In 
#' addition, a possible sample identifier row needs to be named "Sample_ID",
#' or to be specified in \code{id_column}, and the values need to be unique, 
#' with an exception of QC samples: if there are any "QC" identifiers, they 
#' will be replaced with "QC_1", "QC_2" and so on.
#' If a "Sample_ID" column is not found, it will be created using the 
#' \code{id_prefix} and injection order or by renaming \code{id_column}.
#'
#' @examples
#' data(toy_notame_set)
#' ex_set <- toy_notame_set
#' rowData(ex_set)$Flag <- NULL
# 'Flag' column is created in feature data
#' fixed <- fix_object(ex_set)
#' 
#' @export
fix_object <- function(object, id_prefix = "ID_", id_column = NULL,
                       split_by = NULL, name = NULL, clean = TRUE,
                       split_data = FALSE, assay.type = NULL) {

  from <- .get_from_name(object, assay.type)
  object_orig <- object
  pheno_data <- .fix_pheno_data(colData(object), id_prefix = "",
                                id_column = id_column, clean = clean,  
                                log_messages = TRUE)
  
  feature_data <- .fix_feature_data(rowData(object), split_by = split_by,
                                    name = name, clean = clean,
                                    log_messages = TRUE)

  assay <- .fix_assay(assay(object, from))
                                         
  rownames(assay) <- rownames(feature_data)
  colnames(assay) <- rownames(pheno_data)
                
  if (split_data) {
    # Split the data by the Split column of feature data
    parts <- unique(feature_data$Split)
    obj_list <- list()
    for (part in parts) {
      fd_tmp <- S4Vectors::DataFrame(feature_data[feature_data$Split == part, ])
      ad_tmp <- assay[fd_tmp$Feature_ID, ]
      obj_list[[part]] <- SummarizedExperiment(assays = ad_tmp,
                                               colData = pheno_data,
                                               rowData = fd_tmp)
      names(assays(obj_list[[part]])) <- from
    }
    return(obj_list)
  } else {
    object <- SummarizedExperiment(assays = assay,
                                   colData = pheno_data,
                                   rowData = feature_data)
    names(assays(object)) <- from
    object
  }
}

#' @noRd
.fix_pheno_data <- function(x, id_prefix = "", id_column = NULL, 
                            log_messages = FALSE, clean = TRUE) {
  # If QC column is not provided explicitly, attempt to create it
  if (!"QC" %in% colnames(x)) {
    qc_found <- apply(x, 1, function(y) {
      any(grepl("QC", y))
    })
    if (any(qc_found)) {
      x$QC <- ifelse(qc_found, "QC", "Sample")
      log_text(paste("QC column generated from rows containing 'QC'"))
    } else {
      warning("QC not found and column can not be generated.", 
              " Please construct one.")
    }
  }
  # Create and populate 'Sample_ID' column
  x <- .create_sample_col(x, id_prefix, id_column, log_messages = TRUE)

  if (clean) {
    pre_clean <- x
    # Select best classes for columns and prepare data.frame
    x <- .best_classes(x)
    x <- as.data.frame(dplyr::select(x, "Sample_ID", dplyr::everything()))
    
    if (!isTRUE(all.equal(as.data.frame(pre_clean), x))) {
      log_text("Pheno data was cleaned")
    }
  }
  rownames(x) <- x$Sample_ID 
  x
}

.fix_assay <- function(assay_, log_messages = FALSE) {
  # Check that all rows are full of numbers
  non_numerics <- assay_ |> apply(1, function(x) !.looks_numeric(x))
  if (sum(non_numerics)) {
    stop(paste("Non-numeric values found in the abundances on rows",
               paste(which(non_numerics), collapse = ", ")))
  }
  # Convert to numeric
  assay_ <- assay_ |> apply(2, as.numeric)

  assay_
}

.fix_feature_data <- function(name = NULL, split_by = NULL, feature_data, 
                              clean = TRUE, log_messages = FALSE) {
  if (!"Flag" %in% colnames(feature_data)) {
    log_text("Initializing 'Flag' column with unflagged features")
    feature_data$Flag <- NA
  }

  if (!"Split" %in% colnames(feature_data)) {
    # If the file only contains one mode, add the mode name as Split column
    if (!is.null(name)) {
      log_text(paste0("Assigning ", name,
                      " as the value of the Split column for each feature"))
      feature_data$Split <- name
      split_by <- "Split"
    } else { # Multiple modes in the file, create Split column to separate modes
      if (!all(split_by %in% colnames(feature_data))) {
        stop(paste0("Couldn't find column(s): ",
                    paste(split_by[!(split_by %in% colnames(feature_data))],
                          collapse = ", ")))
      }
      log_text(paste0("Creating Split column from ",
                      paste0(split_by, collapse = ", ")))
      feature_data <- feature_data |>
        tidyr::unite("Split", split_by, remove = FALSE)
    }
  }

  # Create Feature ID if necessary
  if (!"Feature_ID" %in% colnames(feature_data)) {
    log_text("Feature_ID column not found, creating feature IDs")
    feature_data <- .name_features(feature_data = feature_data)
  }
    
  if (clean) {
    # Reorganise columns and change classes
    pre_clean <- feature_data
    feature_data <- as.data.frame(feature_data) |>
      dplyr::select("Feature_ID", "Split", dplyr::everything()) |>
      .best_classes()
    # Replace dots with underscores in colnames
    colnames(feature_data) <- gsub("[.]", "_", colnames(feature_data)) |>
    # Remove duplicate underscores
      gsub("_{2,}", "_", x = _)
    if (!isTRUE(all.equal(as.data.frame(pre_clean), feature_data))) {
      log_text("Feature data was cleaned")
    }
  }
  rownames(feature_data) <- feature_data$Feature_ID
  feature_data
}

#' Join new columns to feature data
#'
#' Join a new data frame of information to feature data of a 
#' SummarizedExperiment object. The data frame needs to have a column 
#' "Feature_ID". This function is usually used internally by some of the 
#' functions in the package, but can be useful.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' @param dframe a data frame with the new information
#'
#' @examples
#' data(toy_notame_set)
#' new_info <- data.frame(
#'   Feature_ID = rownames(toy_notame_set),
#'   Feature_number = seq_len(nrow(toy_notame_set))
#' )
#' with_new_info <- join_rowData(toy_notame_set, new_info)
#' colnames(rowData(with_new_info))
#'
#' @return A SummarizedExperiment object with the new information added to 
#' rowData(object).
#'
#' @export
join_rowData <- function(object, dframe) {
  with_new <- dplyr::left_join(
    as.data.frame(rowData(object)),
    dframe,
    by = "Feature_ID"
  )
  rowData(object) <- S4Vectors::DataFrame(with_new)
  rownames(object) <- rowData(object)$Feature_ID
  if (validObject(object)) {
    return(object)
  }
}
  
#' Join new columns to pheno data
#'
#' Join a new data frame of information to pheno data of a SummarizedExperiment 
#' object.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param dframe a data frame with the new information
#'
#' @examples
#' data(toy_notame_set)
#' new_info <- data.frame(
#'   Sample_ID = colnames(toy_notame_set),
#'   BMI = stats::runif(ncol(toy_notame_set), 22, 26)
#' )
#' with_new_info <- join_colData(toy_notame_set, new_info)
#' colnames(colData(with_new_info))
#'
#' @return A SummarizedExperiment object with the new information added to 
#' colData(object).
#'
#' @export
join_colData <- function(object, dframe) {
  with_new <- dplyr::left_join(
    as.data.frame(colData(object)),
    dframe,
    by = "Sample_ID"
  )
  colData(object) <- S4Vectors::DataFrame(with_new)
  rownames(colData(object)) <- colData(object)$Sample_ID
  object
}

context("Testing reading data")

data(toy_notame_set, package = "notame")

test_that("Column conversion works", {
  set.seed(38)
  df <- data.frame(
    Injection = 1:100,
    Group = letters[1:4],
    Time = 1:2,
    Sample_ID = sample(letters, size = 100, replace = TRUE),
    few_numbers = c(1.3, 2.5)
  )
  converted <- .best_classes(df)
  converted_classes <- unname(sapply(converted, class))
  expected_classes <- c("integer", "factor", "factor", "character", "numeric")

  expect_equal(converted_classes, expected_classes)
})

test_that("Pheno data fixing works", {
  # No action is taken if "Sample_ID is" supplied (unless there are na's in 
  # only this specific case to not break functionality), safe default behavior # for all columns
  df <- data.frame(Sample_ID = 1:10, QC = rep("QC", 10))
  expect_no_error(.fix_pheno_data(df))
  # User can specify which column should be used for "Sample_ID"
  new_id <- paste0("ID", df$Sample_ID)
  df$id <- new_id
  df <- .fix_pheno_data(df, id_column = "id")
  expect_equal(df$Sample_ID, new_id)

  # Check if conversion to numeric works
  df <- data.frame(Injection_order = c(1:2, "3", 4:9))
  expect_warning(
    expect_visible(.fix_pheno_data(df, id_prefix = "ID_")),
    "QC not found"
  )
  #Check QC generator
  df <- data.frame(
    Injection_order = seq_len(10),
    Sample_ID = c(letters[1:5], letters[1:5])
    )
  expect_warning(
    expect_error(.fix_pheno_data(df), "Sample_ID is not unique"),
    "QC not found"
    )
  df <- data.frame(
    Injection_order = seq_len(5),
    Sample_ID = c(letters[1:5])
    )
  expect_warning(.fix_pheno_data(df), "QC not found")

  df <- data.frame(
    Injection_order = seq_len(10),
    Sample_ID = c(letters[1:5], rep("QC", 5)),
    QC = as.factor(rep(c("Sample", "QC"), each = 5))
  )
  checked <- .fix_pheno_data(df)
  expected <- data.frame(
    Sample_ID = c(letters[1:5], paste0("QC_", 1:5)),
    Injection_order = seq_len(10),
    stringsAsFactors = FALSE,
    QC = as.factor(rep(c("Sample", "QC"), each = 5))
  )
  rownames(expected) <- expected$Sample_ID
  expect_equal(checked, expected)
})

test_that("Pheno data checking works", {
  df <- data.frame(Group = letters[1:2])
  expect_error(.check_pheno_data(df, pheno_injection = TRUE), "'Injection_order' not found")

  df <- data.frame(Injection_order = c(1:5, 3:9))
  expect_error(.check_pheno_data(df, pheno_injection = TRUE), "Injection_order is not unique")
  # Check that error is thrown if injection order is not numeric
  df <- data.frame(Injection_order = c(1:2, "a", 4:9))
  expect_error(.check_pheno_data(df, pheno_injection = TRUE), 
               "'Injection_order' is not numeric")
})

test_that("Feature data fixing works", {
  # No action is taken for "Feature_ID" if supplied, safe default behavior for 
  # all columns
  df <- data.frame(Feature_ID = 1:10, mz = "nope")
  mode <- "HILIC_pos"
  expect_no_error(.fix_feature_data(feature_data = df, name = mode))
  
  df <- data.frame(Split = rep("HILIC_pos", 10))
  expect_error(.fix_feature_data(feature_data = df, name = mode),
               "No mass to charge ratio column found - should match ...")
  mz <- as.numeric(seq_len(nrow(df)))
  df$Average_Mz <- mz
  expect_error(.fix_feature_data(feature_data = df, name = mode),
               "No retention time column found - should match ...")
  rt <- seq_len(nrow(df))   
  df$Average_Rt_min <- rt
  df <- .fix_feature_data(feature_data = df, name = mode)
  
  expect_equal(df$Feature_ID, paste0(mode, "_", mz, "a", rt))
})

test_that("Feature data checking works", {
  df <- data.frame(Feature_ID = 1:5)
  expect_error(.check_feature_data(df, feature_ID = TRUE),
               "Numbers are not allowed as feature IDs")

  df <- data.frame(Feature_ID = c(letters[1:5], letters[3:5]))
  expect_error(.check_feature_data(df, feature_ID = TRUE),
               "Feature_ID values are not unique")

  df <- data.frame(Feature_ID = letters[1:9])
  df$Feature_ID[6] <- NA
  expect_error(.check_feature_data(df, feature_ID = TRUE),
               "Missing values in Feature IDs")
})

test_that("Easy example data is read correctly", {
  # Pheno data
  pd <- data.frame(
    Sample_ID = paste0("TEST_", seq_len(12)),
    Injection_order = seq_len(12),
    Group = as.factor(rep(LETTERS[1:2], times = c(5, 7))),
    QC = as.factor("Sample"),
    easy_Datafile = paste0("190102SR_RP_pos_0", 10:21),
    stringsAsFactors = FALSE
  )
  rownames(pd) <- pd$Sample_ID

  # Feature data
  fd <- data.frame(
    Feature_ID = "",
    Split = as.factor("easy"),
    Alignment = as.numeric(seq_len(10)),
    Mass = 50 * seq_len(10),
    RetentionTime = 0.5 * seq_len(10),
    MS_MS_Spectrum = as.factor(c("(123.45; 678)", rep(NA, 9))),
    Flag = NA,
    stringsAsFactors = FALSE
  )
  fd <- .name_features(fd)
  rownames(fd) <- fd$Feature_ID

  # Assay data
  ad <- matrix(0, nrow = 10, ncol = 12)
  for (i in seq_len(10)) {
    ad[i, ] <- 50000 - (i - 1) * 3000 + 1000 * (0:11)
  }
  dimnames(ad) <- list(rownames(fd), rownames(pd))

  # Read the file
  read <- import_from_excel(system.file("testdata", "easy_data.xlsx", package = "notame"),
    sheet = 1,
    corner_row = 4, corner_column = "D",
    name = "easy", id_prefix = "TEST_"
  )

  # Test that the parts are read as expected
  expect_equal(assay(read), ad)
  expect_equal(as.data.frame(colData(read)), pd)
  expect_equal(as.data.frame(rowData(read)), fd)
})

test_that("Data is split correctly", {
  # Pheno data
  pd <- data.frame(
    Sample_ID = paste0("TEST_", seq_len(12)),
    Injection_order = seq_len(12),
    Group = factor(
      rep(LETTERS[1:2], times = c(5, 7)),
      levels = c("A", "B", "QC")
    ),
    QC = "Sample",
    Datafile = paste0("190102SR_RP_pos_0", 10:21),
    stringsAsFactors = FALSE
  )
  rownames(pd) <- pd$Sample_ID
  qc_idx <- c(1, 6, 9, 12)
  pd$Group[qc_idx] <- "QC"
  pd$QC[qc_idx] <- "QC"
  pd$QC <- factor(pd$QC)

  # Feature data
  fd <- data.frame(
    Feature_ID = "",
    Split = as.factor(rep(c("RP_pos", "Hilic_pos", "Hilic_neg", "RP_neg"), each = 4)),
    Alignment = as.numeric(seq_len(16)),
    Mass = 50 * seq_len(16),
    RetentionTime = 0.5 * seq_len(16),
    Column = as.factor(rep(c("RP", "Hilic", "RP"), times = c(4, 8, 4))),
    Mode = as.factor(rep(c("pos", "neg"), each = 8)),
    MS_MS_Spectrum = as.factor(c("(123.45; 678)", rep(NA, 15))),
    Flag = NA,
    stringsAsFactors = FALSE
  )
  fd <- .name_features(fd)
  rownames(fd) <- fd$Feature_ID

  # Assay data
  ad <- matrix(0, nrow = 16, ncol = 12)
  for (i in seq_len(16)) {
    ad[i, ] <- 50000 - (i - 1) * 3000 + 1000 * (0:11)
  }
  row_zeros <- c(1, 3, 3, 4, 5, 8, 8, 10, rep(11, 5))
  col_zeros <- c(3, 7, 12, 1, 5, 10, 12, 1, 1, 3, 5, 8, 12)
  for (j in seq_along(row_zeros)) {
    ad[row_zeros[j], col_zeros[j]] <- 0
  }
  dimnames(ad) <- list(rownames(fd), rownames(pd))


  # Read the file
  read <- import_from_excel(system.file("testdata", "split_data.xlsx", package = "notame"),
    sheet = 1,
    corner_row = 4, corner_column = "F",
    split_by = c("Column", "Mode"), id_prefix = "TEST_"
  )

  # Test that the parts are read as expected
  expect_equal(assay(read), ad)
  expect_equal(as.data.frame(colData(read)), pd)
  expect_equal(as.data.frame(rowData(read)), fd)
})

test_that("Splitting data works as expected", {
  split_by <- c("Ion mode", "gswregh") # Wrong column name
  expect_error(import_from_excel(
    system.file("testdata", "sample_data_whole.xlsx",
      package = "notame"
    ),
    corner_row = 4,
    corner_column = "X",
    split_by = split_by
  ))
  
  # Check that colData is intact
  se <- toy_notame_set
  se <- fix_object(se)  
  se_modes <- fix_object(se, split_data = TRUE)
  expect_true(all(unlist(lapply(se_modes, function(se_mode) {
    identical(colData(se_mode), colData(se))
  }))))

  # Splitting into modes shouldn't work without "Split" column
  rowData(se)$Split <- NULL
  expect_error(fix_object(ms, split_data = TRUE))
})

test_that("Creating dummy injection order works as expected", {
  names <- list("hilic_neg", "hilic_pos", "rp_neg", "rp_pos")
  modes <- list()
  for (name in names) {
    file <- system.file("extdata", paste0(name, "_sample.xlsx"), package = "notame")
    mode <- import_from_excel(file, name = name)
    modes[name] <- SummarizedExperiment(assays = SimpleList(assay(mode)), 
                                        colData = colData(mode),
                                        rowData = rowData(mode))
  }
  # Modify data
  modes$hilic_neg$Injection_order <- modes$hilic_neg$Injection_order + 1
  inj_ord_rn <- modes$rp_neg$Injection_order + 2
  modes$rp_neg$Injection_order <- inj_ord_rn
  inj_ord_rp <- modes$rp_pos$Injection_order[5:50] + 5
  modes$rp_pos$Injection_order[5:50] <- inj_ord_rp
  colnames(modes$hilic_neg)[2] <- "ID_666"
  colData(modes$hilic_neg)$Sample_ID[2] <- "ID_666"
  colnames(modes$rp_pos)[22] <- "ID_999"
  colData(modes$rp_pos)$Sample_ID[22] <- "ID_999"

  expect_warning(merged <- merge_notame_sets(modes),
    regexp = "Sample IDs are not identical|Unequal amount of samples"
  )
  # Dummy injection
  expect_equal(merged$Injection_order, -seq_along(merged$Sample_ID))
  # Original IOs
  expect_equal(
    sort(as.numeric(stats::na.omit(merged$HILIC_neg_Injection_order))),
    modes$hilic_neg$Injection_order
  )
  expect_equal(
    sort(as.numeric(stats::na.omit(merged$HILIC_pos_Injection_order))),
    modes$hilic_pos$Injection_order
  )
  expect_equal(
    sort(as.numeric(stats::na.omit(merged$RP_neg_Injection_order))),
    modes$rp_neg$Injection_order
  )
  expect_equal(
    sort(as.numeric(stats::na.omit(merged$RP_pos_Injection_order))),
    modes$rp_pos$Injection_order
  )
})

context("Testing merging logic")

test_that("Flag column is combined correctly when merging batches", {
    se <- toy_notame_set
    se$Batch <- sort(rep(1:3, length.out = ncol(se)))
    batches <- sapply(unique(se$Batch), \(x) se[, se$Batch == x])
    # Only one flag
    flag(batches[[1]])[1:3] <- "A"
    flag(batches[[2]])[2:3] <- "A"
    merged <- merge_notame_sets(batches, merge = "samples")
    expected_flags <- c(rep("A", 3), rep(NA, nrow(merged) - 3))
    expect_equal(flag(merged), expected_flags)
    # Different flags
    # 1: any match
    # 2: tie, so we take the first one
    # 3: most common is B
    # 4: rest are NA
    flag(batches[[1]])[1:3] <- "A"
    flag(batches[[2]])[2:3] <- "B"
    flag(batches[[3]])[3] <- "B"
    merged <- merge_notame_sets(batches, merge = "samples")
    expected_flags <- c("A", "A", "B", rep(NA, nrow(merged) - 3))
    expect_equal(flag(merged), expected_flags)
})

context("Testing constructor class")

data(toy_notame_set, package = "notame")

test_that("Joining new columns to feature data preserves data", {
  # With more rows in data.frame
  new_info <- data.frame(Feature_ID = c(rownames(toy_notame_set), "additional"),
                         Feature_number = seq_len(nrow(toy_notame_set) + 1))
  with_new_info <- join_rowData(toy_notame_set, new_info)
  expect_equal(rowData(toy_notame_set), 
               rowData(with_new_info)[, 
                 which(colnames(rowData(with_new_info)) != "Feature_number")])
  
  # With fewer rows in data.frame
  new_info <- data.frame(
    Feature_ID = rownames(toy_notame_set)[1:nrow(toy_notame_set) -1],
    Feature_number = seq_len(nrow(toy_notame_set) - 1))
  with_new_info <- join_rowData(toy_notame_set, new_info)
  expect_equal(rowData(toy_notame_set)[1: nrow(toy_notame_set) - 1, ], 
               rowData(with_new_info)[1:nrow(toy_notame_set) - 1, 
                  which(colnames(rowData(with_new_info)) != "Feature_number")])

  # NAs are added to fill the difference if data.frame has fewer rows
  expect_equal(is.na(rowData(with_new_info)[nrow(toy_notame_set),
                                            "Feature_number"]),
               TRUE)
})
               
test_that("Joining new columns to pheno data preserves object", {
  # With more rows in data.frame
  new_info <- data.frame(
    Sample_ID = colnames(toy_notame_set),
    BMI = stats::runif(ncol(toy_notame_set), 22, 26))
  with_new_info <- join_colData(toy_notame_set, new_info)
  expect_equal(colData(toy_notame_set), 
               colData(with_new_info)[, 
                 which(colnames(colData(with_new_info)) != "BMI")])
  
  # With fewer rows in data.frame
  new_info <- data.frame(
    Sample_ID = colnames(toy_notame_set)[1:ncol(toy_notame_set) -1],
    BMI = seq_len(ncol(toy_notame_set) - 1))
  with_new_info <- join_colData(toy_notame_set, new_info)
  expect_equal(colData(toy_notame_set), colData(with_new_info)[, 
                 which(colnames(colData(with_new_info)) != "BMI")])

  # NAs are added to fill the difference if data.frame has fewer rows
  expect_equal(is.na(colData(with_new_info)[ncol(toy_notame_set), "BMI"]),
               TRUE)
})

#' Cluster correlated features originating from the same metabolite
#'
#' Clusters features potentially originating from the same compound. Features 
#' with high Pearson correlation coefficient and small retention time 
#' difference are linked together. Then clusters are formed by setting a 
#' threshold for the relative degree that each node in a cluster needs
#' to fulfil. Each cluster is named after the feature with the highest median 
#' peak area (median abundance). This is a wrapper around numerous functions 
#' that are based on the MATLAB code by David Broadhurst.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param mz_col the column name in feature data that holds mass-to-charge 
#' ratios
#' @param rt_col the column name in feature data that holds retention times
#' @param all_features logical, should all features be included in the 
#' clustering? If FALSE, as the default, flagged features are not included in 
#' clustering
#' @param rt_window the retention time window for potential links
#' NOTE: use the same unit as the retention time
#' @param corr_thresh the correlation threshold required for potential links 
#' between features
#' @param d_thresh the threshold for the relative degree required by each node
#' @param assay.type character, assay to be used in case of multiple assays
#'
#' @return a SummarizedExperiment object, with median peak area 
#' (MPA), the cluster ID, the features in the cluster, and cluster size added 
#' to feature data.
#'
#' @examples
#' data(toy_notame_set)
#' # The parameters are really weird because example data is imaginary
#' clustered <- cluster_features(toy_notame_set, rt_window = 1,
#'   corr_thresh = 0.5, d_thresh = 0.6)
#'
#' @export
cluster_features <- function(object, mz_col = NULL, rt_col = NULL,
                             all_features = FALSE, rt_window = 1 / 60,
                             corr_thresh = 0.9, d_thresh = 0.8, 
                             assay.type = NULL) {
  if (!requireNamespace("igraph", quietly = TRUE)) {
    stop("Package \"igraph\" needed for this function to work.",
         " Please install it.", call. = FALSE)
  }
  .add_citation(paste0("igraph package was used to construct networks of", 
                       " features for feature clustering:"), 
                citation("igraph"))

  # Drop flagged compounds before clustering
  from <- .get_from_name(object, assay.type)
  orig <- .check_object(object)
  object <- drop_flagged(object, all_features)
  object <- .check_object(object, check_limits = TRUE, assay.type = from,
                         feature_cols = c(mz_col, rt_col))

  if (is.null(mz_col) || is.null(rt_col)) {
    cols <- .find_mz_rt_cols(rowData(object))
  }
  mz_col <- mz_col %||% cols$mz_col
  rt_col <- rt_col %||% cols$rt_col

  data <- as.data.frame(t(assay(object, from)))
  features <- as.data.frame(rowData(object))
  # Start log
  log_text(paste("\nStarting feature clustering at", Sys.time()))

  # Find connections between features in each Split
  conn <- data.frame()
  for (s in unique(features$Split)) {
    log_text(paste("Finding connections between features in", s))
    features_tmp <- features[features$Split == s, ]
    data_tmp <- data[, features_tmp$Feature_ID]

    conn_tmp <- find_connections(data = data_tmp, features = features_tmp,
                                 corr_thresh = corr_thresh,
                                 rt_window = rt_window, name_col = "Feature_ID",
                                 mz_col = mz_col, rt_col = rt_col)
    
    conn <- rbind(conn, conn_tmp)
    log_text(paste("Found", nrow(conn_tmp), "connections in", s))
  }
  log_text(paste("Found", nrow(conn), "connections"))


  # Form clusters
  clusters <- find_clusters(conn, d_thresh)
  lens <- vapply(clusters, function(x) length(x$features), integer(1))
  log_text(paste("Found", sum(lens > 1),
                 "clusters of 2 or more features, clustering finished at",
                 Sys.time()))
  # Compute median peak area and assing cluster ID
  features$MPA <- apply(assay(object, from), 1, finite_median)
  features <- assign_cluster_id(data, clusters, features, "Feature_ID")

  # Add cluster IDs to the ORIGINAL object (flagged features still there)
  clustered <- join_rowData(orig, features[c("Feature_ID", "MPA",
                                            "Cluster_ID", "Cluster_size",
                                            "Cluster_features")])
  clustered
}

#' Assign Cluster ID to features
#'
#' Assigns a cluster ID to all features that are part of a cluster with 2 or 
#' more features.
#'
#' @param data data frame of the original LC-MS data
#' @param clusters a list of clusters as returned by find_clusters
#' @param features data frame with feature information, feature data
#' @param name_col character, name of the column in features that contains 
#' feature names
#'
#' @inherit find_connections return examples
#'
#' @return A data frame similar to features, with cluster ID added.
#'
#' @noRd
assign_cluster_id <- function(data, clusters, features, name_col) {
  if (!"MPA" %in% colnames(features)) {
    features$MPA <- vapply(data[, features[, name_col]], 
                           finite_median, numeric(1))
  }

  features$Cluster_ID <- features[, name_col]
  features$Cluster_features <- features[, name_col]
  features$Cluster_size <- 1
  n_clust <- length(clusters)
  for (cluster in clusters) {
    if (length(cluster$features) > 1) {
      # Which features are in the cluster
      idx <- features[, name_col] %in% cluster$features
      # The cluster is named for the feature with the largest median peak area
      features_tmp <- features[idx, ]
      max_mpa_idx <- which(features_tmp$MPA == max(features_tmp$MPA, 
                                                   na.rm = TRUE))[1]
      max_mpa_feature <- features_tmp[max_mpa_idx, name_col]
      # Saving some information about the clusters
      features$Cluster_ID[idx] <- paste0("Cluster_", max_mpa_feature)
      features$Cluster_size[idx] <- length(cluster$features)
      features$Cluster_features[idx] <- paste(sort(cluster$features), 
                                              collapse = ";")
    }
  }
  features
}

#' Compress clusters of features to a single feature
#'
#' This function compresses clusters found by cluster_features, keeping only 
#' the feature with the highest median peak area. The features that were 
#' discarded are recorded in feature data, under Cluster_features.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#'
#' @return A SummarizedExperiment object with only one feature per 
#' cluster.
#'
#' @examples
#' data(toy_notame_set)
#' clustered <- cluster_features(toy_notame_set, 
#'   rt_window = 1, corr_thresh = 0.5, d_thresh = 0.6)
#' compressed <- compress_clusters(clustered)
#'
#' @seealso \code{\link{cluster_features}}
#'
#' @export
compress_clusters <- function(object) {
  object <- .check_object(object)
  cluster_names <- rowData(object)$Cluster_ID
  if (is.null(cluster_names)) {
    stop("No 'Cluster_ID' found in rowData(object), ",
         "please run cluster_features first!")
  }
  # Get only "real" clusters
  clusters <- cluster_names[grepl("^Cluster_", cluster_names)] |>
    gsub("^Cluster_", "", x = _) |>
    unique()
  alone_features <- cluster_names[!grepl("^Cluster_", cluster_names)]
  # This ensures the order of the features stays the same
  idx <- rowData(object)$Feature_ID %in% c(clusters, alone_features)

  object <- object[idx, ]
  log_text(paste("Clusters compressed, left with", nrow(object), "features"))

  object
}

#' Extract information of the features in clusters
#'
#' For each cluster, the LC-MS data of the feature with largest median peak 
#' area is retained, all the features inside every cluster are recorded.
#'
#' @param data data frame of the original LC-MS data
#' @param features data frame holding the feature information
#' @param name_col name_col character, name of the column in features that
#' contains feature names
#'
#' @inherit find_connections return examples
#'
#' @return A list of two items:
#' \itemize{
#' \item cdata: a new data frame with the combined LC-MS data
#' \item cfeatures: data frame, feature information per cluster
#' }
#'
#' @noRd
pull_clusters <- function(data, features, name_col) {
  cluster_names <- features$Cluster_ID
  if (is.null(cluster_names)) {
    stop("No 'Cluster_ID' found in features, ",
         "please run assign_cluster_id first!")
  }
  # Get only "real" clusters
  clusters <- cluster_names[grepl("^Cluster_", cluster_names)] |>
    gsub("^Cluster_", "", x = _) |>
    unique()
  alone_features <- cluster_names[!grepl("^Cluster_", cluster_names)]
  # This ensures the order of the features stays the same
  idx <- features[, name_col] %in% c(clusters, alone_features)
  # Get sample features
  sample_cols <- setdiff(colnames(data), features[, name_col])

  features <- features[idx, ]
  data <- data[, c(sample_cols, features[, name_col])]

  return(list(cdata = data, cfeatures = features))
}


#' Find out which features are correlated within a specified retention time 
#' window
#'
#' A part of the peak clustering algorithm. Iterates over all possible pairs of 
#' features and records a connection between them if a) they have a Pearson 
#' correlation coefficient higher than \code{corr_thresh} and b) their 
#' retention time difference is less than \code{rt_window}. 
#'
#' @param data data frame with the abundances of features, with features as 
#' columns
#' @param features data frame with feature information, feature data
#' @param corr_thresh numeric, the threshold of correlation to use in linking 
#' features
#' @param rt_window numeric, the retention time window to use in linking 
#' features. NOTE: you need to use the same unit as in the retention time column
#' @param name_col character, name of the column in features that contains 
#' feature names
#' @param mz_col character, name of the column in features that contains mass-
#' to-charge ratios
#' @param rt_col character, name of the column in features that contains 
#' retention times
#'
#' @examples 
#' \dontshow{.old_wd <- setwd(tempdir())}
#' data(toy_notame_set)
#' data <- t(assay(toy_notame_set))
#' features <- rowData(toy_notame_set)
#' features$MPA <- sapply(data, finite_median)
#' conn <- find_connections(data = data, features = features,
#'   corr_thresh = 0.4, rt_window = 2,
#'   name_col = "Feature_ID", mz_col = "Average_Mz", rt_col = "Average_Rt_min")
#' clusters <- find_clusters(connections = conn, d_thresh = 0.6)
#' features_clustered <- assign_cluster_id(data, clusters, features,
#'   name_col = "Feature_ID")
#' pulled <- pull_clusters(data, features_clustered, name_col = "Feature_ID")
#' \dontshow{setwd(.old_wd)}
#'
#' @return A data frame of pairs of signals that are linked together:
#' \itemize{
#' \item x & y: indexes and names of the signals
#' \item cor: correlation coefficient
#' \item mz_diff & rt_diff: mass and retention time difference
#' }
#'
#' @noRd
find_connections <- function(data, features, corr_thresh = 0.9,
                             rt_window = 1 / 60, name_col, mz_col, rt_col) {
  if (ncol(data) < 2) {
    stop("Need at least 2 features to do any clustering!")
  }
  n <- nrow(features)

  rt_diff_mat <- abs(outer(features[, rt_col], features[, rt_col], "-"))
  mz_diff_mat <- abs(outer(features[, mz_col], features[, mz_col], "-"))
  cor_mat <- stats::cor(data, use = "na.or.complete")
  cor_mat[lower.tri(cor_mat, diag = TRUE)] <- NA
  idx <- which(rt_diff_mat < rt_window & cor_mat > corr_thresh, arr.ind = TRUE)

  data.frame(
    x = features[idx[, 1], name_col],
    y = features[idx[, 2], name_col],
    cor = cor_mat[idx],
    rt_diff = rt_diff_mat[idx],
    mz_diff = mz_diff_mat[idx]
  )
}

#' Extract the densely connected clusters
#'
#' First forms clusters of compounds that are linked together. Then the 
#' clusters are pruned so that in the final clusters, each feature is linked to 
#' at least a set percentage of the other features in the cluster.
#'
#' @param connections data frame of pairs of signals that are linked together,
#' output of find_connections
#' @param d_thresh numeric, the minimum degree required for each signal in a 
#' cluster expressed as a percentage of the maximum degree in the cluster
#'
#' @inherit find_connections return examples
#'
#' @return  A list of clusters, each a list of:
#' \itemize{
#' \item features: character vector of the names of the features included in 
#' the cluster
#' \item graph: an igraph object of the cluster
#' }
#'
#' @noRd
find_clusters <- function(connections, d_thresh = 0.8) {
  # Construct graph from the given edges
  g <- igraph::graph_from_edgelist(as.matrix(connections[seq_len(2)]), 
                                   directed = FALSE)
  g <- igraph::set.edge.attribute(graph = g, name = "weight", 
                                  value = connections$cor)

  # Initialize list of clusters
  clusters <- list()
  k <- 1

  # Repeatedly extract densely connected clusters from the graph
  while (length(igraph::V(g))) {
    # Connected components of the remaining graph
    comp <- igraph::decompose(g)
    n_comp <- length(comp)
    message(n_comp, " components found")

    # Only keep the densely connected part of each component (subgraph)
    clusters_tmp <- lapply(comp, function(subg) {
      n_nodes <- length(igraph::V(subg))
      d <- igraph::degree(subg)
      # The limit of the degree a node needs to be kept
      d_lim <- round(d_thresh * (n_nodes - 1))
      
      if (n_nodes >= 3) {
        # Remove the node with the smallest degree until all nodes in the
        # cluster have a degree above the limit
        while (any(d < d_lim)) {
          idx <- which(d == min(d))
          if (length(idx) > 1) {
            edgesums <- igraph::strength(subg, vids = igraph::V(subg)$name[idx])
            idx <- idx[which(edgesums == min(edgesums))[1]]
          }
          subg <- igraph::delete.vertices(subg, v = igraph::V(subg)[idx])
          d <- igraph::degree(subg)
          n_nodes <- n_nodes - 1
          d_lim <- round(d_thresh * (n_nodes - 1))
        }
      }

      # Record the final cluster and remove the nodes from the main graph
      list(list(features = names(igraph::V(subg)),
                graph = subg))
    })
    clusters_tmp <- do.call(c, clusters_tmp)

    for (j in seq_along(clusters_tmp)) {
      clusters[[k]] <- clusters_tmp[[j]]
      k <- k + 1
      subg <- clusters_tmp[[j]]$graph
      g <- igraph::delete.vertices(g, v = names(igraph::V(subg)))
    }
  }
  clusters
}

#' Initialize log to a file
#'
#' Initialize a log file with the current data and time.
#' All major operations run after this will be logged to the specified file.
#'
#' @param log_file Path to the log file
#'
#' @return None, the function is invoked for its side effect.
#'
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_name <- "log.txt"
#' init_log(file_name)
#' # Print the contents of the file
#' scan(file_name, sep = "\n", what = "character")
#' \dontshow{setwd(.old_wd)}
#' @seealso \code{\link{log_text}}, \code{\link{finish_log}}
#'
#' @export
init_log <- function(log_file) {
  futile.logger::flog.appender(futile.logger::appender.tee(log_file), 
                               name = "notame")
  log_text("Starting logging")
  # Pass errors to log
  options(error = function() {
    futile.logger::flog.error(geterrmessage(), name = "notame")
  })
}

#' Log text to the current log file
#'
#' The specified text is printed and written to the current log file. Does not 
#' overwrite the file. Also used internally by many functions in the package. 
#'
#' @param text The text to be logged
#'
#' @return None, the function is invoked for its side effect.
#'
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_name <- "log.txt"
#' init_log(file_name)
#' log_text("Hello World!")
#' # Print the contents of the file
#' scan(file_name, sep = "\n", what = "character")
#' \dontshow{setwd(.old_wd)}
#'
#' @seealso \code{\link{init_log}}, \code{\link{finish_log}}
#'
#' @export
log_text <- function(text) {
  futile.logger::flog.info(text, name = "notame")
}

#' Finish a log
#'
#' Logs the current date and time and session info, and switches logging off.
#'
#' @return None, the function is invoked for its side effect.
#'
#' @examples
#' finish_log()
#'
#' @seealso \code{\link{init_log}}, \code{\link{log_text}}
#'
#' @export
finish_log <- function() {
  # Return default option for error
  options(error = NULL)
  # Log end of session info
  futile.logger::flog.info(paste("Finished analysis. ", date(), 
                                 "\nSession info:\n", sep = ""))
  futile.logger::flog.info(utils::capture.output(utils::sessionInfo()))
  invisible(futile.logger::flog.appender(futile.logger::appender.console(),
                                         name = "notame"))
}

context("Testing logging")

test_that("Error/warning handling changes when logging and is set to NULL after logging", {
  # Save current error handling
  orig_error <- options("error")
  orig_warn <- options("warning.expression")
  log_file <- tempfile()
  init_log(log_file)
  # Initiating log should change error handling
  expect_failure(expect_identical(options("error")$error, orig_error$error))
  finish_log()
  # Finishing log should set error to "NULL"
  expect_null(options("error")$error)
  expect_null(options("warning.expression")$warning.expression)
})

test_that("Logging works as expected", {
  log_file <- tempfile()
  log_text("New line")
  # File doesn't exist yet
  expect_warning(expect_error(readLines(log_file), "cannot open the connection"))

  init_log(log_file)
  logs <- readLines(log_file)
  # Last line should match
  expect_identical(gsub("\\[.+\\] ", "", logs[length(logs)]), "INFO Starting logging")
  log_text("New line")
  logs <- readLines(log_file)
  # Last line should match
  expect_identical(gsub("\\[.+\\] ", "", logs[length(logs)]), "INFO New line")
  finish_log()
  logs <- readLines(log_file)
  log_len <- length(logs)
  # This shouldn't log to file anymore
  log_text("New line")
  logs <- readLines(log_file)
  expect_identical(log_len, length(logs))
})

#' \code{notame} package.
#'
#' Provides tabular data preprocessing and data wrangling functionality for 
#' untargeted LC-MS metabolomics research.
#' 
#' @name notame-package
#' @references
#' Klåvus et al. (2020). "notame": Workflow for Non-Targeted LC-MS Metabolic 
#' Profiling. Metabolites, 10: 135.
"_PACKAGE"
NULL

#' @rawNamespace import(ggplot2, except = Position)
#' @importFrom utils citation
#' @import BiocGenerics
#' @import methods
#' @import SummarizedExperiment
NULL

utils::globalVariables(c('i', '.'))

#' Set default color scales on load
#'
#' @param libname,pkgname default parameters
#' @noRd
.onLoad <- function(libname, pkgname) {
  op <- options()
  op_notame <- list(
    notame.citations = list(
      "Preprocessing and analyses were performed using notame package:" =
      utils::citation("notame"),
      "The primary data structure in notame is SummarizedExperiment:" = 
      utils::citation("SummarizedExperiment"),
      "visualizations in notame are built with ggplot2:" =
      utils::citation("ggplot2")),
    notame.color_scale_con = scale_color_viridis_c(),
    notame.color_scale_dis = scale_color_brewer(palette = "Set1"),
    notame.fill_scale_con = scale_fill_viridis_c(),
    notame.fill_scale_dis = scale_fill_brewer(palette = "Set1"),
    notame.fill_scale_div_con = scale_fill_distiller(palette = "RdBu"),
    notame.fill_scale_div_dis = scale_fill_brewer(palette = "RdBu"),
    notame.shape_scale = scale_shape_manual(values = c(16, 17, 15, 3,
                                                       7, 8, 11, 13)))
  toset <- !(names(op_notame) %in% names(op))
  if (any(toset)) {
    options(op_notame[toset])
  }
  invisible()
}

.add_citation <- function(name, ref) {
  cites <- getOption("notame.citations")
  if (!name %in% names(cites)) {
    cites[[name]] <- ref
    options(notame.citations = cites)
  }
}

#' Show citations
#'
#' This function lists citations behind the notame functions that have been 
#' called during the session. All notame functions 
#' update the list automatically. The citations are taken from the call to 
#' '\code{citation("package")}, and complemented with a brief description of 
#' what the package was used for.
#' NOTE: the citations might not point to the correct paper if the package 
#' authors have not supplied correct citation information for their package.
#' The output is written to the current log file, if specified.
#'
#' @return None, the function is invoked for its side effect.
#'
#' @examples
#' citations()
#' data(toy_notame_set)
#' ex_set <- flag_quality(toy_notame_set)
#' # Broadhurst et al.(2018) added to citations
#' citations()
#'
#' @export
citations <- function() {
  cites <- getOption("notame.citations")
  for (i in seq_along(cites)) {
    log_text(names(cites)[i])
    log_text(utils::capture.output(show(cites[[i]])))
  }
}


#' Summary statistics of finite elements
#'
#' These functions first remove non-finite and missing values, then 
#' compute the summary statistic in question. They are helper 
#' functions used for computing quality measurements.
#' 
#' @param x a numeric vector.
#' @param ... other parameters passed to underlying function
#' @return A named, numeric vector with the summary statistic in question.
#'
#' @name finite_helpers
#' @noRd
NULL

finite_sd <- function(x) {
  sd(x[is.finite(x)], na.rm = TRUE)
}

finite_mean <- function(x) {
  if (all(is.na(x))) {
    return(NA_real_)
  }
  mean(x[is.finite(x)], na.rm = TRUE)
}

finite_median <- function(x) {
  stats::median(x[is.finite(x)], na.rm = TRUE)
}

finite_min <- function(x) {
  if (all(is.na(x))) {
    return(NA_real_)
  }
  min(x[is.finite(x)], na.rm = TRUE)
}

finite_max <- function(x) {
  if (all(is.na(x))) {
    return(NA_real_)
  }
  max(x[is.finite(x)], na.rm = TRUE)
}

finite_mad <- function(x) {
  mad(x[is.finite(x)], 
      center = stats::median(x[is.finite(x)], na.rm = TRUE), 
      na.rm = TRUE)
}

finite_quantile <- function(x, ...) {
  unname(stats::quantile(x[is.finite(x)], na.rm = TRUE, ...))
}

# Defaults for NULL values
`%||%` <- function(a, b) {
  if (is.null(a)) {
    b
  } else if (is.na(a)) {
    b
  } else {
    a
  }
}

#' Proportion of NA values in a vector
#'
#' @param x a numeric vector
#'
#' @return A numeric, the proportion of non-missing values in a vector.
#'
#' @examples
#' data(toy_notame_set)
#' ex_set <- mark_nas(toy_notame_set, value = 0)
#' prop_na(assay(ex_set))
#' 
#' @noRd
prop_na <- function(x) {
  sum(is.na(x)) / length(x)
}

#' Proportion of non-missing values in a vector
#'
#' @param x a numeric vector
#' 
#' @return A numeric, the proportion of non-missing values in vector.
#'
#' @examples
#' data(toy_notame_set)
#' ex_set <- mark_nas(toy_notame_set, value = 0)
#' prop_found(assay(toy_notame_set))
#'
#' @noRd
prop_found <- function(x) {
  sum(!is.na(x)) / length(x)
}

.best_class <- function(x) {
  x <- as.character(x)
  x <- gsub("false", "FALSE", x, ignore.case = TRUE)
  x <- gsub("true", "TRUE", x, ignore.case = TRUE)
  x <- utils::type.convert(x, as.is = TRUE, na.strings = c("NA", "", "null"))
  if ((is.character(x) || is.integer(x)) && length(unique(x)) < 9) {
    x <- as.factor(x)
  }
  x
}

.best_classes <- function(x) {
  as.data.frame(lapply(x, .best_class), stringsAsFactors = FALSE)
}

.all_unique <- function(x) {
  !any(duplicated(x))
}

.get_from_to_names <- function(object, assay.type, name) {
  # Input behavior (from)
  # If assay.type is not supplied and there is only one assay in the objcet, 
  # choose the first assay
  if (is.null(assay.type) && length(assays(object)) == 1) {
    from <- 1
  } else if (is.null(assay.type)) {
    stop("When using multiple assays, specify assay.type", call. = FALSE)
  } else if (!assay.type %in% names(assays(object)) & assay.type != 1) {
    stop(assay.type, " was specified but not found in assays", call. = FALSE)
  } else {
    from <- assay.type
  }
  # Output behavior (to)
  if (is.null(name) && length(assays(object)) == 1) {
    to <- 1
  } else if (is.null(name)) {
    stop("When using multiple assays, specify name of new assay", call. = FALSE)
  } else {
    to <- name
  }
  list(from, to)
}

.get_from_name <- function(object, assay.type) {
  # Input behavior (from)
  if (is.null(assay.type) && length(assays(object)) == 1) {
    from <- 1
  } else if (is.null(assay.type)) {
    stop("When using multiple assays, specify assay.type", call. = FALSE)
  } else if (!assay.type %in% names(assays(object)) & assay.type != 1) {
    stop(assay.type, " was specified but not found in assays", call. = FALSE)
  } else {
    from <- assay.type
  }
}

context("Finite helper functions")

test_that("Work correctly for full vectors", {
  x <- 13:20
  expect_equal(finite_sd(x), sd(x))
  expect_equal(finite_mean(x), mean(x))
  expect_equal(finite_median(x), median(x))
  expect_equal(finite_mad(x), mad(x))
})

test_that("NAs and finites are excluded correctly", {
  x <- 1:10
  x[c(2, 7)] <- Inf
  x[c(4, 9)] <- NA
  y <- c(1, 3, 5, 6, 8, 10)
  expect_equal(finite_sd(x), sd(y))
  expect_equal(finite_mean(x), mean(y))
  expect_equal(finite_median(x), median(y))
  expect_equal(finite_mad(x), mad(y))
})

# Check that two SummarizedExperiment objects can be combined
#
# Checks many matching criteria, basically pheno data needs to have similar
# special columns,
# the number of samples needs to be the same and feature data need to have the
# same columns names. Throws an error if any of the criteria is not fulfilled.
#
# @param x,y SummarizedExperiment objects
.check_match <- function(x, y) {
  # Lots of checks to ensure that everything goes smoothly

  # Amount of samples must be equal
  if (nrow(colData(x)) != nrow(colData(y))) {
    warning("Unequal amount of samples")
  }
  # Resulting feature ID must be unique
  feature_id <- c(rowData(x)$Feature_ID, rowData(y)$Feature_ID)
  if (!.all_unique(feature_id)) {
    stop("Merge would result in duplicated feature ID")
  }
  common <- intersect(colnames(x), colnames(y))
  if (!identical(colData(x)[common, "Injection_order"], 
                 colData(y)[common, "Injection_order"])) {
    stop("Injection orders of common samples are not identical")
  }
  if (!identical(colData(x)$Sample_ID, colData(y)$Sample_ID)) {
    warning("Sample IDs are not identical")
    samples_x <- setdiff(colnames(x), colnames(y))
    samples_y <- setdiff(colnames(y), colnames(x))
    log_text("Merging objects with unequal amounts of samples.")
    log_text("Samples only in first object:")
    log_text(paste0(paste(samples_x, collapse = ", "), "\n"))
    log_text("Samples only in second object:")
    log_text(paste0(paste(samples_y, collapse = ", "), "\n"))
  }


  overlap_cols <- intersect(colnames(colData(x)), colnames(colData(y))) |>
    setdiff(c("Sample_ID", "Injection_order"))

  if (length(overlap_cols)) {
    for (overlap_col in overlap_cols) {
      if (!identical(colData(x)[common, overlap_col], 
                     colData(y)[common, overlap_col])) {
        stop("Columns named ", overlap_col, 
             " in pheno data have different content")
      }
    }
  }

  if (!identical(colnames(rowData(x)), colnames(rowData(y)))) {
    stop("Feature data have different column names")
  }
}

# Merge two SummarizedExperiment objects together
.merge_mode_helper <- function(x, y) {
  # Create dummy injection order if original ones differ
  common <- intersect(colnames(x), colnames(y))
  if (!identical(colData(x)[common, "Injection_order"], 
      colData(y)[common, "Injection_order"])) {
    log_text(paste0("Injection order differs between modes.",
                    "Creating dummy injection order"))
    x_modes <- unique(rowData(x)$Split)
    # Save original injection order for first mode
    if (length(x_modes) == 1) {
      colData(x)[, paste0(x_modes, "_Injection_order")] <- x$Injection_order
    }
    dummy_injection <- as.numeric(-seq_along(colData(x)$Sample_ID))
    names(dummy_injection) <- x$Sample_ID
    colData(x)$Injection_order <- dummy_injection
    # Save original injection order in other modes
    colData(y)[, paste0(unique(rowData(y)$Split), "_Injection_order")] <-
      y$Injection_order
    # Update dummy injection
    y_in_x <- y$Sample_ID %in% x$Sample_ID
    new_io <- seq(from = min(dummy_injection) - 1, length.out = sum(!y_in_x))
    names(new_io) <- y$Sample_ID[!y_in_x]
    dummy_injection <- append(dummy_injection, new_io)

    colData(y)$Injection_order <- dummy_injection[match(colData(y)$Sample_ID,
                                                        names(dummy_injection))]
    log_text("Dummy injection order (row numbers) created")
  }
  # Check that the match is ok
  .check_match(x, y)

  merged_coldata <- dplyr::full_join(as.data.frame(colData(x)), 
                                     as.data.frame(colData(y)),
                                     by = intersect(colnames(colData(x)),
                                                    colnames(colData(y)))) |>
    S4Vectors::DataFrame()
  rownames(merged_coldata) <- merged_coldata$Sample_ID
  merged_rowdata <- rbind(rowData(x), rowData(y)) |>
    S4Vectors::DataFrame()
  if (identical(colnames(assay(x)), colnames(assay(y)))) {
    merged_assay <- rbind(assay(x), assay(y))
  } else {
    merged_assay <- dplyr::bind_rows(as.data.frame(assay(x)),
                                     as.data.frame(assay(y))) |> 
      as.matrix()
    rownames(merged_assay) <- rownames(merged_rowdata)
  }

  merged_object <- SummarizedExperiment(assays = merged_assay, 
                                        colData = merged_coldata,
                                        rowData = merged_rowdata)

  merged_object
}

# Convert objects in ... to a list
# OR if a list is given in the first place, preserve that list
.to_list <- function(...) {
  # Combine the objects to a list
  objects <- list(...)
  # If a list is given in the first place, it should move to top level
  if (length(objects) == 1) {
    if (is(objects[[1]], "list")) {
      objects <- objects[[1]]
    }
  }

  objects
}

#' Merge SummarizedExperiment objects together
#'
#' Merges two or more SummarizedExperiment objects together. Can be used to
#' merge analytical modes or batches.
#'
#' @param ... \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' objects or a list of objects
#' @param merge what to merge? features is used for combining analytical modes,
#' samples is used for batches
#' @param assay.type character, assay to be used in case of multiple assays.
#' The same assay needs to be present in all objects to be merged, and the
#' resultant object contains this single assay.
#'
#' @return A merged SummarizedExperiment object.
#'
#' @details When merging samples, sample IDs that begin with "QC" or "Ref" are
#' combined so that they have running numbers on them. This means that if both
#' batches have samples called "QC_1", this will not result in an error,
#' but the sample IDs will be adjusted so that they are unique. Column names in
#' the feature data that are shared between batches but have different content
#' are renamed by adding a suffix to avoid data loss. The suffix is the index
#' of the batch in the input list. The most common flag value across batches
#' is included in the "Flag" column in the merged object.
#'
#' @examples
#' # Merge analytical modes
#' data(hilic_neg_sample, hilic_pos_sample, rp_neg_sample, rp_pos_sample)
#' merged <- merge_notame_sets(
#'   hilic_neg_sample, hilic_pos_sample,
#'   rp_neg_sample, rp_pos_sample
#' )
#' # Merge batches
#' batch1 <- toy_notame_set[, toy_notame_set$Batch == 1]
#' batch2 <- toy_notame_set[, toy_notame_set$Batch == 2]
#' merged <- merge_notame_sets(batch1, batch2, merge = "samples")
#'
#' @export
merge_notame_sets <- function(
  ...,
  merge = c("features", "samples"),
  assay.type = NULL
) {
  merge <- match.arg(merge)
  # Combine the objects to a list
  objects <- .to_list(...)
  # Class check
  if (!all(vapply(objects, class, character(1)) == "SummarizedExperiment")) {
    stop("The arguments should only contain SummarizedExperiment objects")
  }
  # Check assay.type and prepare objects for merge
  from_list <- lapply(objects, .get_from_name, assay.type)
  if (!length(unique(from_list)) == 1) {
    stop(
      "The same assay must be present in all objects or alternatively,
          use objects with a single assay"
    )
  } else {
    from <- unlist(unique(from_list))
  }
  objects <- lapply(objects, function(object) {
    assays(object) <- assays(object)[from]
    object
  })

  # Choose merging function
  if (merge == "features") {
    merged <- Reduce(.merge_mode_helper, objects)
  } else {
    merged <- .merge_batch_helper(objects)
    merged <- .recreate_flag(merged)
  }
  merged
}

.rowdata_batch_helper <- function(list_of_dfs) {
  # Find all common columns
  list_of_dfs <- lapply(list_of_dfs, as.data.frame)
  all_cols <- lapply(list_of_dfs, colnames)
  common_cols <- Reduce(intersect, all_cols)

  # Find columns with same name but different content
  # If not all contents are identical, mark for suffixing
  nonmatching_cols_idx <- vapply(
    common_cols,
    function(col) {
      col_list <- lapply(list_of_dfs, `[[`, col)
      !all(vapply(col_list[-1], identical, logical(1), col_list[[1]]))
    },
    logical(1)
  )

  # Add suffix to nonmatching columns in each dataframe
  nonmatching_cols <- common_cols[nonmatching_cols_idx]
  list_of_dfs <- mapply(
    function(df, idx) {
      cols_to_rename <- colnames(df) %in% nonmatching_cols
      colnames(df)[cols_to_rename] <- paste0(
        colnames(df)[cols_to_rename],
        "_",
        idx
      )
      df
    },
    list_of_dfs,
    seq_along(list_of_dfs),
    SIMPLIFY = FALSE
  )

  # Merge all dataframes
  merged <- Reduce(
    function(x, y) {
      dplyr::full_join(x, y, by = common_cols[!nonmatching_cols_idx])
    },
    list_of_dfs
  )
  merged <- S4Vectors::DataFrame(merged)
  merged
}

# Helper function to recreate the Flag column after merging batches, since the
# original Flag columns are renamed to avoid data loss, we need to combine them
# back into a single Flag column that contains all the information.
.recreate_flag <- function(se) {
  flags <- grep("Flag", colnames(rowData(se)))
  # If we have less than 2 flag columns, we can just return the original object
  # since the merge of Flag columns was successful
  if (length(flags) < 2) {
    return(se)
  }
  # Compute the most common flag for each feature across batches
  # and use that as the final flag for the merged object
  flag_mat <- rowData(se)[, flags, drop = FALSE]
  unique_flags <- c(sort(unique(as.vector(as.matrix(flag_mat)))))
  counts <- apply(flag_mat, 1, function(row) {
    tabulate(factor(row, levels = unique_flags), nbins = length(unique_flags))
  }, simplify = FALSE)
  most_common <- unname(vapply(counts, function(count) {
    if (all(count == 0)) {
      return(NA_character_)
    }
    unique_flags[which.max(count)]
  }, character(1)))
  rowData(se)$Flag <- most_common
  se
}

.check_batch_coldata <- function(cd) {
  if (anyDuplicated(cd$Sample_ID)) {
    log_text(paste0(
      "Found duplicated sample IDs when merging, ",
      "renaming QC and Ref samples"
    ))
    qc_idx <- grepl("QC", cd$Sample_ID)
    cd$Sample_ID[qc_idx] <- paste0(
      "QC_",
      seq_len(sum(qc_idx))
    )
    ref_idx <- grepl("Ref", cd$Sample_ID)
    cd$Sample_ID[ref_idx] <- paste0(
      "Ref_",
      seq_len(sum(ref_idx))
    )
    rownames(cd) <- cd$Sample_ID
  }
  cd
}

.merge_batch_helper <- function(objects) {
  # Handle column data
  cds <- lapply(objects, colData)
  merged_coldata <- do.call(rbind, cds)
  merged_coldata <- .check_batch_coldata(merged_coldata)

  # Handle assay data, bind_rows to preserve all features
  assays <- lapply(objects, assay)
  assays <- lapply(assays, \(df) as.data.frame(t(df)))
  merged_assay <- t(as.matrix(dplyr::bind_rows(assays)))
  colnames(merged_assay) <- rownames(merged_coldata)

  # Handle row data
  rds <- lapply(objects, rowData)
  merged_rowdata <- .rowdata_batch_helper(rds)

  merged_object <- SummarizedExperiment(
    assays = merged_assay,
    colData = merged_coldata,
    rowData = merged_rowdata
  )

  merged_object
}

#' Extract quality information of features
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' 
#' @return A data frame with quality metrics for each feature.
#'
#' @examples
#' data(toy_notame_set)
#' ex_set <- assess_quality(toy_notame_set)
#' quality(ex_set)
#'
#' @export
quality <- function(object) {
  object <- .check_object(object)
  if (!all(c("RSD", "RSD_r", "D_ratio", "D_ratio_r") %in%
      colnames(rowData(object)))) {
    return(NULL)
  }
  as.data.frame(
    rowData(object)[c("Feature_ID", "RSD", "RSD_r", "D_ratio", "D_ratio_r")])
}

.erase_quality <- function(object) {
  if (!all(c("RSD", "RSD_r", "D_ratio", "D_ratio_r") %in%
      colnames(rowData(object)))) {
    return(NULL)
  }
  rowData(object)[c("RSD", "RSD_r", "D_ratio", "D_ratio_r")] <- NULL

  object
}



#' Assess quality information of features
#'
#' Assess features using the quality metrics defined in (Broadhurst 
#' 2018). The quality metrics are described in Details section of 
#' \code{\link{flag_quality}}
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param assay.type character, assay to be used in case of multiple assays
#'
#' @return A SummarizedExperiment object with quality metrics in 
#' feature data.
#' 
#' @examples
#' data(toy_notame_set)
#' ex_set <- assess_quality(toy_notame_set)
#' rowData(ex_set)
#'
#' @export
assess_quality <- function(object, assay.type = NULL) {
  from <- .get_from_name(object, assay.type)
  object <- .check_object(object, pheno_QC = TRUE, assay.type = from)
  # Remove old quality metrics
  if (!is.null(quality(object))) {
    object <- .erase_quality(object)
  }

  qc_idx <- object$QC == "QC"
  mat <- assay(object, from)
  qc_mat <- mat[, qc_idx, drop = FALSE]
  nonqc_mat <- mat[, !qc_idx, drop = FALSE]

  qc_sd <- apply(qc_mat, 1, finite_sd)
  qc_mean <- apply(qc_mat, 1, finite_mean)
  qc_mad <- apply(qc_mat, 1, finite_mad)
  qc_median <- apply(qc_mat, 1, finite_median)
  nonqc_sd <- apply(nonqc_mat, 1, finite_sd)
  nonqc_mad <- apply(nonqc_mat, 1, finite_mad)

  quality_metrics <- data.frame(
    Feature_ID = rownames(object),
    RSD = qc_sd / abs(qc_mean),
    RSD_r = qc_mad / abs(qc_median),
    D_ratio = qc_sd / nonqc_sd,
    D_ratio_r = qc_mad / nonqc_mad,
    stringsAsFactors = FALSE
  )
  
  join_rowData(object, quality_metrics)
}

#' Flag low-quality features
#'
#' Flags low-quality features using the quality metrics defined in (Broadhurst 
#' 2018). The metrics are described in more detain in Details. A condition for 
#' keeping the features is given as a character, which is passed to 
#' \code{\link[dplyr]{filter}}.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param condition character, condition for keeping the features, see Details
#' @param assay.type character, assay to be used in case of multiple assays
#'
#' @details The quality metrics measure two things: internal spread of the QCs,
#' and spread of the QCs compared to the spread of the biological samples.
#' Internal spread is measured with relative standard deviation (RSD), also 
#' known as coefficient of variation (CV).
#' \deqn{RSD = sd(QC) / mean(QC) }
#' Where \eqn{sd(QC)} is the standard deviation of the QC samples and 
#' '\eqn{mean(QC)} is the sample mean of the signal in the QC samples.
#' RSD can also be replaced by a non-parametric, robust version based on the 
#' median and median absolute deviation (MAD):
#'   \deqn{RSD_r = 1.4826 * MAD(QC) / median(QC)}
#' The spread of the QC samples compared to the biological samples is measured 
#' using a metric called D-ratio:
#'   \deqn{D_ratio = sd(QC) / sd(biological)}
#' Or, as before, a non-parametric, robust alternative:
#'   \deqn{D_ratio_r = MAD(QC) / MAD(biolofical) }
#' The default condition keeps features that pass either of the two following 
#' conditions:
#' \deqn{RSD_r < 0.2 \& D_ratio_r < 0.4}
#' \deqn{RSD < 0.1 \& RSD_r < 0.1 \& D_ratio < 0.1}
#'
#' @return a SummarizedExperiment object with the features flagged.
#'
#' @references Broadhurst, David et al. Guidelines and considerations for the 
#' use of system suitability and quality control samples in mass spectrometry 
#' assays applied in untargeted clinical metabolomic studies.
#' Metabolomics : Official journal of the Metabolomic Society vol. 14,6 (2018): 
#' 72. doi:10.1007/s11306-018-1367-3
#'  
#' @usage flag_quality
#' flag_quality(object, assay.type = NULL, condition =
#'   "(RSD_r < 0.2 & D_ratio_r < 0.4) | 
#'   (RSD < 0.1 & RSD_r < 0.1 & D_ratio < 0.1)")
#' @examples
#' data(toy_notame_set)
#' ex_set <- flag_quality(toy_notame_set)
#' rowData(ex_set)
#' # Custom condition
#' ex_set <- flag_quality(toy_notame_set,
#'   condition = "RSD_r < 0.3 & D_ratio_r < 0.6")
#' rowData(ex_set)
#'
#' @export
flag_quality <- function(object, assay.type = NULL, condition = 
  "(RSD_r < 0.2 & D_ratio_r < 0.4) | 
  (RSD < 0.1 & RSD_r < 0.1 & D_ratio < 0.1)") {
  if (is.null(quality(object))) {
    object <- assess_quality(object, assay.type)
  }
  object <- .check_object(object, feature_ID = TRUE)
  .add_citation("Quality metrics were computed as per guidelines in:", paste(
    "Broadhurst, David et al. Guidelines and considerations for the use of",
    "system suitability and quality control samples in mass spectrometry",
    "assays applied in untargeted clinical metabolomic studies.",
    "Metabolomics : Official journal of the Metabolomic Society",
    "vol. 14,6 (2018): 72. doi:10.1007/s11306-018-1367-3"))

  good <- paste0("as.data.frame(rowData(object)) |> 
                 dplyr::filter(", condition, ")") |>
    parse(text = _) |>
    eval()
  good <- good$Feature_ID
  
  idx <- is.na(flag(object)) & !rowData(object)$Feature_ID %in% good
  flag(object)[idx] <- "Low_quality"

  percentage <- scales::percent(sum(flag(object) == "Low_quality", 
                                    na.rm =TRUE) 
                                / nrow(rowData(object)))
  log_text(paste0("\n", percentage, " of features flagged for low quality"))

  object
}

#' Flag features with low detection rate
#'
#' Flags features with too high amount of missing values. There are two 
#' detection rate limits, both defined as the minimum proportion of samples 
#' that need to have a value (not NA) for the feature to be kept. 
#' '\code{qc_limit} is the detection rate limit for QC samples, 
#' '\code{group_limit} is the detection rate limit for the actual study groups.
#' If the group limit is passed for AT LEAST ONE GROUP, then the feature is 
#' kept. Features with low detection rate in QCs are flagged as 
#' "Low_qc_detection", while low detection rate in the study groups is flagged 
#' as "Low_group_detection". The detection rates for all the groups are 
#' recorded in feature data.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param qc_limit the detection rate limit for QC samples
#' @param group_limit the detection rate limit for study groups
#' @param group the columns name in sample information to use as the grouping 
#' variable
#' @param assay.type character, assay to be used in case of multiple assays
#' @return A SummarizedExperiment object with the features flagged.
#'
#' @examples
#' data(toy_notame_set)
#' ex_set <- mark_nas(toy_notame_set, value = 0)
#' ex_set <- flag_detection(ex_set, group = "Group")
#' rowData(ex_set)
#'
#' @export
flag_detection <- function(object, qc_limit = 0.7, group_limit = 0.5,
                           group = NULL, assay.type = NULL) {
  from <- .get_from_name(object, assay.type)
  object <- .check_object(object, pheno_ID = TRUE, pheno_QC = TRUE, 
                         pheno_factors = group, assay.type = from, 
                         feature_ID = TRUE, feature_flag = TRUE)
  found_qc <- apply(assay(object[, object$QC == "QC"], from), 1, prop_found)
  bad_qc <- names(which(found_qc < qc_limit))

  found_qc_df <- data.frame(Feature_ID = names(found_qc),
                            Detection_rate_QC = found_qc,
                            stringsAsFactors = FALSE)

  idx <- is.na(flag(object)) & rowData(object)$Feature_ID %in% bad_qc
  flag(object)[idx] <- "Low_qc_detection"

  # Compute proportions found in each study group
  if (!is.null(group)) {
    proportions <- combined_data(object, from)[, c("Sample_ID", group,
                                               rownames(object))] |>
      tidyr::gather("Feature_ID", "Intensity", rownames(object)) |>
      dplyr::group_by(.data$Feature_ID, !!as.name(group)) |>
      dplyr::summarise(proportion_found = prop_found(.data$Intensity)) |>
      tidyr::spread(!!as.name(group), "proportion_found")
    # Remove a possible QC column
    proportions$QC <- NULL
    colnames(proportions)[-1] <- paste0("Detection_rate_", group, "_",
                                        colnames(proportions)[-1])
    # Check if any group has enough non-missing entries
    proportions$good <- apply(proportions[-1], 1, function(x) {
      any(x >= group_limit)
    })

    idx <- is.na(flag(object)) & (!rowData(object)$Feature_ID %in% 
      proportions$Feature_ID[proportions$good])
    flag(object)[idx] <- "Low_group_detection"
    # Add detection rates to feature data
    proportions <- dplyr::left_join(proportions, found_qc_df, by = "Feature_ID")
    proportions$good <- NULL
  } else {
    proportions <- found_qc_df
  }

  percentage <- scales::percent(sum(flag(object) %in% c("Low_qc_detection",
                                                        "Low_group_detection"), 
                                    na.rm = TRUE) / 
                                nrow(rowData(object)))
  log_text(paste0("\n", percentage, 
                  " of features flagged for low detection rate"))

  object <- join_rowData(object, proportions)

  object
}


#' Flag contaminants based on blanks
#'
#' Flags contaminant features by comparing either median, mean or max of blanks 
#' and biological samples. Biological samples are defined as samples that are 
#' not marked as blanks and are not QCs.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param blank_col character, the column name in colData with blank labels
#' @param blank_label character, the label for blank samples in blank_col
#' @param blank_type character, one of "mean", "median", or "max"
#' @param sample_type character, one of "max", "median", or "mean"
#' @param flag_thresh numeric, the scaled ratio threshold for flagging 
#' contaminants.
#' If blank_type(blanks) * flag_thresh > sample_type(biological samples), the 
#' feature gets flagged.
#' @param flag_label character, the label used when flagging contaminants. Can 
#' be changed if sample processing contaminants and carryover contaminants are 
#' flagged separately.
#' @param assay.type character, assay to be used in case of multiple assays
#' @return A SummarizedExperiment object with contaminant features 
#' flagged.
#'
#' @details
#' If the calculation(biological samples) < the calculation(blanks) times a set 
#' ratio, the feature is flagged as contaminant. Default calculations are "max" 
#' for biological samples and "mean" for blanks.
#' 
#' @examples
#' data(toy_notame_set)
#' # Make a blank sample which has one (first) feature exceeding the threshold
#' ## Abundance matrix
#' max_sample <- max(assay(toy_notame_set)[1, toy_notame_set$QC != "QC"])
#' assay <- matrix(c(max_sample * 0.20 + 1, rep(0, 79)), ncol = 1, nrow = 80, 
#'                   dimnames = list(NULL, "Demo_51"))
#' assay <- cbind(assay(toy_notame_set), assay)
#' ## Sample metadata
#' pheno_data <- colData(toy_notame_set)[1, ]
#' rownames(pheno_data) <- "Demo_51"
#' pheno_data$Sample_ID <- "Demo_51"
#' pheno_data$Injection_order <- 51
#' pheno_data[c("Subject_ID", "Group", "QC", "Time")] <- "Blank"
#' pheno_data <- rbind(colData(toy_notame_set), pheno_data)
#' ## Feature metadata
#' feature_data <- rowData(toy_notame_set)
#'
#' # Construct SummarizedExperiment object with blank sample
#' ex_set <- SummarizedExperiment(assays = assay, 
#'                                colData = pheno_data,
#'                                rowData = feature_data)
#' # Flag contaminant(s)
#' contaminants_flagged <- flag_contaminants(ex_set, blank_col = "QC", 
#'                                           blank_label = "Blank")
#' 
#' @export
flag_contaminants <- function(object, blank_col, blank_label,
                              blank_type = c("mean", "median", "max"),
                              sample_type = c("max", "median", "mean"),
                              flag_thresh = 5, flag_label = "Contaminant",
                              assay.type = NULL) {
  blank_type <- tryCatch(
    match.arg(blank_type),
    error = function(e) stop("Wrong contaminant calculation type selected", 
    call. = FALSE)
  )
  sample_type <- tryCatch(
    match.arg(sample_type),
    error = function(e) stop("Wrong sample calculation type selected", 
    call. = FALSE)
  )
  
  from <- .get_from_name(object, assay.type)
  object <- .check_object(object, pheno_QC = TRUE, pheno_cols = blank_col, 
                         assay.type = from)
  
  blanks <- object[, colData(object)[, blank_col] == blank_label]
  samples <- object[, object$QC != "QC" &
                    colData(object)[, blank_col] != blank_label]
  
  stat_functions <- list(
    median = finite_median,
    mean   = finite_mean,
    max    = finite_max
  )
  
  # select the function to be used
  blank_fun  <- stat_functions[[blank_type]]
  sample_fun <- stat_functions[[sample_type]]

  blank_statistic <- apply(assay(blanks, from), 1, blank_fun)
  sample_statistic <- apply(assay(samples, from), 1, sample_fun)
  blank_flag <- sample_statistic / blank_statistic  < flag_thresh

  idx <- is.na(flag(object)) & !is.na(blank_flag)
  idx <- idx & blank_flag
  flag(object)[idx] <- flag_label

  percentage <- scales::percent(sum(flag(object) == flag_label, na.rm = TRUE) /
                                nrow(object))
  log_text(paste0("\n", percentage, " of features flagged as contaminants"))

  blank_ratio <- data.frame(Feature_ID = rownames(object), 
                            Blank_ratio = round(sample_statistic / blank_statistic,digits = 1),
                            stringsAsFactors = FALSE)
  object <- join_rowData(object, blank_ratio)

  object
}

#' Remove Unwanted Variation (RUV) between batches
#'
#' An interface for \code{link[RUVSeq]{RUVs}} method
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param replicates list of numeric vectors, indexes of replicates
#' @param k The number of factors of unwanted variation to be estimated from 
#' the data.
#' @param assay.type character, assay to be used in case of multiple assays
#' @param name character, name of the resultant assay in case of multiple assays
#' @param ... other parameters passed to \code{link[RUVSeq]{RUVs}}
#' @return A SummarizedExperiment object with the normalized data.
#'
#' @examples
#' data(toy_notame_set)
#' # Batch correction
#' replicates <- list(which(toy_notame_set$QC == "QC"))
#' batch_corrected <- ruvs_qc(toy_notame_set, replicates = replicates)
#' # Evaluate batch correction
#' pca_bhattacharyya_dist(toy_notame_set, batch = "Batch")
#' pca_bhattacharyya_dist(batch_corrected, batch = "Batch")
#'
#' @export
ruvs_qc <- function(object, replicates, k = 3, 
                    assay.type = NULL, name = NULL, ...) {
  if (!requireNamespace("RUVSeq", quietly = TRUE)) {
    stop("Bioconductor package RUVSeq needed for this function to work.",
         " Please install it.", call. = FALSE)
  }
  
  .add_citation("RUVSeq was used for batch correction:", citation("RUVSeq"))    
  
  from_to <- .get_from_to_names(object, assay.type, name)
  object <- .check_object(object, assay.type = from_to[[1]])
  
  # Transform data to pseudo counts for RUVs
  assay(object, from_to[[1]])[assay(object, from_to[[1]]) == 0] <- 1
  assay(object) <- round(assay(object))
  # Pad each replicate vector with -1 and transform to matrix
  max_len <- max(vapply(replicates, length, integer(1)))
  scIdx <- matrix(-1, nrow = length(replicates), ncol = max_len)
  # Populate matrix of replicates
  for (i in seq_along(replicates)) {
    scIdx[i, seq_along(replicates[[i]])] <- replicates[[i]]
  }
  # Perform batch correction
  ruv_results <- RUVSeq::RUVs(x = assay(object), cIdx = rownames(object),
                              k = k, scIdx = scIdx, ...)
  # Include results in object
  assay(object, from_to[[2]]) <- ruv_results$normalizedCounts
  colData(object) <- cbind(colData(object), ruv_results$W)
  
  object
}


#' Bhattacharyya distance between batches in PCA space
#'
#' Computes Bhattacharyya distance between all pairs of batches using 
#' \code{\link[fpc:bhattacharyya.matrix]{fpc:bhattacharyya.matrix}} 
#' after projecting the samples into PCA space with 
#' \code{\link[pcaMethods]{pca}}.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param batch column name of pheno data giving the batch labels
#' @param all_features logical, should all features be used? If FALSE
#' (the default), flagged features are removed before imputation.
#' @param center logical, should the data be centered prior to PCA?
#' (usually yes)
#' @param scale scaling used, as in 
#' \code{\link[pcaMethods]{prep}}. Default is "uv" for unit variance
#' @param nPcs the number of principal components to use
#' @param assay.type character, assay to be used in case of multiple assays
#' @param ... other parameters to \code{\link[pcaMethods]{pca}}
#'
#' @return A matrix of Bhattacharyya distances between batches.
#'
#' @examples
#' data(toy_notame_set)
#' # Batch correction
#' replicates <- list(which(toy_notame_set$QC == "QC"))
#' batch_corrected <- ruvs_qc(toy_notame_set, replicates = replicates)
#' # Evaluate batch correction
#' pca_bhattacharyya_dist(toy_notame_set, batch = "Batch")
#' pca_bhattacharyya_dist(batch_corrected, batch = "Batch")
#'
#' @export
pca_bhattacharyya_dist <- function(object, batch, all_features = FALSE, 
                                   center = TRUE, scale = "uv", nPcs = 3, 
                                   assay.type = NULL, ...){ 
  if (!requireNamespace("fpc", quietly = TRUE)) {
    stop("Package \"fpc\" needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("pcaMethods", quietly = TRUE)) {
    stop("Package \"pcaMethods\" needed for this function to work.", 
         " Please install it.", call. = FALSE)
  }
  .add_citation("PCA was performed using pcaMethods package:",
                citation("pcaMethods"))
  .add_citation("fpc package was used for Bhattacharyaa distance computation:",
                citation("fpc"))
  from <- .get_from_name(object, assay.type)
  # Drop flagged features if not told otherwise
  object <- .check_object(object, pheno_factors = batch, assay.type = from)
  object <- drop_flagged(object, all_features)
  # PCA to 2 dimensions
  pca_res <- pcaMethods::pca(t(assay(object, from)), center = center, 
                             scale = scale, nPcs = nPcs, ...)
  pca_scores <- pcaMethods::scores(pca_res)
  # Split to batches
  batches <- list()
  for (b in unique(colData(object)[, batch])) {
    batches[[b]] <- pca_scores[colData(object)[, batch] == b, ]
  }
  # Compute means and covariance matrices for Bhattacharyya distance
  muarray <- vapply(batches, colMeans, double(nPcs))
  sigmaarray <- array(vapply(batches, stats::cov, double(nPcs * nPcs)), 
                      dim = c(nPcs, nPcs, length(batches)))
  # Compute Bhattacharyya distance
  fpc::bhattacharyya.matrix(muarray, sigmaarray, ipairs = "all",
                            misclassification.bound = FALSE)
}


.pooled_variance <- function(x, group) {
  # Remove missing values
  group <- group[!is.na(x)]
  x <- x[!is.na(x)]
  # Split to groups
  group_list <- split(x, group)
  n_1 <- vapply(group_list, length, integer(1)) - 1 # n - 1
  # Pooled variance
  sum(n_1 * vapply(group_list, var, numeric(1))) / sum(n_1)
}

.between_variance <- function(x, group) {
  # Remove missing values
  group <- group[!is.na(x)]
  x <- x[!is.na(x)]
  # Split to groups
  group_list <- split(x, group)
  n <- vapply(group_list, length, integer(1))
  means <- vapply(group_list, mean, numeric(1))
  k_1 <- length(unique(group)) - 1
  # Between group variance formula
  sum(n * (means - mean(x))^2) / k_1
}

.repeatability <- function(mat, group) {
  # Calculate pooled variance
  pv <- apply(mat, 1, .pooled_variance, group)
  # Calculate between group variance
  bv <- apply(mat, 1, .between_variance, group)
  bv / (bv + pv)
}


#' Compute repeatability measures
#'
#' Computes repeatability for each feature with the following formula:
#' \deqn{\frac{\sigma^2_{between}}{\sigma^2_{between} + \sigma^2_{within}}}
#' The repeatability ranges from 0 to 1. Higher repeatability depicts less
#' variation between batches.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param group column name of pheno data giving the group labels
#' @param assay.type character, assay to be used in case of multiple assays
#'
#' @return A data frame with one row per feature with the repeatability measure.
#'
#' @examples
#' data(toy_notame_set)
#' # Batch correction
#' replicates <- list(which(toy_notame_set$QC == "QC"))
#' batch_corrected <- ruvs_qc(toy_notame_set, replicates = replicates)
#' # Evaluate batch correction
#' rep_orig <- perform_repeatability(toy_notame_set, group = "Group")
#' mean(rep_orig$Repeatability, na.rm = TRUE)
#' rep_corr <- perform_repeatability(batch_corrected, group = "Group")
#' mean(rep_corr$Repeatability, na.rm = TRUE)
#'
#' @export
perform_repeatability <- function(object, group, assay.type = NULL) {
  from <- .get_from_name(object, assay.type)
  object <- .check_object(object, pheno_factors = group, assay.type = from)
  group <- colData(object)[, group]
  
  data.frame(
    "Feature_ID" = rownames(object),
    "Repeatability" = .repeatability(assay(object, from), group)
  )
}

#' Toy data set
#'
#' Contains imaginary data used in testing the package functions. The dataset 
#' has 50 samples and 80 features. This dataset includes multiple observations 
#' from same subjects, sampled at two timepoints in separate batches and 
#' divided to two groups. The analytical modes are also available as separate 
#' SummarizedExperiment objects.
#' Note that across batches, the features don't have different feature ID's, 
#' m/z and retention time as would be the case with real-world data. In 
#' essence, the example data reflects that features were aligned perfectly 
#' between batches. 
"toy_notame_set"

#' @rdname toy_notame_set
"hilic_neg_sample"

#' @rdname toy_notame_set
"hilic_pos_sample"

#' @rdname toy_notame_set
"rp_neg_sample"

#' @rdname toy_notame_set
"rp_pos_sample"

#' Mark specified values as missing
#'
#' Replaces all values in the peak table that equal the specified value 
#' with NA.
#' For example, vendor software might use 0 or 1 to signal a missing value,
#' which is not understood by R.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param value the value to be converted to NA
#' @param assay.type character, assay to be used in case of multiple assays
#' @param name character, name of the resultant assay in case of multiple assays
#'
#' @return SummarizedExperiment object as the one supplied, with 
#' missing values correctly set to NA.
#'
#' @examples
#' data(toy_notame_set)
#' nas_marked <- mark_nas(toy_notame_set, value = 0)
#'
#' @export
mark_nas <- function(object, value, assay.type = NULL, name = NULL) {
  from_to <- .get_from_to_names(object, assay.type, name)
  object <- .check_object(object, assay.type = from_to[[1]])
  ex <- assay(object, from_to[[1]])
  ex[ex == value] <- NA
  assay(object, from_to[[2]]) <- ex

  object
}

#' Transform the MS/MS output to publication ready
#'
#' Change the MS/MS output from MS-DIAL format to publication-ready format.
#' Original spectra is sorted according to abundance percentage and clarified. 
#' See the example below.
#'
#' Original MS/MS spectra from MS-DIAL:
#' m/z:Raw Abundance
#'
#' 23.193:254 26.13899:5 27.50986:25 55.01603:82 70.1914:16 73.03017:941 
#' 73.07685:13 73.13951:120
#'
#' Spectra after transformation:
#' m/z  (Abundance)
#'
#' 73.03 (100), 23.193 (27), 73.14 (12.8), 55.016 (8.7)
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param ms_ms_spectrum_col name of column with original MS/MS spectra
#' @param peak_num maximum number of peak that is kept (Recommended: 4-10)
#' @param min_abund minimum relative abundance to be kept (Recommended: 1-5)
#' @param deci_num maximum number of decimals to m/z value (Recommended: >2)
#'
#' @return A SummarizedExperiment object as the one supplied, with 
#' publication-ready MS/MS peak information.
#'
#' @examples
#' data(toy_notame_set)
#' # Spectra before fixing
#' ex_set <- toy_notame_set
#' rowData(ex_set)$MS_MS_spectrum <- NA
#' rowData(ex_set)[1, ]$MS_MS_spectrum <- 
#'   "28.769:53 44.933:42 52.106:89 69.518:140"
#' rowData(ex_set)$MS_MS_spectrum[
#'   !is.na(rowData(ex_set)$MS_MS_spectrum)]
#' # Fixing spectra with default settings
#' fixed_MSMS_peaks <- fix_MSMS(ex_set)
#' # Spectra after fixing
#' rowData(fixed_MSMS_peaks)$MS_MS_Spectrum_clean[
#'   !is.na(rowData(fixed_MSMS_peaks)$MS_MS_Spectrum_clean)]
#'
#' @export
fix_MSMS <- function(object, ms_ms_spectrum_col = "MS_MS_spectrum",
                     peak_num = 10, min_abund = 5, deci_num = 3) {
  object <- .check_object(object, feature_cols = ms_ms_spectrum_col)
  spec <- rowData(object)[, ms_ms_spectrum_col]
  to_metab <- NULL

  for (i in seq_along(spec)) {
    # Check the feature spectra and skip if it doesn't exist
    if (is.na(spec[i])) {
      to_metab[i] <- NA
      next()
    }
    # Transform format
    spectrum <- spec[i]
    spectrum2 <- t(stringr::str_split(spectrum, pattern = " ", simplify = TRUE))
    spectrum3 <- as.data.frame(stringr::str_split(spectrum2, 
                                                  pattern = ":", 
                                                  simplify = TRUE))
    spectrum3 <- as.data.frame(lapply(spectrum3, as.numeric))
    spectrum3 <- spectrum3[order(spectrum3$V2, decreasing = TRUE), ]

    # Leave n most intense fragment peaks or all peaks if number of peaks < n
    ifelse(nrow(spectrum3) > peak_num, num <- peak_num, num <- nrow(spectrum3))
    spectrum4 <- spectrum3[c(seq_len(num)), ]

    # Round m/z of fragments to n decimals and calculate relative intensity (%)
    spectrum4$V1 <- round(spectrum4$V1, digits = deci_num)
    spectrum4$relative <- round(spectrum4$V2 / max(spectrum3$V2) * 100, 
                                digits = 1)

    # Remove peaks with relative intensity less than n% (recommended: 1-5)
    spectrum5 <- spectrum4[c(spectrum4$relative > min_abund), ]

    # Finalize format and write results
    to_metab[i] <- paste(paste0(spectrum5$V1, " (", spectrum5$relative, ")"),
                         collapse = ", ")
  }

  rowData(object)$MS_MS_Spectrum_clean <- to_metab
  log_text(paste0("Saving fixed MS/MS spectra to column",
                  " \'MS_MS_Spectrum_clean\' in rowData"))

  object
}

#' Drop QC samples
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#'
#' @return A SummarizedExperiment object as the one supplied, 
#' without QC samples.
#'
#' @examples
#' data(toy_notame_set)
#' dim(toy_notame_set)
#' noqc <- drop_qcs(toy_notame_set)
#' dim(noqc)
#'
#' @export
drop_qcs <- function(object) {
  object <- .check_object(object)
  object <- object[, object$QC != "QC"]
  colData(object) <- droplevels(colData(object))

  object
}


#' Drop flagged features
#'
#' Removes all features that have been flagged by quality control functions.
#' Only features that do not have a flag (Flag == NA) are retained.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param all_features logical, should all features be retained? Mainly used by 
#' internal functions
#' 
#' @return A SummarizedExperiment object without the previously 
#' flagged features.
#'
#' @examples
#' data(toy_notame_set)
#' dim(toy_notame_set)
#' flagged <- flag_quality(toy_notame_set)
#' noflags <- drop_flagged(flagged)
#' dim(noflags)
#'
#' @export
drop_flagged <- function(object, all_features = FALSE) {
  object <- .check_object(object)
  if (!all_features) {
    object <- object[is.na(flag(object)), ]
  }
  object
}


#' Partially replace peak table with new values
#'
#' Replaces a subset of data in peak table of an object by new values.
#' Used after an operation such as imputation is computed only for a subset of 
#' features or samples such as only good-quality features that have not been 
#' flagged. This function is mainly used internally, but can be useful.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param y matrix containing new values to be merged into peak table
#'
#' @return A SummarizedExperiment object with the new peak table 
#' values.
#'
#' @examples
#' data(toy_notame_set)
#' ex_set <- toy_notame_set[1:5, 1:5]
#' assay(ex_set)
#' # Create a matrix of replacment values for rows 1, 3, 5 and columns 1, 3, 4
#' replacement <- matrix(1:9,
#'   ncol = 3,
#'   dimnames = list(
#'     rownames(ex_set)[c(1, 3, 5)],
#'     colnames(ex_set)[c(1, 3, 4)]
#'   )
#' )
#' replacement
#' merged <- merge_assay(ex_set, replacement)
#' assay(merged)
#'
#' @noRd
merge_assay <- function(object, y, assay.type = NULL, name = NULL) {
  # Colnames and rownames should be found in the object
  if (!all(colnames(y) %in% colnames(assay(object))) || is.null(colnames(y))) {
    stop("Column names of y do not match column names of assay(object).")
  }
  if (!all(rownames(y) %in% rownames(assay(object))) || is.null(rownames(y))) {
    stop("Row names of y do not match row names of assay(object).")
  }
  from_to <- .get_from_to_names(object, assay.type, name)
  assay_tmp <- assay(object, from_to[[1]])
  assay_tmp[rownames(y), colnames(y)] <- y
  assay(object, from_to[[2]]) <- assay_tmp
  object
}

#' Impute missing values using random forest
#'
#' Impute the missing values in the peak table of the object using a
#' random forest. The estimated error in the imputation is logged.
#' It is recommended to set the seed number for reproducibility
#' (it is called random forest for a reason).
#' This a wrapper around \code{\link[missForest]{missForest}}.
#' Use parallelize = "variables" to run in parallel for faster testing.
#' NOTE: running in parallel prevents user from setting a seed number.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param all_features logical, should all features be used? If FALSE (the 
#' default), flagged features are removed before imputation.
#' @param assay.type character, assay to be used in case of multiple assays
#' @param name character, name of the resultant assay in case of multiple assays
#' @param ... passed to \code{\link[missForest]{missForest}}
#'
#' @return An object as the one supplied, with missing 
#' values imputed.
#'
#' @examples
#' data(toy_notame_set)
#' missing <- mark_nas(toy_notame_set, 0)
#' set.seed(38)
#' imputed <- impute_rf(missing)
#'
#' @seealso \code{\link[missForest]{missForest}} for detail about the algorithm 
#' and the parameters
#'
#' @export
impute_rf <- function(object, all_features = FALSE, assay.type = NULL, 
                      name = NULL, ...) {
  if (!requireNamespace("missForest", quietly = TRUE)) {
    stop("Package \'missForest\' needed for this function to work.",
         " Please install it.", call. = FALSE)
  }
  .add_citation("missForest package was used for random forest imputation:",
                citation("missForest"))
  # Start log
  log_text(paste("\nStarting random forest imputation at", Sys.time()))
  # Drop flagged features
  dropped <- drop_flagged(object, all_features)
  from_to <- .get_from_to_names(object, assay.type, name)
  dropped <- .check_object(dropped, assay.type = from_to[[1]])
  object <- .check_object(object, assay.type = from_to[[1]])

  if (!requireNamespace("missForest", quietly = TRUE)) {
    stop("missForest package not found.")
  }

  # Impute missing values
  mf <- missForest::missForest(xmis = t(assay(dropped, from_to[[1]])), ...)
  imputed <- t(mf$ximp)
  # Log imputation error
  log_text(paste0("Out-of-bag error in random forest imputation: ",
                  round(mf$OOBerror, digits = 3)))
  # Assign imputed data to the droppped
  rownames(imputed) <- rownames(assay(dropped))
  colnames(imputed) <- colnames(assay(dropped))
  # Attach imputed abundances to object
  object <- merge_assay(object, imputed, assay.type = from_to[[1]], 
                        name = from_to[[2]])
  log_text(paste("Random forest imputation finished at", Sys.time(), "\n"))
  
  object
}

#' Simple imputation
#'
#' Impute missing values using a simple imputation strategy. All missing values
#' of a feature are imputed with the same value. It is possible
#' to only impute features with a large number of missing values this way. This 
#' can be useful for using this function before random forest imputation to 
#' speed things up.
#' The imputation strategies available are:
#' \itemize{
#' \item a numeric value: impute all missing values in all features with the 
#' same value, e.g. 1
#' \item "mean": impute missing values of a feature with the mean of observed 
#' values of that feature
#' \item "median": impute missing values of a feature with the median of 
#' observed values of that feature
#' \item "min": impute missing values of a feature with the minimum observed 
#' value of that feature
#' \item "half_min": impute missing values of a feature with half the minimum 
#' observed value of that feature
#' \item "small_random": impute missing values of a feature with random numbers 
#' between 0 and the
#' minimum of that feature (uniform distribution, remember to set the seed 
#' number!).
#' }
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param value the value used for imputation, either a numeric or one of 
#' '"min", "half_min", "small_random", see above
#' @param na_limit only impute features with the proportion of NAs over this 
#' limit. For example, if \code{na_limit = 0.5}, only features with at least 
#' half of the values missing are imputed.
#' @param assay.type character, assay to be used in case of multiple assays
#' @param name character, name of the resultant assay in case of multiple assays
#'
#' @return A SummarizedExperiment object with imputed peak table.
#'
#' @examples
#' data(toy_notame_set)
#' missing <- mark_nas(toy_notame_set, 0)
#' imputed <- impute_simple(missing, value = "min")
#'
#' @export
impute_simple <- function(object, value, na_limit = 0, assay.type = NULL,
                          name = NULL) {
  from_to <- .get_from_to_names(object, assay.type, name)
  object <- .check_object(object, assay.type = from_to[[1]])
  imp <- assay(object, from_to[[1]])
  nas <- apply(imp, 1, prop_na)
  imp <- imp[nas > na_limit, , drop = FALSE]
  if (nrow(imp) == 0) {
    warning("None of the features satisfy the NA limit,", 
            " returning the original object.")
    return(object)
  }

  # Replace all missing values with the given constant
  if (is.numeric(value)) {
    imp[is.na(imp)] <- value
  } else if (value == "mean") {
    imp <- t(apply(imp, 1, function(x) {
      x[is.na(x)] <- finite_mean(x)
      x
    }))
  } else if (value == "median") {
    imp <- t(apply(imp, 1, function(x) {
      x[is.na(x)] <- finite_median(x)
      x
    }))
  } else if (value == "min") {
    imp <- t(apply(imp, 1, function(x) {
      x[is.na(x)] <- finite_min(x)
      x
    }))
  } else if (value == "half_min") {
    imp <- t(apply(imp, 1, function(x) {
      x[is.na(x)] <- finite_min(x) / 2
      x
    }))
  } else if (value == "small_random") {
    imp <- t(apply(imp, 1, function(x) {
      x[is.na(x)] <- stats::runif(n = sum(is.na(x)), 
                                  min = 0, max = finite_min(x))
      x
    }))
  } else {
    stop("value should be a numeric value or one of 'min',",
         " 'half_min', 'small_random'.")
  }

  obj <- merge_assay(object, imp, assay.type = from_to[[1]],
                     name = from_to[[2]])
  
  obj
}

#' Inverse-rank normalization
#'
#' Applies inverse rank normalization to all features to approximate
#' a normal distribution.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param assay.type character, assay to be used in case of multiple assays
#' @param name character, name of the resultant assay in case of multiple assays
#' @return An object as the one supplied, with normalized features.
#'
#' @examples
#' data(toy_notame_set)
#' normalized <- inverse_normalize(toy_notame_set)
#' @export
inverse_normalize <- function(object, assay.type = NULL, name = NULL) {
  from_to <- .get_from_to_names(object, assay.type, name)
  object <- .check_object(object, assay.type = from_to[[1]])
  assay(object, from_to[[2]]) <- assay(object, from_to[[1]]) |>
    apply(1, function(x) {
      stats::qnorm((rank(x, na.last = "keep") - 0.5) / sum(!is.na(x)))
    }) |>
    t()

  object
}


#' A report of flagged features
#'
#' Computes the number of features at each stage of flagging for each mode.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#'
#' @return A data frame with the number of features at each stage of flagging.
#'
#' @examples
#' data(toy_notame_set)
#' flagged <- toy_notame_set |>
#'   mark_nas(0) |>
#'   flag_detection(group = "Group") |>
#'   flag_quality()
#' flag_report(flagged)
#'
#' @export
flag_report <- function(object) {
  object <- .check_object(object, feature_split = TRUE)
  splits <- sort(unique(rowData(object)$Split))
  report <- data.frame()
  flag(object)[is.na(flag(object))] <- "Kept"
  for (split in splits) {
    tmp <- object[rowData(object)$Split == split, ]
    report_row <- flag(tmp) |>
      table() |>
      as.matrix() |>
      t()
    report_row <- data.frame(Split = split, report_row)
    if (is.null(report_row$Kept)) {
      report_row$Kept <- 0
    }
    report_row$Total <- nrow(tmp)
    report_row$Flagged <- report_row$Total - report_row$Kept
    report <- dplyr::bind_rows(report, report_row)
  }
  report
}

#' Probabilistic quotient normalization
#'
#' Apply probabilistic quotient normalization (PQN) to the peak table of a 
#' SummarizedExperiment object. By default, reference is 
#' calculated from high-quality QC samples and the median of the reference is 
#' used for normalization. Check parameters for more options.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param ref character, the type of reference samples to use for normalization.
#' @param method character, the method to use for calculating the reference 
#' sample.
#' @param all_features logical, should all features be used for calculating the 
#' reference sample?
#' @param assay.type character, assay to be used in case of multiple assays
#' @param name character, name of the resultant assay in case of multiple assays
#'
#' @return A SummarizedExperiment object with altered feature 
#' abundances.
#'
#' @examples
#' data(toy_notame_set)
#' pqn_set <- pqn_normalization(toy_notame_set)
#'
#' @export
pqn_normalization <- function(object, ref = c("qc", "all"),
                              method = c("median", "mean"), 
                              all_features = FALSE, assay.type = NULL,
                              name = NULL) {
  log_text("Starting PQN normalization")
  ref <- match.arg(ref)
  method <- match.arg(method)
  # Use only good-quality features for calculating reference spectra
  from_to <- .get_from_to_names(object, assay.type, name)
  object <- .check_object(object, pheno_QC = TRUE, assay.type = from_to[[1]])
  ref_data <- assay(drop_flagged(object, all_features), from_to[[1]])
  
  
  # Select reference samples
  switch(ref, qc = reference <- ref_data[, object$QC == "QC"],
         all = reference <- ref_data)
  if (ncol(reference) == 0 || all(is.na(reference))) {
    stop("No specified reference samples found.")
  }
  # Calculate reference spectrum
  switch(method,
         median = reference_spectrum <- apply(reference, 1, finite_median),
         mean = reference_spectrum <- apply(reference, 1, finite_mean))
  log_text(paste("Using", method, "of", ref, "samples as reference spectrum"))
  # Calculate median of quotients
  quotients <- ref_data / reference_spectrum
  quotient_md <- apply(quotients, 2, finite_median)
  # Do the normalization
  data <- assay(object, from_to[[1]])
  pqn_data <- t(t(data) / quotient_md)
  colnames(pqn_data) <- colnames(data)
  rownames(pqn_data) <- rownames(data)
  assay(object, from_to[[2]]) <- pqn_data
  
  object
}

context("Testing transformations")

data(toy_notame_set, package = "notame")

test_that("Marking NAs works properly", {
  marked <- mark_nas(toy_notame_set, value = 0)
  zero_idx <- assay(toy_notame_set) == 0
  na_idx <- is.na(assay(marked))

  expect_equal(zero_idx, na_idx)
})


test_that("RF imputation works as expected", {
  marked <- mark_nas(toy_notame_set, value = 0)
  imputed <- impute_rf(marked)

  # Check that all missing values are imputed
  expect_equal(sum(is.na(assay(imputed))), 0)

  # Check that non-missing values are unchanged
  na_idx <- is.na(assay(marked))

  non_na_marked <- assay(marked)[!na_idx]
  non_na_imputed <- assay(imputed)[!na_idx]
  expect_equal(non_na_imputed, non_na_marked)
})

test_that("Flagging works as expected", {
  single_change <- matrix(1111,
    nrow = 1, ncol = 1,
    dimnames = list(
      "HILIC_neg_108_1065a2_6121",
      "Demo_1"
    )
  )
  mrg <- merge_assay(toy_notame_set, single_change)
  expect_equal(assay(mrg)[2, 1], 1111)

  block_change <- matrix(stats::runif(9),
    nrow = 3, ncol = 3,
    dimnames = list(
      rownames(assay(toy_notame_set))[4:6],
      colnames(assay(toy_notame_set))[5:7]
    )
  )
  mrg2 <- merge_assay(toy_notame_set, block_change)
  expect_true(all(assay(mrg2)[4:6, 5:7] == block_change))

  scattered_change <- matrix(stats::runif(9),
    nrow = 3, ncol = 3,
    dimnames = list(
      rownames(assay(toy_notame_set))[c(1, 5, 7)],
      colnames(assay(toy_notame_set))[c(2, 5, 8)]
    )
  )
  mrg3 <- merge_assay(toy_notame_set, scattered_change)
  expect_true(all(assay(mrg3)[c(1, 5, 7), c(2, 5, 8)] == scattered_change))

  wrong_names <- matrix(stats::runif(9), nrow = 3)
  expect_error(merge_assay(toy_notame_set, wrong_names), "Column names")

  wrong_names2 <- matrix(stats::runif(9),
    nrow = 3,
    dimnames = list(
      letters[1:3],
      colnames(assay(toy_notame_set))[5:7]
    )
  )
  expect_error(merge_assay(toy_notame_set, wrong_names2), "Row names")
})


test_that("Flagged compounds are not imputed", {
  marked <- mark_nas(toy_notame_set, 0)
  flag(marked)[c(1, 4, 6)] <- "Flagged"

  imputed <- impute_rf(marked)
  nas <- apply(assay(imputed), 1, prop_na)

  expect_true(all(nas[c(1, 4, 6)] > 0))
  expect_true(all(nas[-c(1, 4, 6)] == 0))

  # All feature parameter test
  imputed <- impute_rf(marked, all_features = TRUE)
  nas <- apply(assay(imputed), 1, prop_na)
  expect_true(all(nas == 0))
})

test_that("Inverse normalization works as expected", {
  marked <- mark_nas(toy_notame_set, value = 0)
  imputed <- impute_rf(marked)
  normalized <- inverse_normalize(imputed)

  # Ranks should be identical
  orig_ranks <- assay(imputed) |>
    apply(1, rank) |>
    t()
  norm_ranks <- assay(normalized) |>
    apply(1, rank) |>
    t()
  expect_identical(norm_ranks, orig_ranks)

  # Check that all the rows follow standard normal distribution
  # Zero means up to 0.1 accuracy
  abs_means <- abs(rowMeans(assay(normalized)))
  expect_true(all(abs_means < 0.1))
  # Unit variance
  sd_diffs <- abs(apply(assay(normalized), 1, sd) - 1)
  expect_true(all(sd_diffs < 0.1))

  # Shapiro-Wilk normality tests are non-significant
  shapiro_p <- assay(normalized) |>
    apply(1, function(x) {
      shapiro.test(x)$p.value
    })
  expect_true(all(shapiro_p > 0.9))
})

test_that("Simple imputation works as expected", {
  marked <- mark_nas(toy_notame_set, value = 0)
  imputed <- impute_simple(marked, value = 0)

  # Check that all missing values are imputed
  expect_equal(sum(is.na(assay(imputed))), 0)

  # Check that non-missing values are unchanged
  na_idx <- is.na(assay(marked))

  non_na_marked <- assay(marked)[!na_idx]
  non_na_imputed <- assay(imputed)[!na_idx]
  expect_equal(non_na_imputed, non_na_marked)
})

test_that("Simple imputation with only one feature works", {
  marked <- toy_notame_set
  assay(marked)[2, 5] <- NA
  imputed <- impute_simple(marked, value = 0)
  #
  expect_equal(assay(imputed)[2, 5], 0)
})

test_that("PQN normalization works correctly using median of QC samples as reference", {
  data <- assay(toy_notame_set)
  # Calculate the median of QC samples
  reference <- apply(data[, toy_notame_set$QC == "QC"], 1, finite_median)
  # do the normalization
  quotient <- data / reference
  quotient_md <- apply(quotient, 2, finite_median)
  pqn_data <- t(t(data) / quotient_md)

  expect_equal(pqn_data, assay(pqn_normalization(toy_notame_set)))
})

test_that("PQN normalization works correctly using median of all samples as reference", {
  data <- assay(toy_notame_set)
  # Calculate the median of all samples
  reference <- apply(data, 1, finite_median)
  # do the normalization
  quotient <- data / reference
  quotient_md <- apply(quotient, 2, finite_median)
  pqn_data <- t(t(data) / quotient_md)

  expect_equal(pqn_data, assay(pqn_normalization(toy_notame_set, ref = "all")))
})

test_that("PQN normalization works correctly using mean of QC samples as reference", {
  data <- assay(toy_notame_set)
  # Calculate the mean of QC samples
  reference <- apply(data[, toy_notame_set$QC == "QC"], 1, finite_mean)
  # do the normalization
  quotient <- data / reference
  quotient_md <- apply(quotient, 2, finite_median)
  pqn_data <- t(t(data) / quotient_md)

  expect_equal(pqn_data, assay(pqn_normalization(toy_notame_set, method = "mean")))
})

test_that("PQN normalization works correctly using mean of all samples as reference", {
  data <- assay(toy_notame_set)
  # Calculate the mean of all samples
  reference <- apply(data, 1, finite_mean)
  # do the normalization
  quotient <- data / reference
  quotient_md <- apply(quotient, 2, finite_median)
  pqn_data <- t(t(data) / quotient_md)

  expect_equal(pqn_data, assay(pqn_normalization(toy_notame_set, ref = "all", 
                                                 method = "mean")))
})

test_that("PQN normalization works with flagged features", {
  flagged_mset <- flag_quality(toy_notame_set)
  ref_data <- assay(drop_flagged(flagged_mset))
  reference_spectrum <- apply(ref_data[, toy_notame_set$QC == "QC"], 1, finite_median)
  quotient <- ref_data / reference_spectrum
  quotient_md <- apply(quotient, 2, finite_median)

  data <- assay(flagged_mset)
  pqn_data <- t(t(data) / quotient_md)

  pqn_mset <- pqn_normalization(flagged_mset)
  expect_equal(pqn_data, assay(pqn_mset))
})

test_that("Assay control works for transformations (with drift correction)", {
  ex_set <- toy_notame_set[1:10, ]
  # New assays can be added to objects with one assay
  corrected <- correct_drift(ex_set, name = "corrected")
  # Assay not found with a single unnamed assay
  expect_error(correct_drift(ex_set, assay.type = "original"))
  # Assay not found with multiple assays
  assay(ex_set, "alternative") <- assay(ex_set)
  expect_error(correct_drift(ex_set, 
    assay.type = "nope", name = "corrected"))
  # Only one name can be supplied to assay.type
  expect_error(correct_drift(ex_set,
    assay.type = c("original", "nope"), name = "corrected"))
  # Only one name can be supplied to transformed assay
  expect_error(correct_drift(ex_set,
    assay.type = "original", name = c("A", "B")))
})

test_plot_saving_helper <- function(object, func, title, func_args) {
  prefix <- paste0(tempdir(), "\\test\\")
  tmp <- object[1:5]
  rowData(tmp)$Metabolite_name <- c("Glucose", "Threoline", "5-AVAB", "1/2 acid", "20:0 carbon chain")
  do.call(func, c(list(drop_qcs(tmp), file_path = prefix, format = "emf", title = title), func_args))
  if (is.null(title)) {
    expect_equal(all(list.files(path = prefix) %in% paste0(rownames(tmp), ".emf")), TRUE)
  } else if (title == "Metabolite_name") {
    expect_equal(all(
      list.files(path = prefix) %in% paste0(
        gsub("[:/]", "_", rowData(tmp)$Metabolite_name),".emf")
    ), TRUE)
  } else {
    expect_equal(all(list.files(path = prefix) %in% paste0(rownames(tmp), ".emf")), TRUE)
  }
  unlink(prefix, recursive = TRUE)
}

test_file_extension_helper <- function(fileext) {
  p <- ggplot()
  file <- tempfile(fileext = fileext)
  save_plot(p, file)
  expect_equal(file.exists(file), TRUE)
  unlink(file)
}

# Used to combine and return multiple objects from a loop
.comb <- function(x, ...) {
  mapply(rbind, x, ..., SIMPLIFY = FALSE)
}

.calc_cubic_spline <- function(feature, sample_data, log_transform, 
                               spar, spar_lower, spar_upper) {
  # Spline cannot be fitted if there are les than 4 QC values
  qc_detected <- sample_data$QC == "QC" & !is.na(feature)
  if (sum(qc_detected) < 4) {
    return(rep(NA, length(feature)))
  }
  # Spline regression on the QC samples
  fit <- stats::smooth.spline(x = sample_data$Injection_order[qc_detected], 
                              y = feature[qc_detected], 
                              all.knots = TRUE, spar = spar, 
                              control.spar = 
                              list("low" = spar_lower, "high" = spar_upper))
  # Predicted values for all samples
  predicted <- stats::predict(fit, sample_data$Injection_order)$y
  # Substraction in log space, division in original space
  if (log_transform) {
    corrected <- feature + 
      mean(feature[qc_detected]) - predicted
  } else {
    corr_factors <- predicted[1] / predicted
    corrected <- feature * corr_factors
  }

  corrected
}

#' Fit a cubic spline to correct drift
#'
#' Corrects the drift in the features by applying smoothed cubic spline 
#' regression to each feature separately.
#'
#' @param object a SummarizedExperiment object
#' @param log_transform logical, should drift correction be done on 
#' log-transformed values? See Details
#' @param spar smoothing parameter
#' @param spar_lower,spar_upper lower and upper limits for the smoothing 
#' parameter
#'
#' @return A list including a SummarizedExperiment object
#' with drift corrected features and predicted = matrix of the predicted values 
#' by the cubic spline (used in visualization).
#'
#' @details If \code{log_transform = TRUE}, the correction will be done on 
#' log-transformed values.
#' The correction formula depends on whether the correction is run on original 
#' values or log-transformed values.
#' In log-space: \eqn{corrected = original + mean of QCs - 
#' prediction by cubic spline}.
#' In original space: \eqn{corrected = original * prediction for first QC / 
#' prediction for current point}.
#' We recommend doing the correction in the log-space since the log-transformed 
#' data better follows the assumptions of cubic spline regression. The drift 
#' correction in the original space also sometimes results in negative values, 
#' and results in rejection of the drift corrrection procedure.
#'
#' If \code{spar} is set to \code{NULL} (the default), the smoothing parameter 
#' will be separately chosen for each feature from the range 
#' [\code{spar_lower, spar_upper}] using cross validation.
#'
#' @examples
#' data(toy_notame_set)
#' dc <- dc_cubic_spline(toy_notame_set)
#' corrected <- dc$object
#'
#' @seealso  \code{\link[stats]{smooth.spline}} for details about the 
#' regression, \code{\link{inspect_dc}} for analysing the drift correction 
#' results, \code{\link{save_dc_plots}} for plotting the drift correction 
#' process for each feature
#'
#' @noRd
dc_cubic_spline <- function(object, log_transform = TRUE, spar = NULL,
                            spar_lower = 0.5, spar_upper = 1.5, 
                            assay.type = NULL, name = NULL) {
  # Start log
  log_text("Starting drift correction")
  # Zero values do not behave correctly
  full_data <- assay(object, assay.type)
  if (sum(full_data == 0, na.rm = TRUE)) {
    log_text(paste0("Zero values in feature abundances detected.",
                    " Zeroes will be replaced with 1.1."))
    full_data[full_data == 0] <- 1.1
  }
  
  # log-transform before fiting the cubic spline
  if (log_transform) {
    full_data <- log(full_data)
  }

  # Return both predicted values (for plotting) and drift corrected values
  dc_data <- BiocParallel::bplapply(
    as.data.frame(t(full_data)),
    .calc_cubic_spline,
    colData(object),
    log_transform,
    spar,
    spar_lower,
    spar_upper
  )
  corrected <- do.call(rbind, dc_data)
  colnames(corrected) <- colnames(full_data)
  # Inverse the initial log transformation
  if (log_transform) {
    corrected <- exp(corrected)
  }
  assay(object, name) <- corrected
  # Recompute quality metrics
  log_text("Recomputing quality metrics for drift corrected data")
  object <- assess_quality(object, assay.type = name)
  
  log_text("Drift correction performed")

  object
}


.help_inspect_dc <- function(dc, orig, check_quality, condition,
                             assay.orig, assay.dc) {
  orig_data <- assay(orig, assay.orig)
  dc_data <- assay(dc, assay.dc)
  features <- rownames(orig)
  rowData(dc)$DC_note <- NA
  missing_qcs <- apply(dc_data, 1, \(x) all(is.na(x)))
  rowData(dc)$DC_note[missing_qcs] <- "Missing_QCS"
  negative_dc <- apply(dc_data, 1, \(x) any(x < 0, na.rm = TRUE))
  rowData(dc)$DC_note[negative_dc & !missing_qcs] <- "Negative_DC"
  not_passing <- rep(FALSE, length(features))
  if (check_quality){
    qdiff <- quality(dc)[2:5] - quality(orig)[2:5] |> 
      as.data.frame() |> 
      t()
    pass <- paste0("qdiff |> dplyr::filter(", condition, ")") |>
      parse(text = _) |>
      eval() |>
      rownames()
    not_passing <- (features %in% !pass) & !(missing_qcs | negative_dc)
    rowData(dc)$DC_note[not_passing] <- "Low_quality"
  }
  dc_passing <- !(missing_qcs | negative_dc | not_passing)
  rowData(dc)$DC_note[dc_passing] <- "Drift_corrected"
  #  Replace the features that did not pass with the original values
  assay(dc, assay.dc)[!dc_passing, ] <- orig_data[!dc_passing, ]
  quality_cols <- c("RSD", "RSD_r", "D_ratio", "D_ratio_r")
  rowData(dc)[!dc_passing, quality_cols] <-
    rowData(orig)[!dc_passing, quality_cols]
  dc
}

#' Flag the results of drift correction
#'
#' Determines whether the drift correction worked.
#' The primary reason is to search for features where there were too many 
#' missing values in the QCs, so it was not possible to run drift correction. 
#' If the drift correction is run on the original values (not log-transformed), 
#' then there is also a need to check that the correction did not result
#' in any negative values. This can sometimes happen if the prediction curve 
#' takes an extreme shape.
#'
#' If quality is monitored, a quality condition is checked for each feature. 
#' If the condition is fulfilled, the drift corrected feature is retained,
#' otherwise the original feature is retained and the drift corrected feature 
#' is discarded. The result of this operation is recorded in the feature data.
#'
#' @param orig a SummarizedExperiment object, before drift correction
#' @param dc a SummarizedExperiment object, after drift correction
#' @param check_quality logical, whether quality should be monitored.
#' @param condition a character specifying the condition, see Details
#'
#' @return A SummarizedExperiment object.
#'
#' @details The \code{condition} parameter should be a character giving a 
#' condition compatible with \code{\link[dplyr]{filter}}. The condition is 
#' applied on the \strong{changes} in the quality metrics RSD, RSD_r, D_ratio 
#' and D_ratio_r. 
#' For example, the default is "RSD_r < 0 and D_ratio_r < 0",
#' meaning that both RSD_r and D_ratio_r need to decrease in the drift 
#' correction, otherwise the  drift corrected feature is discarded and the 
#' original is retained.
#'
#' @seealso \code{\link{correct_drift}}, \code{\link{save_dc_plots}}
#'
#' @examples
#' data(toy_notame_set)
#' dc <- dc_cubic_spline(toy_notame_set)
#' corrected <- dc$object
#' inspected <- inspect_dc(
#'   orig = toy_notame_set, dc = corrected,
#'   check_quality = TRUE
#' )
#'
#' @noRd
inspect_dc <- function(orig, dc, check_quality,
                       condition = "RSD_r < 0 & D_ratio_r < 0",
                       assay.orig = NULL, assay.dc = NULL) {
  log_text("Inspecting drift correction results")
  if (is.null(quality(orig))) {
    log_text("Original quality metrics missing, recomputing")
    orig <- assess_quality(orig, assay.type = assay.orig)
  }
  if (is.null(quality(dc))) {
    log_text("Drift corrected quality metrics missing, recomputing")
    dc <- assess_quality(dc, assay.type = assay.dc)
  }
  
  dc <- .help_inspect_dc(dc, orig, check_quality, condition,
                         assay.orig, assay.dc)

  # Log information
  dc_note <- rowData(dc)$DC_note
  note_counts <- table(dc_note) |> unname()
  note_percentage <- note_counts / sum(note_counts)
  note_percentage <- scales::percent(as.numeric(note_percentage))
  note_labels <- table(dc_note) |> names()
  report <- paste(note_labels, note_percentage, sep = ": ", collapse = ",  ")
  log_text(paste0("Drift correction results inspected: ", report))

  dc
}

#' Correct drift using cubic spline
#'
#' A wrapper function for applying cubic spline drift correction and saving
#' before and after plots.
#'
#' @param object a \code{
#' \link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
#' object
#' @param log_transform logical, should drift correction be done on 
#' log-transformed values? See Details
#' @param spar smoothing parameter as in 
#' \code{\link[stats]{smooth.spline}}
#' @param spar_lower,spar_upper lower and upper limits for the smoothing 
#' parameter
#' @param check_quality logical, whether quality should be monitored.
#' @param condition a character specifying the condition used to decide whether 
#' drift correction works adequately, see Details
#' @param file path to the PDF file where the plots should be saved
#' @param width,height width and height of the plots in inches
#' @param color character, name of the column used for coloring the points
#' @param shape character, name of the column used for shape
#' @param color_scale,shape_scale the color and shape scales as returned by a 
#' ggplot function
#' @param assay.type character, assay to be used in case of multiple assays
#' @param name character, name of the resultant assay
#'
#' @return A SummarizedExperiment object as the one supplied, with 
#' drift corrected features.
#'
#' @details If \code{log_transform = TRUE}, the correction will be done on 
#' log-transformed values.
#' The correction formula depends on whether the correction is run on original 
#' values or log-transformed values.
#' In log-space: \eqn{corrected = original + mean of QCs - prediction by cubic 
#' spline}.
#' In original space: \eqn{corrected = original * prediction for first QC / 
#' prediction for current point}.
#' We recommend doing the correction in the log-space since the log-transfomred 
#' data better follows the assumptions of cubic spline regression. The drift 
#' correction in the original space also sometimes results
#' in negative values, and results in rejection of the drift corrrection 
#' procedure.
#' If  \code{check_quality = TRUE}, the \code{condition} parameter should be a 
#' character giving a condition compatible with \code{\link[dplyr]{filter}}. 
#' The condition is applied on the \strong{changes} in the quality metrics
#' RSD, RSD_r, D_ratio and D_ratio_r. For example, the default is "RSD_r < 0 
#' and D_ratio_r < 0", meaning that both RSD_r and D_ratio_r need to decrease
#' in the drift correction, otherwise the drift corrected feature is discarded 
#' and the original is retained.
#' By default, the column used for color is also used for shape.
#'
#' @examples
#' data(toy_notame_set)
#' corrected <- correct_drift(mark_nas(toy_notame_set[1:5, ], value = 0))
#'
#' @seealso \code{\link[stats]{smooth.spline}} for details about the regression
#'
#' @export
correct_drift <- function(object, log_transform = TRUE, spar = NULL, 
                          spar_lower = 0.5, spar_upper = 1.5,
                          check_quality = FALSE, 
                          condition = "RSD_r < 0 & D_ratio_r < 0",
                          file = NULL, width = 16, 
                          height = 8, color = "QC", shape = color, 
                          color_scale = getOption("notame.color_scale_dis"),
                          shape_scale = scale_shape_manual(values = c(15, 16)),
                          assay.type = NULL, name = NULL) {
  from_to <- .get_from_to_names(object, assay.type, name)
  object <- .check_object(object, pheno_injection = TRUE, pheno_QC = TRUE, 
                          assay.type = from_to[[1]])
                          
  # Fit cubic spline and correct
  corrected <- dc_cubic_spline(object, log_transform = log_transform, 
                               spar = spar, spar_lower = spar_lower,
                               spar_upper = spar_upper, 
                               assay.type = from_to[[1]],
                               name = from_to[[2]])
  # Only keep corrected versions of features with increased quality
  inspected <- inspect_dc(orig = object, dc = corrected, 
                          check_quality = check_quality, condition = condition, 
                          assay.orig = from_to[[1]], assay.dc = from_to[[2]])
  # Return the final version
  inspected
}

---
title: "Non-targeted metabolomics preprocessing and data wrangling"
author: "Anton Klåvus, Vilhelm Suksi"
date: "`r Sys.Date()`"
output:
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
    number_sections: true
vignette: >
  %\VignetteIndexEntry{Non-targeted metabolomics preprocessing and wrangling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
biblio-style: apalike
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "##"
)
# To improve runtime on Windows
BiocParallel::register(BiocParallel::SerialParam())
```

# Motivation
From the perspective of metabolites as the continuation of the central dogma of 
biology, metabolomics provides the closest link to many phenotypes of interest. 
This makes metabolomics research promising in teasing apart the complexities of 
living systems, attracting many new practitioners.

The notame R package was developed in parallel with a protocol 
article as a general guideline for data analysis in untargeted 
metabolomics studies [@notame]. The notame package was split into notame, 
notameViz and notameStats for Bioconductor. 
The primary goal is identifying quality, interesting features for 
laborious downstream steps relating to biological context, such as 
metabolite identification and pathway analysis. Bioconductor packages with 
complementary functionality in Bioconductor include pmp, phenomis and qmtools; 
notame packages brings partially overlapping and new functionality to the 
table. There are also Bioconductor packages for preprocessing, metabolite 
identification and pathway analysis. Together, notame, Bioconductor's 
dependency management and other Bioconductor functionality allow for quality, 
reproducible metabolomics research.  Please see the [notame website](https://
hanhineva-lab.github.io/notame/) and protocol article for more information 
[@notame].

# Installation
To install notame, install BiocManager first, if it is not installed. 
Afterwards use the install function from BiocManager.

```{r, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("notame")
library(notame)
```

## Drift correction 
Untargeted LC-MS metabolomics data usually suffers from signal intensity drift, 
i.e. systematic variation in the signal intensity as a function of injection 
order. Features of a dataset often have very different drift patterns. This 
means that to correct for the drift, we need a tool that can adapt to different 
drift patterns.

The approach used in this package models the drift separately for each feature, 
using pooled QC samples and cubic spline regression. The smoothing parameter 
controls the "wiggliness" of the function: the larger the smoothing parameter, 
the closer the spline is to a linear function. By default, the smoothing 
parameter is chosen automatically using cross validation, and can vary between 
0.5 and 1.5, where 1.5 is large enough to guarantee that the curve is 
practically linear. This interval seems to work well in practice.

We run the drift correction on log-transformed data. Log-transformed data 
follows the assumptions of the drift function model better, and avoids some 
practical issues. The corrected feature abundance for a sample with injection 
order $i$ is then computed using the following formula:
$$x_{corrected}(i) = x_{original}(i) + \overline{x}_{QC} - f_{drift}(i)$$

```{r}
library(notame)
data(toy_notame_set, package = "notame")
se <- toy_notame_set
names(assays(se)) <- "rawbundances"
# Mark missing values as NA
se <- mark_nas(se, value = 0)
# Correct drift
se <- correct_drift(se, name = "dc")
```

## Flagging low quality compounds
LC-MS metabolomics datasets contain a significant number of low-quality 
features. We use three established quality metrics based on pooled QC samples 
to flag those features: detection rate, relative standard deviation and 
dispersion ratio[@qcguidelines]. We are using the robust versions of relative 
standard deviation and dispersion ratio as they are less affected by a single 
outlying QC and since the original RSD works best if the data is normally 
distributed, which is often not the case with our data. In addition, one can 
use `flag_contaminants()` to, well, flag contaminants. 

By default, we use a slightly more lenient detection threshold, 0.5. The 
recommended limits for RSD and D-ratio are 0.2 and 0.4, respectively. Since we 
are using the robust alternatives, we have added an additional condition: 
signals with classic RSD, RSD\* and basic D-ratio all below 0.1 are kept. This 
additional condition prevents the removal of signals with very low values in 
all but a few samples. These signals tend to have a very high value of D-
ratio\*, since the median absolute deviation of the biological samples is not 
affected by the large concentration in a handful of samples, causing the D-
ratio* to overestimate the significance of random errors in measurements of QC 
samples. Thus, other quality metrics are applied, with a conservative limit of 
0.1 to ensure that only good quality signals are kept this way.

```{r}
se <- flag_detection(se, assay.type = "dc")
se <- flag_quality(se, assay.type = "dc")
head(rowData(se)$Flag)
```

Flagged features are not removed from the data, but the flag information is 
stored in the "Flag" column of `rowData(se)`. In notame, notameViz and 
notameStats packages, the flagged features are silently ignored in operations 
that involve muliple features at a time, including random forest imputation, 
PCA visualization and multivariate statistical models. For feature-wise 
statistics, the statistical model is fit for flagged features, but FDR 
correction is only done for non-flagged features. One can also set 
`all_features = TRUE` in these functions to include all features in the model.

## Imputation of missing values
LC-MS data often contain missing values. Note that even if some statistical 
methods state that they "can deal with missing values", they sometimes just 
ignore samples or variables with at least one missing value, or impute the 
missing values before applying the actual method. 

There are many different imputation strategies out there. We have tested a 
bunch of them on our data and we concluded that random forest imputation works 
best for our data [@kokla2019random]. To be brief, the method predicts the 
missing part of the data by fitting a random forest on the observed part of the 
data [@missForest]. The downside of random forest imputation is that it is 
rather slow. Thus, we recommend running the imputation separately for each 
analysis mode if the dataset has more than 100 samples, or if there are lots of 
missing values.

```{r}
set.seed(2025)
se <- impute_rf(se, assay.type = "dc", name = "imputed")
```

Simple imputation strategies included in `impute_simple()` are imputation by 
constant value, imputation by mean, median, minimum or half minimum value or by 
a small random value. Simple imputation strategies rely on the assumption that 
missingness is mostly caused by the fact that the metabolite concentration is 
below the detection limit.

## Utilities
Data can be read with `import_from_excel()`, which includes checks and 
preparation of metadata. To accommodate typical output from peak-picking 
software such as Agilent's MassHunter or MS-DIAL, the output is transformed 
into a spreadsheet for `import_from_excel()`. Alternatively, data in R can be 
wrangled and passed to the `SummarizedExperiment()` constructor.

```{r, echo = FALSE}
#| fig.cap = "Structure of spreadsheet for import_from_excel()."
knitr::include_graphics("Data_input.png")
```

General utilities include `combined_data()` for representing the instance in a
`data.frame` suitable for plotting and various functions for data wrangling. 
For keeping track of the analysis, notame offers a logging system operated 
using `init_log()`, `log_text()` and `finish_log()`. notame also keeps track 
of all the external packages used, offering you references for each. To see and 
log a list of references, use `citations()`.

Parallellization is used in many feature-wise calculations and is provided by 
the BiocParallel package. BiocParallel defaults to a parallel backend. For 
small-scale testing on Windows, it can be quicker to use serial execution:

```{r, eval = FALSE} 
BiocParallel::register(BiocParallel::SerialParam())
```

# Authors & Acknowledgements

The first version of notame was written by Anton Klåvus for his master's 
thesis in Bioinformatics at Aalto university (published under former name Anton 
Mattsson), while working for University of Eastern Finland and Afekta 
Technologies. The package is inspired by analysis scripts written by Jussi 
Paananen and Oskari Timonen. The algorithm for clustering molecular features 
originating from the same compound is based on MATLAB code written by David 
Broadhurst, Professor of Data Science & Biostatistics in the School of Science, 
and director of the Centre for Integrative Metabolomics & Computational Biology 
at the Edith Covan University.

If you find any bugs or other things to fix, please submit an issue on GitHub! 
All contributions to the package are always welcome!

# Session information

```{r, echo = FALSE}
sessionInfo()
```

# References

---
title: "Project example"
author: "Anton Klåvus, Vilhelm Suksi"
date: "`r Sys.Date()`"
output:
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Project example}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
biblio-style: apalike
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  warning = FALSE,
  message = FALSE,
  results = FALSE,
  comment = "##"
)
# To improve runtime on Windows
BiocParallel::register(BiocParallel::SerialParam())
```

In this project example, core functionality of notame, notameViz and 
notameStats is demonstrated in unison with the associated protocol article 
[@notame].

# Project setup

## Set up path and logging
Let's set up a path and start logging our project. Many functions in the notame 
packages will log information when they have finished. This helps in monitoring 
the analysis process in case something goes wrong and in reviewing the results 
later. The log will also print to console, although we'll exclude console 
output for the rest of the document for brevity.

```{r, results = 'markup'}
library(notame)
library(notameViz)
library(notameStats)
library(dplyr)
ppath <- tempdir()
init_log(log_file = file.path(ppath, "log.txt"))
```

## Read data
The data is mock data with two balanced groups and two balanced time points. 
The time points also correspond to two batches. There are 50 samples with ten 
regularly interspersed QC samples and 80 features divided into four analytical 
modes.

```{r}
data <- import_from_excel(
  file = system.file("extdata", "toy_notame_set.xlsx", package = "notame"), 
  sheet = 1, corner_row = 11, corner_column = "H", 
  split_by = c("Column", "Ion_mode"))
  
names(assays(data)) <- "abundances"
```

The function `import_from_excel()` returns a SummarizedExperiment object with 
three main parts and corresponding accessors for streamlined data wrangling.

- `assay`: feature abundances across the samples  
- `colData`: sample information  
- `rowData`: feature information

The example data contains four analytical modes, so we separate them using 
`fix_object()`, which also conveniently checks the object for compatibility 
with all of notame. 

```{r}
modes <- fix_object(data, split_data = TRUE, assay.type = "abundances")
```

Now we have a list of objects, one per mode (LC column x 
ionization mode) or however many unique values were specified in
`rowData(data)$Split`.

```{r, results = 'markup'}
names(modes)
sapply(modes, class)
```

# Tabular data preprocessing

## Preprocessing by mode
Tabular data preprocessing is performed to complete the dataset by way of 
reducing unwanted variation and data preparation dependent on downstream 
methods. Steps performed separately for each mode include marking missing 
values as NA, flagging features with low detection rate [@qcguidelines], drift 
correction [@driftcorrection], and flagging of low-quality features 
[@qcguidelines]. Visualizations are drawn to monitor the process and explore 
the data globally. The `save_QC_plots()` wrapper does not include
feature-wise plots as it may not be feasible to inspect them at this 
stage of the analysis. For example, drift correction visualizations are best 
drawn for a subset of interesting features after feature selection.

```{r}
# Initialize empty list for processed objects
processed <- list()
for (i in seq_along(modes)) {
  name <- names(modes)[i]
  mode <- modes[[i]]
  # Set all zero abundances to NA
  mode <- mark_nas(mode, value = 0)
  # Flag features with low detection rate
  mode <- flag_detection(mode, group = "Group")
  # Visualize data before drift correction
  save_QC_plots(mode, prefix = paste0(ppath, "figures/", name, "_ORIG"),
                 perplexity = 5, group = "Group", time = "Time", 
                 id = "Subject_ID", color = "Group")
  # Correct drift
  corrected <- correct_drift(mode)
  # Visualize data after drift correction
  save_QC_plots(corrected, prefix = paste0(ppath, "figures/", name, "_DRIFT"),
                perplexity = 5, group = "Group", time = "Time", 
                id = "Subject_ID", color = "Group")
  # Flag low-quality features
  corrected <- corrected %>% 
    assess_quality() %>% 
    flag_quality()
  # Visualize data after removal of low-quality features
  save_QC_plots(corrected, prefix = paste0(ppath, "figures/", name, "_CLEAN"),
                perplexity = 5, group = "Group", time = "Time",
                id = "Subject_ID", color = "Group")
  # Save result of iteration
  processed[[i]] <- corrected
}
```

Feature-wise flagging information, quality metrics, and brief drift correction 
notes are included in the feature information after the above steps.

```{r, results = 'markup'}
rowData(processed[[1]])$DC_note
```

## Preprocessing for the complete dataset

Next, it is time to merge the modes together and visualize the complete dataset.

```{r}
merged <- merge_notame_sets(processed)
save_QC_plots(merged, prefix = paste0(ppath, "figures/_FULL"),
              group = "Group", time = "Time",
              id = "Subject_ID", color = "Group")
```

Then, QC samples can (and should) be removed, as they are no longer needed, and 
the dataset is visualized anew.

```{r}
merged_no_qc <- drop_qcs(merged)
save_QC_plots(merged_no_qc, prefix = paste0(ppath, "figures/FULL_NO_QC"),
              group = "Group", time = "Time",
              id = "Subject_ID", color = "Group")
```

If there are any missing values in the data, they need to be imputed. Let's use 
random forest imputation to impute these values and visualize the dataset one 
last time before statistical analysis. Seed number should be set before random 
forest imputation to guarantee reproducibility of results.

```{r}
set.seed(2024)
imputed <- impute_rf(merged_no_qc)
save_QC_plots(imputed, prefix = paste0(ppath, "figures/FULL_IMPUTED"),
              group = "Group", time = "Time",
              id = "Subject_ID", color = "Group")
```

By default, the imputation procedure only operates on good quality features, 
i.e. those that have not been flagged. To use flagged features in statistical 
analysis, they should be imputed as well. This can be achieved through a second 
round of imputation, now with all features included. This two-step imputation 
makes sure that low-quality features don't affect the imputation of quality 
features. Imputation could also be performed separately for each mode to reduce 
execution time, especially if the amounts of features still allows for good 
imputation results.

```{r}
base <- impute_rf(imputed, all_features = TRUE)
```

It is a good idea to save the merged and processed data, so experimenting with 
different statistical analyses becomes easier.

```{r}
save(base, file = paste0(ppath, "full_data.RData"))
```

# Feature selection

## Univariate analysis
First, we'll use a linear model to evaluate the features in terms of the 
probability of obtaining the observed abundances given the null hypothesis, 
namely that there is no difference in feature abundance across study groups.

```{r}
assay(base, "log") <- log(assay(base))
lm_results <- perform_lm(base, assay.type = "log", 
  formula_char = "Feature ~ Group")
base <- join_rowData(base, lm_results)
```

Next, univariate and supervised rankings of features could be combined in a 
final ranking. Such a final ranking would combine the qualities of univariate 
analysis and supervised learning. Next, carefully selected features would 
undergo labour-intensive scrutiny for biological meaning, for example by 
identification and pathway analysis.

That concludes our project example. The last thing to do is to write the 
results to an Excel file and finish the logging. Thanks!

```{r}
write_to_excel(base, file = paste0(ppath, "results.xlsx"))
finish_log()
```

# Session information

```{r, echo = FALSE, results = 'markup'}
sessionInfo()
```

# References

# Contributions
Contributors and maintainers are expected to abide by the [Bioconductor Code of Conduct](https://bioconductor.github.io/bioc_coc_multilingual/en-US.html). 

When creating an issue, use a clear and descriptive title, detail the issue and context in a verbose manner and include code for demonstration. 

We follow a git flow approach for changes. Fork and clone the repository. Create a new branch and develop against current Bioc-devel version. Document documentation changes using devtools::document(). Check and resolve errors and warnings from devtools::check() and BiocCheck::BiocCheck(). Make sure that your branch is synced with the target branch and create a pull request. Detail the changes and context in a verbose manner and tag one of the maintainers for review. Also check the automatic build reports in the pull request.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/feature_clustering.R
\name{cluster_features}
\alias{cluster_features}
\title{Cluster correlated features originating from the same metabolite}
\usage{
cluster_features(
  object,
  mz_col = NULL,
  rt_col = NULL,
  all_features = FALSE,
  rt_window = 1/60,
  corr_thresh = 0.9,
  d_thresh = 0.8,
  assay.type = NULL
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{mz_col}{the column name in feature data that holds mass-to-charge 
ratios}

\item{rt_col}{the column name in feature data that holds retention times}

\item{all_features}{logical, should all features be included in the 
clustering? If FALSE, as the default, flagged features are not included in 
clustering}

\item{rt_window}{the retention time window for potential links
NOTE: use the same unit as the retention time}

\item{corr_thresh}{the correlation threshold required for potential links 
between features}

\item{d_thresh}{the threshold for the relative degree required by each node}

\item{assay.type}{character, assay to be used in case of multiple assays}
}
\value{
a SummarizedExperiment object, with median peak area 
(MPA), the cluster ID, the features in the cluster, and cluster size added 
to feature data.
}
\description{
Clusters features potentially originating from the same compound. Features 
with high Pearson correlation coefficient and small retention time 
difference are linked together. Then clusters are formed by setting a 
threshold for the relative degree that each node in a cluster needs
to fulfil. Each cluster is named after the feature with the highest median 
peak area (median abundance). This is a wrapper around numerous functions 
that are based on the MATLAB code by David Broadhurst.
}
\examples{
data(toy_notame_set)
# The parameters are really weird because example data is imaginary
clustered <- cluster_features(toy_notame_set, rt_window = 1,
  corr_thresh = 0.5, d_thresh = 0.6)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{fix_object}
\alias{fix_object}
\title{Fix object for functioning of notame}
\usage{
fix_object(
  object,
  id_prefix = "ID_",
  id_column = NULL,
  split_by = NULL,
  name = NULL,
  clean = TRUE,
  split_data = FALSE,
  assay.type = NULL
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{id_prefix}{character, prefix for autogenerated sample IDs, see Details}

\item{id_column}{character, column name for unique identification of samples}

\item{split_by}{character vector, in the case where all the modes are in the 
same object, the column names of feature data used to separate the modes 
(usually Mode and Column)}

\item{name}{in the case where object only contains one mode, the 
name of the mode, such as "Hilic_neg"}

\item{clean}{boolean, whether to select best classes, reorder columns and 
consistently rename columns in pheno and feature}

\item{split_data}{logical, whether to split data by analytical mode recorded 
in the "Split" column of feature data. If TRUE (the default), will return a 
list of objects, one per analytical mode. If FALSE, will return a 
single object.}

\item{assay.type}{character, assay to be used in case of multiple assays}
}
\value{
A new SummarizedExperiment object with a single peak table. If 
split_data = TRUE, a list containing separate objects for analytical modes.
}
\description{
Attempts to create missing columns needed for notame in pheno and feature 
data. Optionally cleans the object and splits the object by mode.
}
\details{
Only specify one of \code{split_by} and \code{name}. The feature 
data will contain columns named "Split", used to separate features from 
different modes, and "Flag" for recording flagged features. Unless a column 
named "Feature_ID" is found in feature data, a feature ID will be generated 
based on the value of "Split", mass and retention time. The function will 
try to find columns for mass and retention time by looking at a few common 
alternatives, and throw 
an error if no matching column is found. Sample information needs to contain 
a row called "Injection_order", and the values need to be unique. In 
addition, a possible sample identifier row needs to be named "Sample_ID",
or to be specified in \code{id_column}, and the values need to be unique, 
with an exception of QC samples: if there are any "QC" identifiers, they 
will be replaced with "QC_1", "QC_2" and so on.
If a "Sample_ID" column is not found, it will be created using the 
\code{id_prefix} and injection order or by renaming \code{id_column}.
}
\examples{
data(toy_notame_set)
ex_set <- toy_notame_set
rowData(ex_set)$Flag <- NULL
fixed <- fix_object(ex_set)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{drop_flagged}
\alias{drop_flagged}
\title{Drop flagged features}
\usage{
drop_flagged(object, all_features = FALSE)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{all_features}{logical, should all features be retained? Mainly used by 
internal functions}
}
\value{
A SummarizedExperiment object without the previously 
flagged features.
}
\description{
Removes all features that have been flagged by quality control functions.
Only features that do not have a flag (Flag == NA) are retained.
}
\examples{
data(toy_notame_set)
dim(toy_notame_set)
flagged <- flag_quality(toy_notame_set)
noflags <- drop_flagged(flagged)
dim(noflags)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{flag_report}
\alias{flag_report}
\title{A report of flagged features}
\usage{
flag_report(object)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}
}
\value{
A data frame with the number of features at each stage of flagging.
}
\description{
Computes the number of features at each stage of flagging for each mode.
}
\examples{
data(toy_notame_set)
flagged <- toy_notame_set |>
  mark_nas(0) |>
  flag_detection(group = "Group") |>
  flag_quality()
flag_report(flagged)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{package}
\name{notame-package}
\alias{notame}
\alias{notame-package}
\title{\code{notame} package.}
\description{
Provides tabular data preprocessing and data wrangling functionality for 
untargeted LC-MS metabolomics research.
}
\references{
Klåvus et al. (2020). "notame": Workflow for Non-Targeted LC-MS Metabolic 
Profiling. Metabolites, 10: 135.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/hanhineva-lab/notame}
  \item \url{https://hanhineva-lab.github.io/notame/}
  \item Report bugs at \url{https://github.com/hanhineva-lab/notame/issues}
}

}
\author{
\strong{Maintainer}: Vilhelm Suksi \email{vksuks@utu.fi} (\href{https://orcid.org/0009-0005-1108-518X}{ORCID})

Authors:
\itemize{
  \item Anton Klåvus (\href{https://orcid.org/0000-0003-2612-0230}{ORCID}) [copyright holder]
  \item Jussi Paananen (\href{https://orcid.org/0000-0001-5100-4907}{ORCID}) [copyright holder]
  \item Oskari Timonen (\href{https://orcid.org/0000-0002-6317-6260}{ORCID}) [copyright holder]
  \item Atte Lihtamo
  \item Retu Haikonen (\href{https://orcid.org/0000-0003-0830-3850}{ORCID})
  \item Leo Lahti (\href{https://orcid.org/0000-0001-5537-637X}{ORCID})
  \item Kati Hanhineva (\href{https://orcid.org/0000-0001-6834-7375}{ORCID})
}

Other contributors:
\itemize{
  \item Ville Koistinen (\href{https://orcid.org/0000-0003-1587-8361}{ORCID}) [contributor]
  \item Olli Kärkkäinen (\href{https://orcid.org/0000-0003-0825-4956}{ORCID}) [contributor]
  \item Artur Sannikov (\href{https://orcid.org/0000-0001-7765-123X}{ORCID}) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{impute_rf}
\alias{impute_rf}
\title{Impute missing values using random forest}
\usage{
impute_rf(object, all_features = FALSE, assay.type = NULL, name = NULL, ...)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{all_features}{logical, should all features be used? If FALSE (the 
default), flagged features are removed before imputation.}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{name}{character, name of the resultant assay in case of multiple assays}

\item{...}{passed to \code{\link[missForest]{missForest}}}
}
\value{
An object as the one supplied, with missing 
values imputed.
}
\description{
Impute the missing values in the peak table of the object using a
random forest. The estimated error in the imputation is logged.
It is recommended to set the seed number for reproducibility
(it is called random forest for a reason).
This a wrapper around \code{\link[missForest]{missForest}}.
Use parallelize = "variables" to run in parallel for faster testing.
NOTE: running in parallel prevents user from setting a seed number.
}
\examples{
data(toy_notame_set)
missing <- mark_nas(toy_notame_set, 0)
set.seed(38)
imputed <- impute_rf(missing)

}
\seealso{
\code{\link[missForest]{missForest}} for detail about the algorithm 
and the parameters
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_correction.R
\name{perform_repeatability}
\alias{perform_repeatability}
\title{Compute repeatability measures}
\usage{
perform_repeatability(object, group, assay.type = NULL)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{group}{column name of pheno data giving the group labels}

\item{assay.type}{character, assay to be used in case of multiple assays}
}
\value{
A data frame with one row per feature with the repeatability measure.
}
\description{
Computes repeatability for each feature with the following formula:
\deqn{\frac{\sigma^2_{between}}{\sigma^2_{between} + \sigma^2_{within}}}
The repeatability ranges from 0 to 1. Higher repeatability depicts less
variation between batches.
}
\examples{
data(toy_notame_set)
# Batch correction
replicates <- list(which(toy_notame_set$QC == "QC"))
batch_corrected <- ruvs_qc(toy_notame_set, replicates = replicates)
# Evaluate batch correction
rep_orig <- perform_repeatability(toy_notame_set, group = "Group")
mean(rep_orig$Repeatability, na.rm = TRUE)
rep_corr <- perform_repeatability(batch_corrected, group = "Group")
mean(rep_corr$Repeatability, na.rm = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{mark_nas}
\alias{mark_nas}
\title{Mark specified values as missing}
\usage{
mark_nas(object, value, assay.type = NULL, name = NULL)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{value}{the value to be converted to NA}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{name}{character, name of the resultant assay in case of multiple assays}
}
\value{
SummarizedExperiment object as the one supplied, with 
missing values correctly set to NA.
}
\description{
Replaces all values in the peak table that equal the specified value 
with NA.
For example, vendor software might use 0 or 1 to signal a missing value,
which is not understood by R.
}
\examples{
data(toy_notame_set)
nas_marked <- mark_nas(toy_notame_set, value = 0)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{citations}
\alias{citations}
\title{Show citations}
\usage{
citations()
}
\value{
None, the function is invoked for its side effect.
}
\description{
This function lists citations behind the notame functions that have been 
called during the session. All notame functions 
update the list automatically. The citations are taken from the call to 
'\code{citation("package")}, and complemented with a brief description of 
what the package was used for.
NOTE: the citations might not point to the correct paper if the package 
authors have not supplied correct citation information for their package.
The output is written to the current log file, if specified.
}
\examples{
citations()
data(toy_notame_set)
ex_set <- flag_quality(toy_notame_set)
# Broadhurst et al.(2018) added to citations
citations()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality_metrics.R
\name{quality}
\alias{quality}
\title{Extract quality information of features}
\usage{
quality(object)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}
}
\value{
A data frame with quality metrics for each feature.
}
\description{
Extract quality information of features
}
\examples{
data(toy_notame_set)
ex_set <- assess_quality(toy_notame_set)
quality(ex_set)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{write_to_excel}
\alias{write_to_excel}
\title{Write results to Excel file}
\usage{
write_to_excel(object, file, ...)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{file}{path to the file to write}

\item{...}{Additional parameters passed to
\code{\link[openxlsx]{write.xlsx}}}
}
\value{
None, the function is invoked for its side effect.
}
\description{
Writes all the data in a SummarizedExperiment object to an 
Excel spreadsheet.
The format is similar to the one used to read data in, except for the fact 
that EVERYTHING NEEDS TO BE WRITTEN AS TEXT. To fix numeric values in Excel,
choose any cell with a number, press Ctrl + A, then go to the dropdown menu
in upper left corner and choose "Convert to Number". This will fix the file,
but can take quite a while.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
data(toy_notame_set)
write_to_excel(toy_notame_set, file = "toy_notame_set.xlsx")
\dontshow{setwd(.old_wd)}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{join_rowData}
\alias{join_rowData}
\title{Join new columns to feature data}
\usage{
join_rowData(object, dframe)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}}

\item{dframe}{a data frame with the new information}
}
\value{
A SummarizedExperiment object with the new information added to 
rowData(object).
}
\description{
Join a new data frame of information to feature data of a 
SummarizedExperiment object. The data frame needs to have a column 
"Feature_ID". This function is usually used internally by some of the 
functions in the package, but can be useful.
}
\examples{
data(toy_notame_set)
new_info <- data.frame(
  Feature_ID = rownames(toy_notame_set),
  Feature_number = seq_len(nrow(toy_notame_set))
)
with_new_info <- join_rowData(toy_notame_set, new_info)
colnames(rowData(with_new_info))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{pqn_normalization}
\alias{pqn_normalization}
\title{Probabilistic quotient normalization}
\usage{
pqn_normalization(
  object,
  ref = c("qc", "all"),
  method = c("median", "mean"),
  all_features = FALSE,
  assay.type = NULL,
  name = NULL
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{ref}{character, the type of reference samples to use for normalization.}

\item{method}{character, the method to use for calculating the reference 
sample.}

\item{all_features}{logical, should all features be used for calculating the 
reference sample?}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{name}{character, name of the resultant assay in case of multiple assays}
}
\value{
A SummarizedExperiment object with altered feature 
abundances.
}
\description{
Apply probabilistic quotient normalization (PQN) to the peak table of a 
SummarizedExperiment object. By default, reference is 
calculated from high-quality QC samples and the median of the reference is 
used for normalization. Check parameters for more options.
}
\examples{
data(toy_notame_set)
pqn_set <- pqn_normalization(toy_notame_set)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{join_colData}
\alias{join_colData}
\title{Join new columns to pheno data}
\usage{
join_colData(object, dframe)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{dframe}{a data frame with the new information}
}
\value{
A SummarizedExperiment object with the new information added to 
colData(object).
}
\description{
Join a new data frame of information to pheno data of a SummarizedExperiment 
object.
}
\examples{
data(toy_notame_set)
new_info <- data.frame(
  Sample_ID = colnames(toy_notame_set),
  BMI = stats::runif(ncol(toy_notame_set), 22, 26)
)
with_new_info <- join_colData(toy_notame_set, new_info)
colnames(colData(with_new_info))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logging.R
\name{log_text}
\alias{log_text}
\title{Log text to the current log file}
\usage{
log_text(text)
}
\arguments{
\item{text}{The text to be logged}
}
\value{
None, the function is invoked for its side effect.
}
\description{
The specified text is printed and written to the current log file. Does not 
overwrite the file. Also used internally by many functions in the package.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_name <- "log.txt"
init_log(file_name)
log_text("Hello World!")
# Print the contents of the file
scan(file_name, sep = "\n", what = "character")
\dontshow{setwd(.old_wd)}

}
\seealso{
\code{\link{init_log}}, \code{\link{finish_log}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality_metrics.R
\name{flag_detection}
\alias{flag_detection}
\title{Flag features with low detection rate}
\usage{
flag_detection(
  object,
  qc_limit = 0.7,
  group_limit = 0.5,
  group = NULL,
  assay.type = NULL
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{qc_limit}{the detection rate limit for QC samples}

\item{group_limit}{the detection rate limit for study groups}

\item{group}{the columns name in sample information to use as the grouping 
variable}

\item{assay.type}{character, assay to be used in case of multiple assays}
}
\value{
A SummarizedExperiment object with the features flagged.
}
\description{
Flags features with too high amount of missing values. There are two 
detection rate limits, both defined as the minimum proportion of samples 
that need to have a value (not NA) for the feature to be kept. 
'\code{qc_limit} is the detection rate limit for QC samples, 
'\code{group_limit} is the detection rate limit for the actual study groups.
If the group limit is passed for AT LEAST ONE GROUP, then the feature is 
kept. Features with low detection rate in QCs are flagged as 
"Low_qc_detection", while low detection rate in the study groups is flagged 
as "Low_group_detection". The detection rates for all the groups are 
recorded in feature data.
}
\examples{
data(toy_notame_set)
ex_set <- mark_nas(toy_notame_set, value = 0)
ex_set <- flag_detection(ex_set, group = "Group")
rowData(ex_set)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality_metrics.R
\name{assess_quality}
\alias{assess_quality}
\title{Assess quality information of features}
\usage{
assess_quality(object, assay.type = NULL)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{assay.type}{character, assay to be used in case of multiple assays}
}
\value{
A SummarizedExperiment object with quality metrics in 
feature data.
}
\description{
Assess features using the quality metrics defined in (Broadhurst 
2018). The quality metrics are described in Details section of 
\code{\link{flag_quality}}
}
\examples{
data(toy_notame_set)
ex_set <- assess_quality(toy_notame_set)
rowData(ex_set)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{import_from_msdial}
\alias{import_from_msdial}
\title{Import data from a TSV file created by MS-DIAL}
\usage{
import_from_msdial(
  file,
  id_column = NULL,
  id_prefix = "ID_",
  split_by = NULL,
  name = NULL,
  mz_limits = c(10, 2000),
  rt_limits = c(0, 20),
  skip_checks = FALSE
)
}
\arguments{
\item{file}{path to the TSV file}

\item{id_column}{character, column name for unique identification of samples}

\item{id_prefix}{character, prefix for autogenerated sample IDs}

\item{split_by}{character vector, in the case where all the modes are in the 
same Excel file, the column names of feature data used to separate the modes 
(usually Mode and Column)}

\item{name}{in the case where the Excel file only contains one mode, the 
name of the mode, such as "Hilic_neg"}

\item{mz_limits}{numeric vector of two, all m/z values should be in between 
these}

\item{rt_limits}{numeric vector of two, all retention time values should be 
in between these}

\item{skip_checks}{logical: skip checking and fixing of data integrity. Not 
recommended, but sometimes useful when you just want to read the data in as 
is and fix errors later. The data integrity checks are important for 
functioning of notame.}
}
\value{
A SummarizedExperiment with assays, rowData and colData filled from
the TSV file.
}
\description{
Reads a tab-delimited peak table (TSV) exported from MS-DIAL. This works 
similarly to \code{\link{import_from_excel}}, but removes the need to 
manually open and re-save the TSV files in Excel. Excludes any summary
statistics computed by MS-DIAL, such as average abundances, since these are
not valid after notame processing.
}
\examples{
\donttest{
# Read a TSV peak table
# data <- import_from_msdial(file = "path/to/peak_table.tsv", name = "RP_neg")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{fix_MSMS}
\alias{fix_MSMS}
\title{Transform the MS/MS output to publication ready}
\usage{
fix_MSMS(
  object,
  ms_ms_spectrum_col = "MS_MS_spectrum",
  peak_num = 10,
  min_abund = 5,
  deci_num = 3
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{ms_ms_spectrum_col}{name of column with original MS/MS spectra}

\item{peak_num}{maximum number of peak that is kept (Recommended: 4-10)}

\item{min_abund}{minimum relative abundance to be kept (Recommended: 1-5)}

\item{deci_num}{maximum number of decimals to m/z value (Recommended: >2)}
}
\value{
A SummarizedExperiment object as the one supplied, with 
publication-ready MS/MS peak information.
}
\description{
Change the MS/MS output from MS-DIAL format to publication-ready format.
Original spectra is sorted according to abundance percentage and clarified. 
See the example below.
}
\details{
Original MS/MS spectra from MS-DIAL:
m/z:Raw Abundance

23.193:254 26.13899:5 27.50986:25 55.01603:82 70.1914:16 73.03017:941 
73.07685:13 73.13951:120

Spectra after transformation:
m/z  (Abundance)

73.03 (100), 23.193 (27), 73.14 (12.8), 55.016 (8.7)
}
\examples{
data(toy_notame_set)
# Spectra before fixing
ex_set <- toy_notame_set
rowData(ex_set)$MS_MS_spectrum <- NA
rowData(ex_set)[1, ]$MS_MS_spectrum <- 
  "28.769:53 44.933:42 52.106:89 69.518:140"
rowData(ex_set)$MS_MS_spectrum[
  !is.na(rowData(ex_set)$MS_MS_spectrum)]
# Fixing spectra with default settings
fixed_MSMS_peaks <- fix_MSMS(ex_set)
# Spectra after fixing
rowData(fixed_MSMS_peaks)$MS_MS_Spectrum_clean[
  !is.na(rowData(fixed_MSMS_peaks)$MS_MS_Spectrum_clean)]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{impute_simple}
\alias{impute_simple}
\title{Simple imputation}
\usage{
impute_simple(object, value, na_limit = 0, assay.type = NULL, name = NULL)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{value}{the value used for imputation, either a numeric or one of 
'"min", "half_min", "small_random", see above}

\item{na_limit}{only impute features with the proportion of NAs over this 
limit. For example, if \code{na_limit = 0.5}, only features with at least 
half of the values missing are imputed.}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{name}{character, name of the resultant assay in case of multiple assays}
}
\value{
A SummarizedExperiment object with imputed peak table.
}
\description{
Impute missing values using a simple imputation strategy. All missing values
of a feature are imputed with the same value. It is possible
to only impute features with a large number of missing values this way. This 
can be useful for using this function before random forest imputation to 
speed things up.
The imputation strategies available are:
\itemize{
\item a numeric value: impute all missing values in all features with the 
same value, e.g. 1
\item "mean": impute missing values of a feature with the mean of observed 
values of that feature
\item "median": impute missing values of a feature with the median of 
observed values of that feature
\item "min": impute missing values of a feature with the minimum observed 
value of that feature
\item "half_min": impute missing values of a feature with half the minimum 
observed value of that feature
\item "small_random": impute missing values of a feature with random numbers 
between 0 and the
minimum of that feature (uniform distribution, remember to set the seed 
number!).
}
}
\examples{
data(toy_notame_set)
missing <- mark_nas(toy_notame_set, 0)
imputed <- impute_simple(missing, value = "min")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drift_correction.R
\name{correct_drift}
\alias{correct_drift}
\title{Correct drift using cubic spline}
\usage{
correct_drift(
  object,
  log_transform = TRUE,
  spar = NULL,
  spar_lower = 0.5,
  spar_upper = 1.5,
  check_quality = FALSE,
  condition = "RSD_r < 0 & D_ratio_r < 0",
  file = NULL,
  width = 16,
  height = 8,
  color = "QC",
  shape = color,
  color_scale = getOption("notame.color_scale_dis"),
  shape_scale = scale_shape_manual(values = c(15, 16)),
  assay.type = NULL,
  name = NULL
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{log_transform}{logical, should drift correction be done on 
log-transformed values? See Details}

\item{spar}{smoothing parameter as in 
\code{\link[stats]{smooth.spline}}}

\item{spar_lower, spar_upper}{lower and upper limits for the smoothing 
parameter}

\item{check_quality}{logical, whether quality should be monitored.}

\item{condition}{a character specifying the condition used to decide whether 
drift correction works adequately, see Details}

\item{file}{path to the PDF file where the plots should be saved}

\item{width, height}{width and height of the plots in inches}

\item{color}{character, name of the column used for coloring the points}

\item{shape}{character, name of the column used for shape}

\item{color_scale, shape_scale}{the color and shape scales as returned by a 
ggplot function}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{name}{character, name of the resultant assay}
}
\value{
A SummarizedExperiment object as the one supplied, with 
drift corrected features.
}
\description{
A wrapper function for applying cubic spline drift correction and saving
before and after plots.
}
\details{
If \code{log_transform = TRUE}, the correction will be done on 
log-transformed values.
The correction formula depends on whether the correction is run on original 
values or log-transformed values.
In log-space: \eqn{corrected = original + mean of QCs - prediction by cubic 
spline}.
In original space: \eqn{corrected = original * prediction for first QC / 
prediction for current point}.
We recommend doing the correction in the log-space since the log-transfomred 
data better follows the assumptions of cubic spline regression. The drift 
correction in the original space also sometimes results
in negative values, and results in rejection of the drift corrrection 
procedure.
If  \code{check_quality = TRUE}, the \code{condition} parameter should be a 
character giving a condition compatible with \code{\link[dplyr]{filter}}. 
The condition is applied on the \strong{changes} in the quality metrics
RSD, RSD_r, D_ratio and D_ratio_r. For example, the default is "RSD_r < 0 
and D_ratio_r < 0", meaning that both RSD_r and D_ratio_r need to decrease
in the drift correction, otherwise the drift corrected feature is discarded 
and the original is retained.
By default, the column used for color is also used for shape.
}
\examples{
data(toy_notame_set)
corrected <- correct_drift(mark_nas(toy_notame_set[1:5, ], value = 0))

}
\seealso{
\code{\link[stats]{smooth.spline}} for details about the regression
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{inverse_normalize}
\alias{inverse_normalize}
\title{Inverse-rank normalization}
\usage{
inverse_normalize(object, assay.type = NULL, name = NULL)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{name}{character, name of the resultant assay in case of multiple assays}
}
\value{
An object as the one supplied, with normalized features.
}
\description{
Applies inverse rank normalization to all features to approximate
a normal distribution.
}
\examples{
data(toy_notame_set)
normalized <- inverse_normalize(toy_notame_set)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_correction.R
\name{ruvs_qc}
\alias{ruvs_qc}
\title{Remove Unwanted Variation (RUV) between batches}
\usage{
ruvs_qc(object, replicates, k = 3, assay.type = NULL, name = NULL, ...)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{replicates}{list of numeric vectors, indexes of replicates}

\item{k}{The number of factors of unwanted variation to be estimated from 
the data.}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{name}{character, name of the resultant assay in case of multiple assays}

\item{...}{other parameters passed to \code{link[RUVSeq]{RUVs}}}
}
\value{
A SummarizedExperiment object with the normalized data.
}
\description{
An interface for \code{link[RUVSeq]{RUVs}} method
}
\examples{
data(toy_notame_set)
# Batch correction
replicates <- list(which(toy_notame_set$QC == "QC"))
batch_corrected <- ruvs_qc(toy_notame_set, replicates = replicates)
# Evaluate batch correction
pca_bhattacharyya_dist(toy_notame_set, batch = "Batch")
pca_bhattacharyya_dist(batch_corrected, batch = "Batch")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{combined_data}
\alias{combined_data}
\title{Retrieve both sample information and features}
\usage{
combined_data(object, assay.type = NULL)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{assay.type}{character, assay to be used in case of multiple assays}
}
\value{
A data frame with sample information plus all features as columns,
one row per sample.
}
\description{
Retrieve both sample information and features
}
\examples{
data(toy_notame_set)
combined_data(toy_notame_set)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{drop_qcs}
\alias{drop_qcs}
\title{Drop QC samples}
\usage{
drop_qcs(object)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}
}
\value{
A SummarizedExperiment object as the one supplied, 
without QC samples.
}
\description{
Drop QC samples
}
\examples{
data(toy_notame_set)
dim(toy_notame_set)
noqc <- drop_qcs(toy_notame_set)
dim(noqc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logging.R
\name{init_log}
\alias{init_log}
\title{Initialize log to a file}
\usage{
init_log(log_file)
}
\arguments{
\item{log_file}{Path to the log file}
}
\value{
None, the function is invoked for its side effect.
}
\description{
Initialize a log file with the current data and time.
All major operations run after this will be logged to the specified file.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_name <- "log.txt"
init_log(file_name)
# Print the contents of the file
scan(file_name, sep = "\n", what = "character")
\dontshow{setwd(.old_wd)}
}
\seealso{
\code{\link{log_text}}, \code{\link{finish_log}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logging.R
\name{finish_log}
\alias{finish_log}
\title{Finish a log}
\usage{
finish_log()
}
\value{
None, the function is invoked for its side effect.
}
\description{
Logs the current date and time and session info, and switches logging off.
}
\examples{
finish_log()

}
\seealso{
\code{\link{init_log}}, \code{\link{log_text}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_correction.R
\name{pca_bhattacharyya_dist}
\alias{pca_bhattacharyya_dist}
\title{Bhattacharyya distance between batches in PCA space}
\usage{
pca_bhattacharyya_dist(
  object,
  batch,
  all_features = FALSE,
  center = TRUE,
  scale = "uv",
  nPcs = 3,
  assay.type = NULL,
  ...
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{batch}{column name of pheno data giving the batch labels}

\item{all_features}{logical, should all features be used? If FALSE
(the default), flagged features are removed before imputation.}

\item{center}{logical, should the data be centered prior to PCA?
(usually yes)}

\item{scale}{scaling used, as in 
\code{\link[pcaMethods]{prep}}. Default is "uv" for unit variance}

\item{nPcs}{the number of principal components to use}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{...}{other parameters to \code{\link[pcaMethods]{pca}}}
}
\value{
A matrix of Bhattacharyya distances between batches.
}
\description{
Computes Bhattacharyya distance between all pairs of batches using 
\code{\link[fpc:bhattacharyya.matrix]{fpc:bhattacharyya.matrix}} 
after projecting the samples into PCA space with 
\code{\link[pcaMethods]{pca}}.
}
\examples{
data(toy_notame_set)
# Batch correction
replicates <- list(which(toy_notame_set$QC == "QC"))
batch_corrected <- ruvs_qc(toy_notame_set, replicates = replicates)
# Evaluate batch correction
pca_bhattacharyya_dist(toy_notame_set, batch = "Batch")
pca_bhattacharyya_dist(batch_corrected, batch = "Batch")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{toy_notame_set}
\alias{toy_notame_set}
\alias{hilic_neg_sample}
\alias{hilic_pos_sample}
\alias{rp_neg_sample}
\alias{rp_pos_sample}
\title{Toy data set}
\format{
An object of class \code{SummarizedExperiment} with 80 rows and 50 columns.

An object of class \code{SummarizedExperiment} with 20 rows and 50 columns.

An object of class \code{SummarizedExperiment} with 20 rows and 50 columns.

An object of class \code{SummarizedExperiment} with 20 rows and 50 columns.

An object of class \code{SummarizedExperiment} with 20 rows and 50 columns.
}
\usage{
toy_notame_set

hilic_neg_sample

hilic_pos_sample

rp_neg_sample

rp_pos_sample
}
\description{
Contains imaginary data used in testing the package functions. The dataset 
has 50 samples and 80 features. This dataset includes multiple observations 
from same subjects, sampled at two timepoints in separate batches and 
divided to two groups. The analytical modes are also available as separate 
SummarizedExperiment objects.
Note that across batches, the features don't have different feature ID's, 
m/z and retention time as would be the case with real-world data. In 
essence, the example data reflects that features were aligned perfectly 
between batches.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality_metrics.R
\name{flag_contaminants}
\alias{flag_contaminants}
\title{Flag contaminants based on blanks}
\usage{
flag_contaminants(
  object,
  blank_col,
  blank_label,
  blank_type = c("mean", "median", "max"),
  sample_type = c("max", "median", "mean"),
  flag_thresh = 5,
  flag_label = "Contaminant",
  assay.type = NULL
)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{blank_col}{character, the column name in colData with blank labels}

\item{blank_label}{character, the label for blank samples in blank_col}

\item{blank_type}{character, one of "mean", "median", or "max"}

\item{sample_type}{character, one of "max", "median", or "mean"}

\item{flag_thresh}{numeric, the scaled ratio threshold for flagging 
contaminants.
If blank_type(blanks) * flag_thresh > sample_type(biological samples), the 
feature gets flagged.}

\item{flag_label}{character, the label used when flagging contaminants. Can 
be changed if sample processing contaminants and carryover contaminants are 
flagged separately.}

\item{assay.type}{character, assay to be used in case of multiple assays}
}
\value{
A SummarizedExperiment object with contaminant features 
flagged.
}
\description{
Flags contaminant features by comparing either median, mean or max of blanks 
and biological samples. Biological samples are defined as samples that are 
not marked as blanks and are not QCs.
}
\details{
If the calculation(biological samples) < the calculation(blanks) times a set 
ratio, the feature is flagged as contaminant. Default calculations are "max" 
for biological samples and "mean" for blanks.
}
\examples{
data(toy_notame_set)
# Make a blank sample which has one (first) feature exceeding the threshold
## Abundance matrix
max_sample <- max(assay(toy_notame_set)[1, toy_notame_set$QC != "QC"])
assay <- matrix(c(max_sample * 0.20 + 1, rep(0, 79)), ncol = 1, nrow = 80, 
                  dimnames = list(NULL, "Demo_51"))
assay <- cbind(assay(toy_notame_set), assay)
## Sample metadata
pheno_data <- colData(toy_notame_set)[1, ]
rownames(pheno_data) <- "Demo_51"
pheno_data$Sample_ID <- "Demo_51"
pheno_data$Injection_order <- 51
pheno_data[c("Subject_ID", "Group", "QC", "Time")] <- "Blank"
pheno_data <- rbind(colData(toy_notame_set), pheno_data)
## Feature metadata
feature_data <- rowData(toy_notame_set)

# Construct SummarizedExperiment object with blank sample
ex_set <- SummarizedExperiment(assays = assay, 
                               colData = pheno_data,
                               rowData = feature_data)
# Flag contaminant(s)
contaminants_flagged <- flag_contaminants(ex_set, blank_col = "QC", 
                                          blank_label = "Blank")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quality_metrics.R
\name{flag_quality}
\alias{flag_quality}
\title{Flag low-quality features}
\usage{
flag_quality
flag_quality(object, assay.type = NULL, condition =
  "(RSD_r < 0.2 & D_ratio_r < 0.4) | 
  (RSD < 0.1 & RSD_r < 0.1 & D_ratio < 0.1)")
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{assay.type}{character, assay to be used in case of multiple assays}

\item{condition}{character, condition for keeping the features, see Details}
}
\value{
a SummarizedExperiment object with the features flagged.
}
\description{
Flags low-quality features using the quality metrics defined in (Broadhurst 
2018). The metrics are described in more detain in Details. A condition for 
keeping the features is given as a character, which is passed to 
\code{\link[dplyr]{filter}}.
}
\details{
The quality metrics measure two things: internal spread of the QCs,
and spread of the QCs compared to the spread of the biological samples.
Internal spread is measured with relative standard deviation (RSD), also 
known as coefficient of variation (CV).
\deqn{RSD = sd(QC) / mean(QC) }
Where \eqn{sd(QC)} is the standard deviation of the QC samples and 
'\eqn{mean(QC)} is the sample mean of the signal in the QC samples.
RSD can also be replaced by a non-parametric, robust version based on the 
median and median absolute deviation (MAD):
  \deqn{RSD_r = 1.4826 * MAD(QC) / median(QC)}
The spread of the QC samples compared to the biological samples is measured 
using a metric called D-ratio:
  \deqn{D_ratio = sd(QC) / sd(biological)}
Or, as before, a non-parametric, robust alternative:
  \deqn{D_ratio_r = MAD(QC) / MAD(biolofical) }
The default condition keeps features that pass either of the two following 
conditions:
\deqn{RSD_r < 0.2 \& D_ratio_r < 0.4}
\deqn{RSD < 0.1 \& RSD_r < 0.1 \& D_ratio < 0.1}
}
\examples{
data(toy_notame_set)
ex_set <- flag_quality(toy_notame_set)
rowData(ex_set)
# Custom condition
ex_set <- flag_quality(toy_notame_set,
  condition = "RSD_r < 0.3 & D_ratio_r < 0.6")
rowData(ex_set)

}
\references{
Broadhurst, David et al. Guidelines and considerations for the 
use of system suitability and quality control samples in mass spectrometry 
assays applied in untargeted clinical metabolomic studies.
Metabolomics : Official journal of the Metabolomic Society vol. 14,6 (2018): 
72. doi:10.1007/s11306-018-1367-3
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{import_from_excel}
\alias{import_from_excel}
\title{Read formatted Excel files}
\usage{
import_from_excel(
  file,
  sheet = 1,
  id_column = NULL,
  corner_row = NULL,
  corner_column = NULL,
  id_prefix = "ID_",
  split_by = NULL,
  name = NULL,
  mz_limits = c(10, 2000),
  rt_limits = c(0, 20),
  skip_checks = FALSE
)
}
\arguments{
\item{file}{path to the Excel file}

\item{sheet}{the sheet number or name}

\item{id_column}{character, column name for unique identification of samples}

\item{corner_row}{integer, the bottom row of sample information,
usually contains data file names and feature info column names.
If set to NULL, will be detected automatically.}

\item{corner_column}{integer or character, the corresponding column number 
or the column name (letter) in Excel.
If set to NULL, will be detected automatically.}

\item{id_prefix}{character, prefix for autogenerated sample IDs}

\item{split_by}{character vector, in the case where all the modes are in the 
same Excel file, the column names of feature data used to separate the modes 
(usually Mode and Column)}

\item{name}{in the case where the Excel file only contains one mode, the 
name of the mode, such as "Hilic_neg"}

\item{mz_limits}{numeric vector of two, all m/z values should be in between 
these}

\item{rt_limits}{numeric vector of two, all retention time values should be 
in between these}

\item{skip_checks}{logical: skip checking and fixing of data integrity. Not 
recommended, but sometimes useful when you just want to read the data in as 
is and fix errors later. The data integrity checks are important for 
functioning of notame.}
}
\value{
A SummarizedExperiment with assays, rowData and colData filled from
the Excel file.
}
\description{
Reads data from an Excel file of the following format:
\itemize{
  \item Left side of the sheet contains information about the features, size 
features x feature info columns
  \item Top part contains sample information, size sample info variables x 
samples
  \item The middle contains the actual abundances, size features x samples
}
Exported data from MS-DIAL converted to an excel file creates such format.
}
\examples{
data <- import_from_excel(
  file = system.file("extdata", "toy_notame_set.xlsx", 
  package = "notame"), sheet = 1, corner_row = 11, corner_column = "H",
  split_by = c("Column", "Ion_mode"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/feature_clustering.R
\name{compress_clusters}
\alias{compress_clusters}
\title{Compress clusters of features to a single feature}
\usage{
compress_clusters(object)
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}
}
\value{
A SummarizedExperiment object with only one feature per 
cluster.
}
\description{
This function compresses clusters found by cluster_features, keeping only 
the feature with the highest median peak area. The features that were 
discarded are recorded in feature data, under Cluster_features.
}
\examples{
data(toy_notame_set)
clustered <- cluster_features(toy_notame_set, 
  rt_window = 1, corr_thresh = 0.5, d_thresh = 0.6)
compressed <- compress_clusters(clustered)

}
\seealso{
\code{\link{cluster_features}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merger.R
\name{merge_notame_sets}
\alias{merge_notame_sets}
\title{Merge SummarizedExperiment objects together}
\usage{
merge_notame_sets(..., merge = c("features", "samples"), assay.type = NULL)
}
\arguments{
\item{...}{\code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
objects or a list of objects}

\item{merge}{what to merge? features is used for combining analytical modes,
samples is used for batches}

\item{assay.type}{character, assay to be used in case of multiple assays.
The same assay needs to be present in all objects to be merged, and the
resultant object contains this single assay.}
}
\value{
A merged SummarizedExperiment object.
}
\description{
Merges two or more SummarizedExperiment objects together. Can be used to
merge analytical modes or batches.
}
\details{
When merging samples, sample IDs that begin with "QC" or "Ref" are
combined so that they have running numbers on them. This means that if both
batches have samples called "QC_1", this will not result in an error,
but the sample IDs will be adjusted so that they are unique. Column names in
the feature data that are shared between batches but have different content
are renamed by adding a suffix to avoid data loss. The suffix is the index
of the batch in the input list.
}
\examples{
# Merge analytical modes
data(hilic_neg_sample, hilic_pos_sample, rp_neg_sample, rp_pos_sample)
merged <- merge_notame_sets(
  hilic_neg_sample, hilic_pos_sample,
  rp_neg_sample, rp_pos_sample
)
# Merge batches
batch1 <- toy_notame_set[, toy_notame_set$Batch == 1]
batch2 <- toy_notame_set[, toy_notame_set$Batch == 2]
merged <- merge_notame_sets(batch1, batch2, merge = "samples")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class_constructor.R
\name{flag}
\alias{flag}
\alias{flag<-}
\title{Get and set the values in the flag column}
\usage{
flag(object)

flag(object) <- value
}
\arguments{
\item{object}{a \code{
\link[SummarizedExperiment:SummarizedExperiment-class]{SummarizedExperiment}}
object}

\item{value}{character vector, values for flag column}
}
\value{
Character vector of feature flags.

For the endomorphism, an object with a modified flag column.
}
\description{
Get and set the values in the flag column
}
\examples{
data(toy_notame_set)
# Get values in flag column of rowData
flag(toy_notame_set)

data(toy_notame_set)
# Flag a suspicious feature manually
flag(toy_notame_set)[1] <- "Contaminant, known from experience"
}
